/*************************************************************************/
/*                        program eigenray.c                             */
/*************************************************************************/
/*
 *      Copyright (C) 1992,  Woods Hole Oceanographic Institution
 *    See LICENSE file for use, copying and distribution conditions
 *
 *
 *  jim bowlin may 1993.
 *
 * Changes: Rich Pawlowicz Aug 93
 *           - create_eigenangles_geo *completely* rewritten so that it
 *             now actually finds eigenrays (i.e. it shoots more rays
 *             until the correct angle is found). Diffracted ray search
 *             will now no longer work; I suppose it should be fixed
 *             eventually but I do not see a need right now.
 */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#include "utils.h"
#include "ray.h"
#include "outray.h"
#include "eigenray.h"

#define GOLD_R 0.61803399
#define GOLD_C ( 1.0 - GOLD_R)

time_t      start_timing(long n);                               /* ray.c     */
void        end_timing(long n,time_t start_time);               /* ray.c     */
rinfos      *shoot_all_angles(params *param);

/************************************************************************/
/*                     airy function                                    */
/************************************************************************/

#define Max_airy 8

static double X3n[Max_airy];
static double Airy_c1[] = {
    0.355028,
   0.0591713,
  0.00197238,
 2.73941e-05,
 2.07531e-07,
 9.88245e-10,
 3.22956e-12,
 7.68942e-15,	
};

static double Airy_c2[] = {
    0.258819,
   0.0215683,
 0.000513531,
  5.7059e-06,
 3.65763e-08,
 1.52401e-10,
 4.45617e-13,
 9.64539e-16,
};

double airy(double x)
/*
 * Returns airy(x)
 */
{
double ret,absx,x_4,x_32,x3;
int    i;
	
	absx = fabs(x);
	x_4  = pow(absx,-0.25) / sqrt(Pi);
	x_32 = 2.0/3.0 * pow(absx,1.5);
   if (x < -3.0) 
		return( x_4 * cos( x_32 - Pi/4));

	if ( x > 3.0) 
		return( exp( - x_32 ) * x_4 /2.0);

	ret = 0;
	x3   = x * x * x;
	X3n[0] = 1.0;
	for ( i = 1; i < Max_airy; i++)
		X3n[i] = X3n[i-1] * x3;
	for(i = Max_airy -1 ; i >= 0; i--) {
		ret += X3n[i] * (Airy_c1[i] - x * Airy_c2[i]);
	}
	return(ret);
}

double airy2(double x)
{
	x = airy(x);
	return( x * x);
}

/************************************************************************/
/*                 end of airy function                                 */
/************************************************************************/

/************************************************************************/
/*                    rinfo merge                                       */
/************************************************************************/


void merge_rinfos(rinfos **first,rinfos *link, rinfos *new)
{
rinfos *next;

	while( new != NULL){
		next  = new->next;
		merge_1rinfo(first,link,new);
		link = new;
		new = next;
	}
}

void merge_1rinfo(rinfos **first,rinfos *rinfo, rinfos *new)
/*
 *  Merges new into first.  Returns a pointer to the new start of first.
 *  Sorted by source angle.  We will be speedy if the two lists start out
 *  sorted.
 *  ASSUMES first list is already sorted by sangle.
 */
{
rinfos *next;
double angle;

   if (first == NULL) return;
	if (*first == NULL) {
		*first = new;
		return;
	}
   if (  new == NULL) return;
   if (rinfo == NULL) rinfo = *first;
	
	angle = new->sangle;
	if ( angle < rinfo->sangle) {        /* search backward */
		while( angle < rinfo->sangle) { 
			rinfo = rinfo->prev;
			if (rinfo == NULL) break;
		}
		if ( rinfo == NULL) {        /* then link at head */
			new->prev      = NULL;
			new->next      = *first;
			(*first)->prev = new;
			*first = new;
			return;
		}
		next        = rinfo->next;
		new->prev   = rinfo;
		new->next   = next;
		rinfo->next = new;
		if ( next != NULL) next->prev  = new;
		return;
	}
		
	next = rinfo->next;
	while( next != NULL)  {  /* search forward  ... */
		if ( next->sangle > angle) break;
		rinfo = next;
		next = rinfo->next;            
	}  
	if ( next == NULL) {              /* then link at tail */
		new->prev     = rinfo;
		new->next     = NULL;
		rinfo->next   = new;
		return;
	}
	new->prev   = rinfo;
	new->next   = rinfo->next;
	rinfo->next = new;
	next->prev  = new;
}

/************************************************************************/
/*                end of rinfo merge                                    */
/************************************************************************/

rinfos *prime_2rinfos(rinfos *rinfo,rinfos **rinfo1,rinfos **rinfo2,double r0)
{
   if (rinfo == NULL) return(rinfo);
   while (rinfo->rrange != r0) {
      rinfo = rinfo->next;
      if (rinfo == NULL) return(rinfo);
   }
   *rinfo1 = rinfo;
   do{
      rinfo  = rinfo->next;
      if (rinfo == NULL) return(rinfo);
   }  while (rinfo->rrange != r0);
   *rinfo2 = rinfo;
   return(rinfo);
}

void add_a_dub(dubs **first,dubs **last,double value)
/*
 *   Adds a dubs structure to the list starting with *first
 *   and ending with *last.
 */
{
dubs *dub;

      Safe_calloc(dub,dubs,1);
      if (*first == NULL) {
         *first = dub;
         *last = dub;
      }
      else {
         (*last)->next = dub;
         *last = dub;
      }
      dub->d = value;
}

int fit_parab(double p[3],double x1,double x2,double x3,
                           double y1,double y2,double y3)
/*
 *   Inputs:  x1,x2,x3, y1,y2,y3
 *   Outputs: p[3]
 *   Finds    p[] such that
 *   fits (yi - p[2]) = p[0] * (xi - p[1])^2
 *                y0      R           x0
 */
{
double dy0  = y1 - y2;
double dy2  = y3 - y2;
double dx01 = x1 - x2;
double dx21 = x3 - x2;
double dx02 = x1 - x3;
double f1,f2,dx0;

   if (dx02 ==0.0 || dx01 == 0.0 || dx21 == 0.0) return(1);
   f2  = ( dy0 / dx01 - dy2 /dx21 ) / dx02;
   f1  = 0.5 * (dy0 / dx01 + dy2 / dx21 - f2 * (dx01 + dx21));

   dx0 = - f1 / (2.0 * f2);
   p[0]  = f2;
   p[1]  = dx0 + x2;
   p[2]  = y2 + dx0 * ( f1 + dx0 * f2);
   return(0);
}

void fit_rinfo_parab(double p[],rinfos *left,rinfos *mid,rinfos *right)
{
   fit_parab(p,left->sangle,mid->sangle,right->sangle,
				   left->rdepth,mid->rdepth,right->rdepth);
}

double geo_angle(rinfos *rinfo1,rinfos *rinfo2,double rdepth)
{
double depth1 = rinfo1->rdepth;
double depth2 = rinfo2->rdepth;
double lambda;
   lambda = (fabs(depth2 -depth1) > 1.0e-6) ?
            (depth2 - rdepth)/(depth2 - depth1) : 0.5;
   return(lambda * rinfo1->sangle + (1.0 - lambda) * rinfo2->sangle);
}

/************************************************************************/
/*                   golden search                                      */
/************************************************************************/

int check_dz(params *param,double oldz[],double olda[],
				 int i0,int i1,int i2,int i3)
/*
 * Checks angles to see if min spacing is less than min_da.
 * Checks depths to see if maximum distance from a parabolic
 * fit extremum is less than min_dz.
 * If either is true we return true (1) else return false (0).
 */ 
{
double  min_dz    = param->add_extrema.min_dz;
double  min_da    = param->add_extrema.da;
int     min_cnt   = param->add_extrema.min_cnt;
double z[4],a[3],p[3];
double z0,da1,da2,da;
int    i;
int    cnt = 0;

   z[0] = oldz[i0]; z[1] = oldz[i1]; z[2] = oldz[i2];z[3] = oldz[i3];
   a[0] = olda[i0]; a[1] = olda[i1]; a[2] = olda[i2];
	
	da1 = fabs(a[0] - a[1]);
	da2 = fabs(a[1] - a[2]);
	da  = Min(da1,da2);
	if ( da < min_da) return(1);
	if ( min_dz == 0.0) return(0);	
	fit_parab(p,a[0],a[1],a[2],z[0],z[1],z[2]);
	z0 = p[2];
	for (i = 0; i < 4; i++) /* count how many are within min_z of z0 */
		if ( fabs(z[i] - z0) < min_dz) cnt++;
	
	if ( cnt >= min_cnt) return(1);
	return(0);
}
		
int golden_search(params *param,rinfos *left,rinfos *mid,rinfos *right,
						rinfos **first,rinfos **link)

/*
 * Searches for an extrema ( either max or min ) in the wavefront
 * given 3 rinfos: *left, *mid, *right.  Parameters as to when to
 * stop searching are in *param. The rinfo from each ray that is
 * shot is added to the rinfo linked list starting at **first.
 * Returns the number of new rays shot.
 */
{
double  minmax,extreme;
double  z[4],a[4];
double  lz,mz,rz,ma,la,ra;
int     iter;
int     max_iter  = param->add_extrema.max_iter;
rinfos  *new;
int     count  = 1;

   lz = left->rdepth; mz = mid->rdepth; rz = right->rdepth;
   la = left->sangle; ma = mid->sangle; ra = right->sangle;
   extreme = (mz - lz);
   if (extreme == 0.0) extreme = (mz - rz);
   if (extreme == 0.0) return(0);

   minmax = ( extreme > 0.0 ) ? 1.0 : -1.0;

   z[0] = lz;
   z[3] = rz;
   a[0] = left->sangle;
   a[3] = right->sangle;
   if ( fabs( mz - lz) > fabs(mz - rz)) {
      a[1] = ma + GOLD_C * (la - ma);
      a[2] = ma;
      new   = shoot_one_ray(&Param, a[1],1,0);
		merge_1rinfo(first,*link,new);
      z[1]  = new->rdepth;
      z[2]  = mid->rdepth;
   }
   else {
      a[1] = ma;
      a[2] = ma + GOLD_C * (ra - ma);
      new  = shoot_one_ray(&Param, a[2],1,0);
		merge_1rinfo(first,*link,new);		
      z[1]  = mid->rdepth;
      z[2]  = new->rdepth;
   }

   for ( iter = 2; iter <= max_iter; iter++) {
      if (  (z[1] - z[2]) * minmax  < 0.0) {
			if (check_dz(param,z,a,1,2,3,0) ) break;
         SHIFT3(a[0],a[1],a[2]);
         a[2] = GOLD_R * a[1] + GOLD_C * a[3];
         SHIFT3(z[0],z[1],z[2]);
         new =  shoot_one_ray(&Param, a[2],iter,0);
			merge_1rinfo(first,*link,new);		
         z[2] = new->rdepth;
      }
      else {
			if (check_dz(param,z,a,0,1,2,3) ) break;
         SHIFT3(a[3],a[2],a[1]);
         a[1] = GOLD_R * a[2] + GOLD_C * a[0];
         SHIFT3(z[3],z[2],z[1]);
         new =  shoot_one_ray(&Param, a[1],iter,0);
			merge_1rinfo(first,*link,new);		
         z[1] = new->rdepth;
      }
		count++;
   }
	*link = new;
   return(count);
}

/************************************************************************/
/*                 end of golden search                                 */
/************************************************************************/

/************************************************************************/
/*                 geometric eigenrays                                  */
/************************************************************************/

void eigen_warning(rinfos *r1,rinfos *r2,char *s)
{
   if ( s == NULL) s = "unknown error";   
   fprintf(Log_file, "%s",s);
   fprintf(Log_file, " between angles, %10.10f and %10.10f\n",
                       r1->sangle * 57.29578, r2->sangle * 57.29578);
   if ( ! Verbose) return;
   fprintf(stdout, "%s",s);
   fprintf(stdout, " between angles, %10.10f and %10.10f\n",
                       r1->sangle * 57.29578, r2->sangle * 57.29578);
}

void eigen_error(rinfos *r1,rinfos *r2,char *s)
{
   eigen_warning(r1,r2,s);
   vlputs("search aborted.\n");
}
   
void geo_eigensearch(params *param,rinfos *r1,rinfos *r2,double z0,
                     searches *search)
{
rinfos  *rinfo;
int      search_cnt = 1;
double   angle,dz,dz1,dz2;
double   geo_miss   = param->eigen.geo_miss;
double   r0         = param->rrange;
int      search_max = param->eigen.search_max;
   
   if( ( dz1 = z0 - r1->rdepth) * ( dz2 = z0 - r2->rdepth) > 0 ) return;
/*
 * Check for lucky hit.  If so add best angle to search list and return.
 */
   if ( dz1 < 0) {                /* swap r1,r2   dz1,dz2 */
      SHIFT4(rinfo,r1,r2,rinfo);
      SHIFT4(dz,dz1,dz2,dz);
   }
         
   dz1 = fabs(dz1);
   dz2 = fabs(dz2);
   dz  = Min(dz1,dz2);
   if (dz < geo_miss) {
      angle = ( dz1 < dz2) ? r1->sangle : r2->sangle;
      search->angle_count++;
      add_a_dub(&search->first_angle,&search->last_angle,angle);
      return;
   }

   for(;;){

/*  Half way between binary search and linear interpolation. */

      angle = geo_angle(r1,r2,z0) * 0.5 + (r1->sangle + r2->sangle) * 0.25;
      rinfo = shoot_one_ray(param, angle,search_cnt,0);
      search->numshot++;

      if ( rinfo->rrange != r0) {
         eigen_error(r1,r2,"ray did not reach receiver");
         return;
      }
      dz = z0 - rinfo->rdepth;
      if (fabs(dz) < geo_miss) {
         search->angle_count++;
         add_a_dub(&search->first_angle,&search->last_angle,angle);
         return;
      }
      if ( search_cnt++ > search_max) {
         eigen_error(r1,r2,"search_max exceeded");
         return;
      }
      if ( dz > dz1 && dz > dz2 ) 
         eigen_warning(r1,r2,"depth not linear with angle");
		
      if ( dz > 0 ) {
			r1 = rinfo; dz1 = dz;
		}
      else {
        r2 = rinfo; dz2 = dz;
		}
   }
}

void find_1eigenrays_geo(params *param,rinfos *rinfo,double z0,searches *search)
{
rinfos   *left,*mid;
double    r0 = param->rrange;


   rinfo = prime_2rinfos(rinfo, &left, &mid,r0);
   if (rinfo == NULL) return;
   for(;;) {
      if ( (rinfo  = rinfo->next) == NULL) break;
      if (rinfo->rrange != r0) continue;
      geo_eigensearch(param,left,mid,z0,search);
      SHIFT3(left,mid,rinfo);
   }     
   geo_eigensearch(param,left,mid,z0,search);
}

int find_eigenrays_geo(params *param,rinfos *rinfo)
{
dubs     *rdepths = param->rdepths;
time_t   start_time;
searches  search  = {NULL,NULL,0,0};

   param->saving_pinfo = 0;	
   start_time=start_timing((long)0);
   if (rdepths == NULL)
		 find_1eigenrays_geo(param,rinfo,param->rdepth,&search);
	else {
	for(; rdepths !=NULL; rdepths = rdepths->next)
		find_1eigenrays_geo(param,rinfo,rdepths->d,&search);
	}
   end_timing(search.numshot,start_time);
   param->angles     = search.first_angle;
   param->num_angles = search.angle_count;	
	return(search.angle_count);
}

/************************************************************************/
/*                 end of geometric eigenrays                           */
/************************************************************************/

/************************************************************************/
/*                   shadow zone rays                                   */
/************************************************************************/

/*
 *                    cos(sangle)    
 * shadow amp =    --------------    2 pi sqrt(ka/rc) airy(ka z_perp)
 *                 r0 cos(rangle)   
 *
 *   ka = ( 2 k0^2/r_mean)^(1/3)
 *   r_mean = rc * rr / (rc + rr)
 *   rc derived from fitting parabola z - z0 = rc ( sangle - sangle0)^2
 *   rr = c/ ( dc/dz) is local curvature of rays.
 */

rinfos *create_shadowray(params *param,
								 rinfos *left,rinfos * mid, rinfos *right)
/*
 * Assumes left,mid, right describe a local extremum in the wave front.
 * and that there is a shadow zone with the receiver in it. Creates a
 * new rinfo structure with parameters modified to descibe the shadow
 * zone arrival.
 */
{
rinfos *rinfo;
double p[3],cof[3],r_c,z_c;
 	
	if( left==NULL || mid==NULL || right==NULL) return(NULL);
	/** don't try to make shadow zone if # of bounces changes **/
	
	if (left->top_bonks != mid->top_bonks) return(NULL);
	if (left->bot_bonks != mid->bot_bonks) return(NULL);
	if (left->top_bonks != right->top_bonks) return(NULL);
	if (left->bot_bonks != right->bot_bonks) return(NULL);

   Safe_calloc(rinfo,rinfos,1);
	Memcopy(rinfo,mid,sizeof(rinfos));
	fit_rinfo_parab(p,left,mid,right);
	r_c           = p[0];
	rinfo->sangle = p[1];
   z_c           = p[2];
/*   get_cofs(cof,left,mid,right,rinfo); */

	return(rinfo);
}
		
int find_shadowrays(params *param,rinfos **first)
{
rinfos  *rinfo = *first;
rinfos  *left,*mid,*right;
int     count  = 0;
double  z0     = param->rdepth;
double  r0     = param->rrange;
double  dmiss  = param->eigen.diff_miss;
double  dz0,dz1,dz2;
rinfos *shad,*link = NULL,*list = NULL;

   if (rinfo == NULL) return(0);
   param->saving_pinfo = 0;
	
   rinfo = prime_2rinfos(rinfo, &mid, &right,r0);
   if ( rinfo == NULL) return(0);
   for(;;) {
      rinfo = rinfo->next;
      if (rinfo == NULL) break;
      if (rinfo->rrange != r0) continue;
      SHIFT4(left,mid,right,rinfo);

      dz1  = mid->rdepth - left->rdepth;
		dz2  = right->rdepth - mid->rdepth;
		dz0   = z0 - left->rdepth;
		
/* if slope of both segments are the same  - or - 
 * if we are curved in the wrong direction then no shadowray
 */			      
      if ( dz1 * dz2 >  0.0) continue;      
      if ( dz1 * dz0  < 0.0) continue;
      if( fabs( z0 - mid->rdepth ) > dmiss) continue;
		
      shad = create_shadowray(param,left,mid,right);
		merge_1rinfo(&list,link,shad);
		link = shad;
		count++;
   }
   return(count);
}

/************************************************************************/
/*                end ofshadow zone rays                                */
/************************************************************************/

void find_eigenrays(params *param,rinfos **first)
{
int     count;
rinfos *eigenrays;

   if ( param->eigen.diff_miss > 0.0) {
      vlputs("Looking for shadow rays, shooting extra rays ...\n");
		count = find_shadowrays(param,first);
		fprintf(Log_file,"%d shadowrays were found.\n",count);
		if(Verbose) printf("%d shadowrays were found.\n",count);
	}
	vlputs("Looking for eigenrays, shooting extra rays ...\n");	
   count = find_eigenrays_geo(param,*first);
	fprintf(Log_file,"%d eigenrays were found.\n",count);
	if(Verbose) printf("%d eigenrays were found.\n",count);
   if (count == 0) return;
   eigenrays = shoot(&Eigenfan,param->vname.eigenrays,Eigenfan.selected);
   add_geo_amplitudes(param,eigenrays);  
   save_rinfo(Mat_file,Asc_file,eigenrays,Param.vname.eigenfront);
}

int add_extrema(params *param,rinfos **first)
/*
 *   Looks for local minima and maxima of receiver depth in the
 *   rinfo list.  The trys to extremize these local mins and maxes.
 *   This should allow for courser fan spacing.
 */
{
rinfos  *left,*right,*mid;
rinfos  *rinfo    = *first;
double  r0        = param->rrange;
double  z0        = param->rdepth;
double  z_miss    = param->add_extrema.z_miss;
int    count     = 0;
double dz1,dz2,dz0,p[3];
rinfos *list = NULL,*link = NULL;

   param->saving_pinfo = 0;
   rinfo = prime_2rinfos(rinfo, &(mid), &(right),r0);
   if ( rinfo == NULL) return(0);
   for(;;) {
      if ( (rinfo  = rinfo->next) == NULL) break;
      if (rinfo->rrange != r0) continue;
      SHIFT4(left,mid,right,rinfo);
      dz1 = mid->rdepth - left->rdepth;
		dz2 = right->rdepth - mid->rdepth;
		dz0   = z0 - left->rdepth;
		
      if( dz1 * dz2  >= 0.0)    /* are slopes of both segments the same ? */
			continue;
		
		if ( z_miss > 0.0) {      /* if user specified z_miss */
		   if ( dz1 * dz0  < 0.0) /* are we pointed in right direction ? */
				continue;
			fit_rinfo_parab(p,left,mid,right);
         /* p[2] = z0, a guess at the depth of the extremum */
			if ( fabs( p[2] - z0) > z_miss)  
				continue;
		}
      count += golden_search(param,left,mid,right,&list,&link);
   }
	merge_rinfos(first,*first,list);
   return(count);
}

/**************************************************************************/
/*               geometric Amplitude section                              */
/**************************************************************************/

void add_1geo_amplitude(rinfos *rinfo1,rinfos *rinfo,rinfos *rinfo2,double r0)
/*
 *                    cos(sangle) | sangle2 - sangle1 |
 *       amp =    -------------------------------------
 *                 r0 cos(rangle) | rdepth2 - rdepth1 |
 */
{
double   a,z,atten,intens,denom,numer;
double   dz1 = 0.0;
double   dz2 = 0.0;
double   da1 = 0.0;
double   da2 = 0.0;

   if ( rinfo == NULL) return;
   if ( rinfo->rrange != r0) return;
   z  = rinfo->rdepth;
   a  = rinfo->sangle;
   if ( rinfo1 != NULL) {
      if ( rinfo1->rrange == r0) {
         dz1 = rinfo1->rdepth - z;
         da1 = rinfo1->sangle - a;
      }
   }
   
   if ( rinfo2 != NULL) {
      if ( rinfo2->rrange == r0) {
         dz2 = z - rinfo2->rdepth;
         da2 = a - rinfo2->sangle;        
      }
   }
   
   denom =  fabs(dz1) + fabs(dz2);
   denom *= r0 *  cos( rinfo->rangle);
   
   if ( da1 * da2 < 0.0)  
      ldbug("Warning angles not bracketed in \"add_1geo_amplitude()\"");

   numer =  ( fabs(da1) + fabs(da2) ) * cos(rinfo->sangle);
   if ( denom > 1.0e-40 ) intens = numer / denom;
   else intens = 1.0e40;
   atten = pow( 10.0, -fabs(rinfo->bottom_loss)/10.0 );
   rinfo->geo_intens = intens;   
   rinfo->tot_intens = intens * atten;
	rinfo->tot_amp = sqrt( rinfo->tot_intens * rinfo->c_at_r / rinfo->c_at_s);
}

void add_all_geo_amps(params *param,rinfos *rinfo)
/*
 *   For every ray in the rinfo list we calculate the geometric
 *   arrival amplitude using the two neighboring rays.
 */
{
rinfos *rinfo1,*rinfo2,*rinfo3;
double r0     = param->rrange;

   rinfo = prime_2rinfos(rinfo,&rinfo2,&rinfo3,r0);
   add_1geo_amplitude(NULL,rinfo2,rinfo3,r0);    
   if (rinfo == NULL) return;
   for( ;; ) {
      if ( (rinfo  = rinfo->next) == NULL) break;
      if (rinfo->rrange != r0) continue;
      SHIFT4(rinfo1,rinfo2,rinfo3,rinfo);
      add_1geo_amplitude(rinfo1,rinfo2,rinfo3,r0); 
   }
   add_1geo_amplitude(rinfo2,rinfo3,NULL,r0);    
   if (Verbose) printf("\n");
}

void add_geo_amplitudes(params *param,rinfos *first)
/*
 *   For every ray in the list rinfo, we shoot two extra rays
 *   at rinfo->sangle +/- dangle.  We then use these extra rays
 *   to compute the geometric amplitude of the rays in the rinfo list.
 */
{
double angle;
rinfos *rinfo,*rinfo1,*rinfo2;
double r0     = param->rrange;
double dangle = param->eigen.geo_dangle;
int    maxrays = 0;
int    raynum  = 0;

   if ( dangle == 0.0 ) return;
   for(rinfo = first; rinfo != NULL; rinfo = rinfo->next) maxrays += 2;

   fprintf(Log_file,"adding amplitude to %d rays.\n",maxrays/2);
   if (Verbose) printf("adding amplitude to %d rays.\n",maxrays/2);

   param->saving_pinfo = 0;      
   for(rinfo = first ; rinfo != NULL; rinfo = rinfo->next) {
      if (rinfo->rrange != r0) continue;
      angle = rinfo->sangle + dangle;
      raynum++;
      rinfo1 = shoot_one_ray(&Param,angle,raynum,maxrays);
      if (rinfo1->rrange != r0) continue;    
      angle = rinfo->sangle - dangle;
      raynum++;
      rinfo2 = shoot_one_ray(&Param,angle,raynum,maxrays);
      if (rinfo2->rrange != r0) continue;
      add_1geo_amplitude(rinfo1,rinfo,rinfo2,r0); 
   }
   if (Verbose) printf("\n");
}

/****************************************************************************/
/*                       End of eigenray.c                                  */
/****************************************************************************/
