/****************************************************************************/
/*                 program init.c                                           */
/****************************************************************************/
/*
 *    Copyright (C) 1992,  Woods Hole Oceanographic Institution
 *    See LICENSE file for use, copying and distribution conditions
 *
 *    Jim Bowlin jbowlin@whoi.edu              May 11, 1993
 *
 *    Provides two functions for reading and writing human
 *    readable ascii initializations files:
 *
 * void read_main_tab(iofiles *iof,intabs *intab);
 *
 *    Initializes variables listed in intab with the input specified
 *    in the (previously opened) iofile *iof.
 *
 * void print_main_tab(FILE *file,intabs *intab);
 *
 *    Prints out the current values in *intab to the FILE *file.
 *    This file can then be editted and read back in via read_main_tab().
 *
 */      

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "utils.h"
#include "reada.h"
#include "init.h"

#if !defined(__MSDOS__)
#include <unistd.h> 
#include <memory.h>
#endif

typedef void (*read_structs)(iofiles *iof, intabs *intab);
typedef void (*print_structs)(intabs *intab,int indent);
typedef dubs *(*save_structs)(intabs *intab,dubs *dub);


/***     these function declarations are used in some arrays of ***/
/***     function pointers defined below.                       ***/

void read_none(iofiles *iof, intabs *intab);
void read_tab(iofiles *iof, intabs *intab);
void read_enum(iofiles *iof, intabs *intab);
void read_double(iofiles *iof, intabs *intab);
void read_dub_array(iofiles *iof, intabs *intab);
void read_int(iofiles *iof, intabs *intab);
void read_str(iofiles *iof, intabs *intab);
void read_flag(iofiles *iof, intabs *intab);

void print_none(intabs *intab,int indent);
void print_tab(intabs *intab,int indent);
void print_double(intabs *intab,int indent);
void print_dub_array(intabs *intab,int indent);
void print_int(intabs *intab,int indent);
void print_enum(intabs *intab,int indent);
void print_str(intabs *intab,int indent);
void print_flag(intabs *intab,int indent);

dubs *save_none(intabs *intab,dubs *dub);
dubs *save_tab(intabs *intab,dubs *dub);
dubs *save_double(intabs *intab,dubs *dub);
dubs *save_dub_array(intabs *intab,dubs *dub);
dubs *save_int(intabs *intab,dubs *dub);
dubs *save_enum(intabs *intab,dubs *dub);
dubs *save_str(intabs *intab,dubs *dub);
dubs *save_flag(intabs *intab,dubs *dub);

static   FILE *Prt_file;
static   int   Comment_out;
printits Printit;
static   char  Prt_buf[256];

/****************************************************************************/
/*            arrays of pointers to functions                               */
/****************************************************************************/

/*
 *  These arrays are used for reading, printing, and writing (matlab binary)
 *  the intabs structures.
 */

/*
 *   Array of routines to read intabs
 */

read_structs read_struct[] = {             /* [intab.type] */
   read_none,
   read_tab,
   read_double,
   read_dub_array,
   read_int,
   read_enum,
   read_str,
   read_flag
};

/*
 *   Array of routines to print intabs to stdout.
 */

print_structs print_struct[] = {
   print_none,
   print_tab,
   print_double,
   print_dub_array,
   print_int,
   print_enum,
   print_str,
   print_flag
};

/*
 *   Array of routines to write intabs to matlab file.
 */

save_structs save_struct[] = {
   save_none,
   save_tab,
   save_double,
   save_none,   /* someday change to save_dub_array */
   save_int,
   save_enum,
   save_none,
   save_none,
};

/****************************************************************************/
/*          end of arrays of pointers to functions                          */
/****************************************************************************/

/****************************************************************************/
/*                      end of GLOBAL VARIABLES                             */
/****************************************************************************/

/****************************************************************************/
/*                 routines for reading intabs                              */
/****************************************************************************/
/*
 *  The routines in this section are used for reading the initialization
 *  file.
 */

void want_unique(iofiles *iof,intabs *intab)
/*
 *   Checks to see of this intab has already been set.
 *   If so we show where the problem is.
 */
{
   if ((intab->set_flag)++ == 0) return;
   iof_err(iof,"this parameter was already specified");
}

intabs *lookup_tab(iofiles *iof,intabs *intab,char *tok)
/*
 *  Tries to find *tok in the array (intab->list)[].name
 *  if it is found we return a pointer to the intab that contains *tok
 *  otherwise we report an error and return NULL.
 */
{
int i;
intabs *tab;
char *temp = iof->err_buf;
int   targ_len = 60;

   for(i=0; i<intab->num; i++) {
      tab = intab->list + i;
      if(strcmp(tok,tab->name) == 0) break;
   }
   if (i < intab->num) return(tab);
   strcpy(temp,"Expected ");
   for(i=0; i<intab->num; i++) {
      tab = intab->list + i;
		if (strlen(temp) + strlen(tab->name) +3 > targ_len)  {
			strcat(temp,"\n\t");
			targ_len += 70;
		}
      strcat(temp,"\"");
      strcat(temp,tab->name);
      strcat(temp,"\", ");
      if (i == intab->num-2)   strcat(temp,"or ");		
   }
   strcat(temp," here.");
   iof_err(iof,NULL);
   skip_past_eos(tok,iof);
   return(NULL);
}

double tab_value(iofiles *iof,intabs *tab)
/*
 *  Returns value pointed to by intab->where.  Used by lookup_units().
 */
{
   if (tab->type == Double) return( *(double*) tab->where);
   if (tab->type == Int )   return( *(int*) tab->where);
   sprintf(iof->err_buf,"Internal: incorrect type passed to \"tab_value\"");
   iof_err(iof,NULL);
   return(0.0);
}

void lookup_units(iofiles *iof,intabs *intab)
/*
 *   Reads the  " ... (unit) ..." part of an initialization file.
 *   and tries to find "unit" in intab->list (which should be pointing
 *   to one of the unit conversion intabs.   Then multiplies intab->where
 *   by the proper unit conversion factor.
 */
{
char   *tok;
intabs *tab;
double  mult;
dubs    *dub;

   if (intab->list == NULL) return;
   want_tok(iof,"(");
   tok = get_token(iof);
   tab = lookup_tab(iof,intab,tok);
   want_tok(iof,")");
   if (tab == NULL) return;
   mult = tab_value(iof,tab);
   if ( intab->type == Double)
     *(double*) intab->where *= mult;
   else if ( intab->type == Dub_array )
      for( dub = *(dubs**) intab->where; dub != NULL; dub = dub->next)
         dub->d *= mult;
   intab->def_unit = tab - intab->list;
}

void read_none(iofiles *iof,intabs *intab)
/*
 *    This is just filler and should never be called.
 */
{

   sprintf(iof->err_buf,"Internal: tried to read_none in %s",
      (intab == NULL) ?  "a NULL table" : intab->name);
   iof_err(iof,NULL);
}

void read_tab(iofiles *iof,intabs *intab)
/*
 *  Reads the part of an initialization file right after a new
 *  structure has been started as in:
 *         source {  range = .....
 *                ^
 *                | starts reading here.
 *
 *    Wants an open brace for multiple entries or a single entry from
 *    intab->list.
 */
{
char *tok;
intabs *tab;

   tok = get_token(iof);
   if (tok[0] != '{') {
      tab = lookup_tab(iof,intab,tok);
      if (tab != NULL) read_struct[tab->type](iof,tab);
      return;
   }
   for(tok = get_token_eos(iof); tok[0] != '}'; tok = get_token_eos(iof) ) {
      if (iof->end_of_file != 0) {
         eof_err(iof);
         printf("inside \"%s { ... }\"\n",intab->name);
         return;
      }
      tab = lookup_tab(iof,intab,tok);
      if (tab != NULL) read_struct[tab->type](iof,tab);
   }
}

void read_enum(iofiles *iof,intabs *intab)
/*
 *  Reads a choice parameter from the initialization file starting
 *  at the equal sign and ending at the semi-colon.
 */
{
intabs *tab;
char *tok;

   want_unique(iof,intab);
   want_tok(iof,"=");
   tok = get_token(iof);
   tab = lookup_tab(iof,intab,tok);
   if (tab == NULL) return;
   if(intab->where == NULL)
      iof_err(iof,"Internal error: NULL pointer in \"read_enum\"");
   
   else *(int*) (intab->where) = tab - intab->list;
/***    dont do this in order to get units reading correctly. -jbb 3/3/94
   if (tab->type != None) read_struct[tab->type](iof,tab);
   else
***/
   want_tok(iof,";");
}

void read_double(iofiles *iof, intabs *intab)
/*
 *  Reads in a real number from the initialization file starting
 *  at the equal sign  and ending at the semi-colon.
 *  Calls lookup_units() to read in the units and perform unit conversion.
 *  Sets *(intab->where) to the value read in.
 */
{
   want_unique(iof,intab);
   want_tok(iof,"=");
   *(double*) intab->where = get_a_dub(iof);

   /*** add error detection here ZZZ***/

   lookup_units(iof,intab);
   want_tok(iof,";");
}

int is_a_number(iofiles *iof,char *token)
{
char c = token[0];

   if ( ! ( isdigit(c) || c == '.' ||  c == '-') ) {
      iof_err(iof,"expected a number here.");
      return(0);
   }
   return(1);
}

dubs *link_double(dubs *first,double d)
/*
 *  Adds values to a linked list of doubles.
 *  Start list by sending a NULL first. There
 *  after first just gets cycled through.
 */
{
static dubs *last;
dubs *next;
   
   Safe_calloc(next,dubs,1);
   if (first == NULL) first = next;
   else last->next = next;
   last = next;
   next->d = d;
   return(first);
}

void read_dub_array(iofiles *iof, intabs *intab)
/*
 *  Reads in an array of doubles from the initialization file starting
 *  at the equal sign. wants something of the form:
 *
 *      ... = { d1, d2 , d3,  ... dn} (units);
 *
 *  Any combination of white space or commas are allowed between numbers.
 *  Creates a linked list of dubs containing all of the values read in.
 */
{
char c,*token;
dubs *dub = NULL;
double val1,val,val3,step,span;

   want_unique(iof,intab);
   want_tok(iof,"=");
   want_tok(iof,"{");

   for(;;) {
      token = get_token(iof);
      c = token[0];
      if ( c == '}') break;
      if ( c == ',') continue;

/***  want a number ***/

      if ( ! is_a_number(iof,token) ) continue;

/*** put the number on the list ***/

      dub = link_double(dub, val1 = atof(token));

/*** want ':'  '['  '}'  ',' or a number ***/

      token = get_token(iof);
      c = token[0];
      if ( c == '}') break;
      if ( c == ',') continue;
      if ( c == ':') {
         if ( ! is_a_number(iof,token = get_token(iof) )) continue;
         step = atof(token);
         if ( !want_tok(iof,":") ) continue;
         if ( ! is_a_number(iof,token = get_token(iof) )) continue;
         val3 = atof(token);
         if ( (val3 - val1) * step < 0.0 ) step = -step;
         for( val = val1 + step;(val3 - val) * (val - val1) >= 0.0; val+=step)
            dub = link_double(dub, val);
         continue;
      }
      if ( c == '#') {
         if ( ! is_a_number(iof,token = get_token(iof) )) continue;
         span = floor( fabs(atof(token)));
         if (span < 2.0 ) span = 2.0;
         if ( ! want_tok(iof,"#")) continue;
         if ( ! is_a_number(iof,token = get_token(iof) )) continue;
         val3 = atof(token);
         step = (val3 - val1) / ( span - 1.0);
         for(val = val1; span > 1; span--) 
            dub = link_double(dub, val+= step );
         continue;
      }

   }
   *(dubs**) (intab->where) = dub;
   lookup_units(iof,intab);
   want_tok(iof,";");
}

void read_int(iofiles *iof, intabs *intab)
/*
 *  Reads an integer from the initialization file.
 *  Sets *(intab->where) to the value read in.
 */
{
   want_unique(iof,intab);
   want_tok(iof,"=");
   *(int*) (intab->where) = get_an_int(iof);
   want_tok(iof,";");
}

void read_flag(iofiles *iof, intabs *intab)
/*
 *   Reads a flag parameter from the initialization file.
 *   Sets *(intab->where) to one.
 */
{
   want_unique(iof,intab);
   *(int*) (intab->where) = 1;
   want_tok(iof,";");
}

void read_str(iofiles *iof, intabs *intab)
/*
 *  Reads in a string parameter from the initialization file.
 *  Creates a copy of the string.
 *  Sets *(intab->where) to point to the copy.
 */
{
char *temp;

   want_unique(iof,intab);
   want_tok(iof,"=");
   temp = get_a_string(iof);
   if (temp != NULL) *(char**) (intab->where) = Strdup(temp);
   want_tok(iof,";");
}

void read_main_tab(iofiles *iof,intabs *intab)
/*
 *  Reads the entire initialization file via recursion
 *  using the read_struct[] array of routines.
 */
{
char *tok;
intabs *tab;

   for(;;) {
      tok = get_token_eos(iof);
      if (iof->end_of_file != 0) break;
      tab = lookup_tab(iof,intab,tok);
      if (tab != NULL) read_struct[tab->type](iof,tab);
   }
}

/****************************************************************************/
/*             end of routines for reading intabs                           */
/****************************************************************************/

/****************************************************************************/
/*             routines to print intabs                                     */
/****************************************************************************/

/*
 *  These routines are the "inverse" of the read_struct[] routines above.
 *  The read_struct[] routines read an initialization file and fill in
 *  locations specified by intab->where with values read in.
 *  The print_struct[] routines take the values from the locations
 *  intab->where and print them out to stdout.
 *  They are used for creating a sample initialization file which shows
 *  all of the default values I am using.
 */

void do_indent(int indent)
{
int i;
   for(i=0;i<indent;i++) Printit(" ",Prt_file);
}

void start_comment(intabs *intab)
{
   if (intab->set_flag == 0 && Comment_out)
		Printit("// ",Prt_file);
}

void indent_comment(int indent,int comflag)
/*
 * Indents indent spaces.  If comflag is set then
 * last three spaces are replaced with "// ".
 */
{
   if (comflag) {
		do_indent(indent-3);
	   Printit("// ",Prt_file);
   }
   else
		do_indent(indent);
}
						 
void end_comment(intabs *intab)
{
	Printit(";\n",Prt_file);	
	return;
   if (intab->set_flag == 0 && Comment_out)
		Printit("; */\n",Prt_file);
   else
		Printit(";\n",Prt_file);
}

void print_double(intabs *intab,int indent)
/*
 *  prints out the value *(intab->where) as a double.
 *  includes the name of the intab, the equal sign, the units
 *  and the closing semi-colon.
 */
{
double value;
intabs *utab;

   utab = intab->list;
   if (utab != NULL) utab += intab->def_unit;
   value = *(double*) intab->where;

   indent_comment(indent,intab->set_flag == 0 && Comment_out);

   if (utab != NULL) value /= *(double*) utab->where;
   sprintf(Prt_buf,"%-15s = %10.10g",intab->name,value);
   Printit(Prt_buf,Prt_file);
   if (utab != NULL) {
      sprintf(Prt_buf," (%s)",utab->name);
      Printit(Prt_buf,Prt_file);
   }
   end_comment(intab);
}

void print_dub_array(intabs *intab,int indent)
/*
 *  prints out the values stored in a linked list of
 *  dubs starting at *(intab->where) as  doubles.
 *  includes the name of the intab, the equal sign,
 *  the opening and closing braces, the units
 *  and the closing semi-colon.
 */
{
intabs *utab;
double value;
dubs *dub;

   utab = intab->list;
   if (utab != NULL) utab += intab->def_unit;
   dub = *(dubs**) (intab->where);
		
	indent_comment(indent,intab->set_flag == 0 && Comment_out);

   sprintf(Prt_buf,"%-15s = {",intab->name);
   Printit(Prt_buf,Prt_file);
   while( dub != NULL) {
      value = dub->d;
      if (utab != NULL) value /= *(double*) utab->where;
      sprintf(Prt_buf,"%10.10g",value);
      Printit(Prt_buf,Prt_file);
      dub = dub->next;
		if ( dub != NULL)  {
			Printit(",\n",Prt_file);
		   do_indent(indent+19);
		}
   }
   Printit(" }",Prt_file);
   if (utab != NULL) {
      sprintf(Prt_buf," (%s)",utab->name);
      Printit(Prt_buf,Prt_file);
   }
   end_comment(intab);
}

void print_int(intabs *intab,int indent)
/*
 *  prints out the value *(intab->where) as an int.
 *  includes the name of the intab, the equal sign, the units
 *  and the closing semi-colon.
 */

{
	indent_comment(indent,intab->set_flag == 0 && Comment_out);
	
   sprintf(Prt_buf,"%-15s = %10d",intab->name,*(int *)intab->where);
   Printit(Prt_buf,Prt_file);
   end_comment(intab);
}

void print_flag(intabs *intab,int indent)
/*
 *  Prints out the name of this flag.  Encloses it in comments
 *  if it was not set.
 */
{
	indent_comment(indent,*(int*) intab->where == 0);
	sprintf(Prt_buf,"%s;\n",intab->name);
   Printit(Prt_buf,Prt_file);
}

void print_str(intabs *intab,int indent)
/*
 *   Prints out the string in *(intab->where).
 */
{
char *s;

   s = *(char**) intab->where;
	indent_comment(indent, s == NULL);
   if ( s!= NULL)
		sprintf(Prt_buf,"%-15s = \"%s\";\n",intab->name,s);
   else
		sprintf(Prt_buf,"%-15s = \"\";\n",intab->name);
   Printit(Prt_buf,Prt_file);
}

void print_none(intabs *intab,int indent)
{
   do_indent(indent);
   sprintf(Prt_buf,"%s\n",intab->name);
   Printit(Prt_buf,Prt_file);
}

void print_tab(intabs *intab,int indent)
/*
 *  Starts printing a structure.  Prints structure by
 *  recursing through print_struct[](), then finishes
 *  structure with "};"
 */
{
int i;
intabs *tab;

   Printit("\n",Prt_file);
   do_indent(indent);
   sprintf(Prt_buf,"%s {\n",intab->name);
   Printit(Prt_buf,Prt_file);
   for(i=0;i<intab->num;i++) {
      tab = intab->list + i;
      print_struct[tab->type](tab,indent+3);
   }
   do_indent(indent);
   Printit("};\n",Prt_file);
}

void print_enum(intabs *intab,int indent)
/*
 *  Prints out choice parameter which was selected in *(intab->where).
 */
{
intabs *tab,*tab2;
int  i,offset = *(int*) intab->where;

   tab = intab->list + offset;
   do_indent(indent);
	
   sprintf(Prt_buf,"%-15s = %10s;",intab->name,tab->name);
   Printit(Prt_buf,Prt_file);

   Printit(" // ",Prt_file);
   for(i = 0; i < intab->num; i++) {
      if (i == offset) continue;
         tab2 = intab->list + i;
         sprintf(Prt_buf,"%s ",tab2->name);
    Printit(Prt_buf,Prt_file);
   }
   Printit("\n",Prt_file);
/*** prevent printout of unit while reading units jbb 3/3/94
   if (tab->type != None) print_struct[tab->type](tab,indent+3);
***/
}

long Default_printit(char *s,FILE *file);

long Default_printit(char *s,FILE *file)
{
   fputs(s,file);
   return(strlen(s));
}

void print_main_tab(printits printit,FILE *file,intabs *intab)
/*
 *  Prints out a sample initialization file reflecting the values
 *  that are currently in memory.
 */
{
int i;
intabs *tab;
FILE *old_file = file;

   Printit  = (printit == NULL) ? Default_printit : printit;
   Prt_file = (file == NULL)    ? stdout          : file;
   for(i=0;i<intab->num;i++) {
      tab = intab->list + i;
      print_struct[tab->type](tab,0);
   }
   Prt_file = old_file;
}

/****************************************************************************/
/*            end of routines to print intabs                               */
/****************************************************************************/

/****************************************************************************/
/*             routines to save intabs in a matlab file                     */
/****************************************************************************/
/*
 *   These routines are used to create a linked list of doubles (dubs)
 *   which can then be written to a matlab file.
 *   The only twist is that they are all passed the last dub in the
 *   linked list and if they are supposed to add a value to this list
 *   they create a new dub, link it to the end of the list and then
 *   return this dub, otherwise they pass the dub that they were given.
 */

dubs *save_double(intabs *intab,dubs *dub)
{

   Safe_calloc(dub->next,dubs,1);
   dub = dub->next;
   dub->d = *(double *) intab->where;
   return(dub);
}

dubs *save_int(intabs *intab,dubs *dub)
{
int    i;

   Safe_calloc(dub->next,dubs,1);
   dub = dub->next;
   i = *(int *)intab->where;
   dub->d = i;
   return(dub);
}

dubs *save_none(intabs *intab,dubs *dub)
{
   if (intab->type == Double) printf("save_none error.\n");
   return(dub);
}

dubs *save_tab(intabs *intab,dubs *dub)
{
int i;
intabs *tab;

   for(i=0;i<intab->num;i++) {
      tab = intab->list + i;
      dub = save_struct[tab->type](tab,dub);
   }
   return(dub);
}

dubs *save_enum(intabs *intab,dubs *dub)
{

   dub = save_int(intab,dub);
   dub = save_tab(intab->list,dub);
   return(dub);
}

dubs *__save_main_tab__(intabs *intab)
/*
 *  Saves all (except dub_array) initialization values in a matlab
 *  variable called "inits".  We do this by using the above routines
 *  to create a linked list of all of the initialization values.
 *  The order of the values in this list will always be identical to the
 *  order of the parameters in the sample initialization file created by
 *  the print_struct[] routines above.
 */
{
int i;
intabs *tab;
dubs   *first,*dub;

   if (intab == NULL) return(NULL);
   Safe_calloc(dub,dubs,1);
   first = dub;
   for(i=0;i<intab->num;i++) {
      tab = intab->list + i;
      dub = save_struct[tab->type](tab,dub);
   }
   return(first);
}

/****************************************************************************/
/*         end of routines to save intabs                                   */
/****************************************************************************/
