/****************************************************************************/
/*                 program initray.c                                        */
/****************************************************************************/
/*
 *    Copyright (C) 1992,  Woods Hole Oceanographic Institution
 *    See LICENSE file for use, copying and distribution conditions
 *
 *    Jim Bowlin Aug 1992.
 *
 *    Entry via initialize_ray().  Reads and processes command line
 *    parameters, reads and processes initialization file.  The format
 *    of the initialization file is determined by the arrays of intabs
 *    below.
 *
 * Changes: Rich Pawlowicz Aug 93
 *            - changed some things in resolve_eigenrays, set_defaults_params,
 *              and eigen_lst to reflect changes in params.eigen
 *              structure changes in ray.h
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <math.h>
#include "utils.h"
#include "reada.h"
#include "init.h"
#include "ray.h"
#include "readray.h"
#include "outray.h"
#include "helpray.h"
#include "initvar.h"

#if !defined(__MSDOS__)
#include <unistd.h>
#include <memory.h>
#endif


/*------------------- local flags -----------------------------------*/

int Parse;          /* print parsed initfile (boring) */
int Makeinit;       /* print sample initfile          */
int Makefull;       /* print full   initfile          */
int Help;           /* print extensive                */
int Whatsnew;       /* print whats new                */
int License;        /* print license agreement        */
int Comment_out;    /* comment out unset intabs when printing initfile */
int Quiet;          /* supress messages to stdout     */

/*----------------- end of local flags ------------------------------*/

iofiles   *Init_file;
long       Num_prt;


/*----------- Initialization file format defined by intabs below -----------*/


intabs source_lst[] = {
   "range",   KMETERS(Param.srange),
   "depth",    METERS(Param.sdepth),
};

intabs receiver_lst[] = {
   "range",      KMETERS(Param.rrange),
   "depth",       METERS(Param.rdepth),
	"depths", METER_ARRAY(Param.rdepths),
};

intabs units_lst[] = {
   "bath_range",ENUM(Meters,Param.units.bth_range.offset),
   "bath_depth",ENUM(Meters,Param.units.bth_depth.offset),
   "prof_range",ENUM(Meters,Param.units.ssp_range.offset),
   "prof_depth",ENUM(Meters,Param.units.ssp_depth.offset),
   "prof_speed",ENUM(Velocity,Param.units.ssp_speed.offset),
   "loss_range",ENUM(Meters,Param.units.loss_range.offset),
   "loss_angle",ENUM(Radians,Param.units.loss_angle.offset),
   "loss_atten",ENUM(Attenuation,Param.units.loss_atten.offset),
};

intabs input_lst[] = {
   "bath_file",   STR(Param.fname.bath),
   "prof_file",   STR(Param.fname.prof),
   "loss_file",   STR(Param.fname.loss),
   "units",       STRUCT(units_lst), 
};

intabs output_lst[] = {
   "mat_file",             STR(Param.fname.mat),
   "ascii_file",           STR(Param.fname.ascii),
   "initialization",      FLAG(Param.saveflg.initialization),
   "filenames",           FLAG(Param.saveflg.filenames),
   "sound_speeds",        FLAG(Param.saveflg.sound_speeds),
   "bathymetry",          FLAG(Param.saveflg.bathymetry),
   "fan",                 FLAG(Param.saveflg.fan),
   "wavefront",           FLAG(Param.saveflg.wavefront),
   "eigenrays",           FLAG(Param.saveflg.eigenrays),
   "everything",          FLAG(Param.saveflg.everything),
   "environment_only",    FLAG(Param.saveflg.e_only),
   "turning_points",      FLAG(Param.saveflg.turning_points)
};


intabs angle_lst [] = {
   "first",          DEGREES(Param.angle1),
   "last",           DEGREES(Param.angle2),
   "number",             INT(Param.num_angles),
   "specific",  DEGREE_ARRAY(Param.angles),
};

intabs pinfo_col_lst [] = {
   "range",       FLAG(Fan.saveflg[0]),
   "depth",       FLAG(Fan.saveflg[1]),
   "time",        FLAG(Fan.saveflg[2]),
   "angle",       FLAG(Fan.saveflg[3]),
   "speed",       FLAG(Fan.saveflg[4]),
   "grad",        FLAG(Fan.saveflg[5]),
   "length",      FLAG(Fan.saveflg[6]),
   "length_c2",   FLAG(Fan.saveflg[7]),
   "p_z",         FLAG(Fan.saveflg[8]),
   "top_bounces", FLAG(Fan.saveflg[9]),
   "bot_bounces", FLAG(Fan.saveflg[10]),
   "tot_bounces", FLAG(Fan.saveflg[11]),	
   "upper_tp",    FLAG(Fan.saveflg[12]),
   "lower_tp",    FLAG(Fan.saveflg[13]),
   "bottom_loss", FLAG(Fan.saveflg[14]),	
   "ray_number",  FLAG(Fan.saveflg[15]),
   "everything",  FLAG(Fan.save_everything),
};

intabs fan_lst [] = {
   "min_range",       KMETERS(Fan.min_range),
   "max_range",       KMETERS(Fan.max_range),
   "dr",              METERS(Fan.dr),
/***
   "dt",              SECONDS(Fan.dt),
   "dl",              METERS(Fan.dl),
***/   
   "include_bounces", FLAG(Fan.inc_bonks),
   "columns",         STRUCT(pinfo_col_lst),
};

intabs eigen_col_lst [] = {
   "range",       FLAG(Eigenfan.saveflg[0]),
   "depth",       FLAG(Eigenfan.saveflg[1]),
   "time",        FLAG(Eigenfan.saveflg[2]),
   "angle",       FLAG(Eigenfan.saveflg[3]),
   "speed",       FLAG(Eigenfan.saveflg[4]),
   "grad",        FLAG(Eigenfan.saveflg[5]),
   "length",      FLAG(Eigenfan.saveflg[6]),
   "length_c2",   FLAG(Eigenfan.saveflg[7]),
   "p_z",         FLAG(Eigenfan.saveflg[8]),
   "top_bounces", FLAG(Eigenfan.saveflg[9]),
   "bot_bounces", FLAG(Eigenfan.saveflg[10]),
	"tot_bounces", FLAG(Eigenfan.saveflg[11]),
   "upper_tp",    FLAG(Eigenfan.saveflg[12]),
   "lower_tp",    FLAG(Eigenfan.saveflg[13]),
	"bottom_loss", FLAG(Eigenfan.saveflg[14]),	
   "ray_number",  FLAG(Eigenfan.saveflg[15]),
   "everything",  FLAG(Eigenfan.save_everything),
};

intabs eigen_lst [] = {
   "min_range",       KMETERS(Eigenfan.min_range),
   "max_range",       KMETERS(Eigenfan.max_range),
   "geo_miss",        METERS(Param.eigen.geo_miss),
   "search_max",      INT(Param.eigen.search_max),
   "amp_dangle",      DEGREES(Param.eigen.geo_dangle), 
   "dr",              METERS(Eigenfan.dr),
/***
   "diff_miss",       METERS(Param.eigen.diff_miss),
   "dt",              SECONDS(Eigenfan.dt),
   "dl",              METERS(Eigenfan.dl),
***/   
   "include_bounces", FLAG(Eigenfan.inc_bonks),
   "columns",         STRUCT(eigen_col_lst),
};

intabs ran_dep_lst[] = {
   "full",    NONE,
   "none",    NONE,
   "grad_z",  NONE,
};

intabs bottom_type_lst[] = {
   "reflecting",    NONE,
   "absorbing",     NONE,
};

intabs step_size_lst[] = {
   "multiplier",     UDOUBLE(Param.step.mult),
   "max",             METERS(Param.step.max),
   "min",             METERS(Param.step.min),
   "cos_factor",     UDOUBLE(Param.step.cos_factor),
};

intabs bias_lst[] = {
   "tolerance",     VELOCITY(Param.debias.tolerance),
   "iteration",          INT(Param.debias.iteration),
   "exact",             FLAG(Param.debias.exact)
};

intabs smoothing_lst[] = {
   "levitus",          FLAG(Param.smooth.levitus),
   "auto_max",         FLAG(Param.smooth.auto_max),
   "specific",  METER_ARRAY(Param.smooth.specific)
};

intabs extrema_lst[] = {
   "min_angle",      DEGREES(Param.add_extrema.da),
   "iterations",     INT(Param.add_extrema.max_iter),
   "min_dz",         METERS(Param.add_extrema.min_dz),
   "min_cnt",        INT(Param.add_extrema.min_cnt),
   "z_miss",         METERS(Param.add_extrema.z_miss),		
			
};

intabs default_lst[] = {
   "range_depend",       ENUM(ran_dep_lst, Param.range_depend),
   "bath_smoothing",     KMETERS(Param.bath_smooth),
   "bottom_depth",       METERS(Param.bottom),
   "earth_radius",       KMETERS(Param.earth_radius),
   "z_tolerance",        METERS(Param.z_tol),
   "max_bounces",        INT(Param.max_bonks),
   "max_angle",          DEGREES(Param.max_angle),
    "bottom_type",       ENUM(bottom_type_lst,Param.bottom_type),
    "margins",           INT(Param.margin)
};

intabs mat_name_lst[] = {
   "fan",         STR(Param.vname.fan),   
   "ctab",        STR(Param.vname.ctab),
   "btab",        STR(Param.vname.btab),
   "rtab",        STR(Param.vname.rtab),
   "idex",        STR(Param.vname.idex),
   "init",        STR(Param.vname.init),
   "wavefront",   STR(Param.vname.wavefront),
   "eigenrays",   STR(Param.vname.eigenrays),
   "eigenfront",  STR(Param.vname.eigenfront),
};


intabs main_lst[] = {
   "input",           STRUCT(input_lst),
   "output",          STRUCT(output_lst),
   "source",          STRUCT(source_lst),
   "receiver",        STRUCT(receiver_lst),
   "angles",          STRUCT(angle_lst),
   "fan",             STRUCT(fan_lst),	
   "add_extrema",     STRUCT(extrema_lst),
   "eigenrays",       STRUCT(eigen_lst),
   "step_size",       STRUCT(step_size_lst),
   "debias",          STRUCT(bias_lst),
   "prof_smoothing",  STRUCT(smoothing_lst),
   "model",           STRUCT(default_lst),
   "mat_names",       STRUCT(mat_name_lst),
};

intabs Main_tab[] = {
   "main", STRUCT(main_lst)
};

/*------------------ small initialization file ----------------------------*/

intabs source_sml[] = {
   "range",   KMETERS(Param.srange),
   "depth",    METERS(Param.sdepth),
};

intabs receiver_sml[] = {
   "range",      KMETERS(Param.rrange),
   "depth",       METERS(Param.rdepth),
	"depths", METER_ARRAY(Param.rdepths),
};


intabs input_sml[] = {
   "bath_file",   STR(Param.fname.bath),
   "prof_file",   STR(Param.fname.prof),
   "loss_file",   STR(Param.fname.loss),
};

intabs output_sml[] = {
   "mat_file",             STR(Param.fname.mat),
   "ascii_file",           STR(Param.fname.ascii),
   "turning_points",      FLAG(Param.saveflg.turning_points)
};

intabs angle_sml [] = {
   "specific",  DEGREE_ARRAY(Param.angles),
};


intabs fan_sml [] = {
   "dr",              METERS(Fan.dr),
   "include_bounces", FLAG(Fan.inc_bonks),
};

intabs eigen_sml [] = {
   "geo_miss",        METERS(Param.eigen.geo_miss),
	"search_max",      INT(Param.eigen.search_max),
   "dr",              METERS(Eigenfan.dr),
   "include_bounces", FLAG(Eigenfan.inc_bonks),
};

intabs bottom_type_sml[] = {
   "reflecting",    NONE,
   "absorbing",     NONE,
};

intabs step_size_sml[] = {
   "multiplier",     UDOUBLE(Param.step.mult),
   "max",             METERS(Param.step.max),
   "min",             METERS(Param.step.min),
   "cos_factor",     UDOUBLE(Param.step.cos_factor),
};

intabs smoothing_sml[] = {
   "levitus",          FLAG(Param.smooth.levitus),
   "auto_max",         FLAG(Param.smooth.auto_max),
   "specific",  METER_ARRAY(Param.smooth.specific)
};

intabs extrema_sml[] = {
   "min_angle",      DEGREES(Param.add_extrema.da),
   "iterations",     INT(Param.add_extrema.max_iter),
   "min_dz",         METERS(Param.add_extrema.min_dz),
   "z_miss",         METERS(Param.add_extrema.z_miss),			
};

intabs default_sml[] = {
   "bath_smoothing",     KMETERS(Param.bath_smooth),
   "bottom_depth",       METERS(Param.bottom),
   "max_bounces",        INT(Param.max_bonks),
    "bottom_type",       ENUM(bottom_type_sml,Param.bottom_type),
};

intabs main_sml[] = {
   "input",           STRUCT(input_sml),
   "output",          STRUCT(output_sml),
   "source",          STRUCT(source_sml),
   "receiver",        STRUCT(receiver_sml),
   "angles",          STRUCT(angle_sml),
   "fan",             STRUCT(fan_sml),	
   "add_extrema",     STRUCT(extrema_sml),
   "eigenrays",       STRUCT(eigen_sml),
   "step_size",       STRUCT(step_size_sml),
   "prof_smoothing",  STRUCT(smoothing_sml),
   "model",           STRUCT(default_sml),
};

intabs Small_tab[] = {
   "main", STRUCT(main_sml)
};

/*----------------- main initialization routines ---------------------------*/

void disp_banner(void)

{
   printf("\n/************************* %s *************************/\n",Prog_name);
   printf("/*    Version %4.2f %s\n",Version, Prog_date);
   printf(" *    Copyright (C) 1992-1994,  Woods Hole Oceanographic Institution.\n");
   printf(" *    Contact Jim Bowlin (jbowlin@whoi.edu) for information and updates.\n");
   printf(" *    Use -l flag for use, copying and distribution conditions.\n");
   printf(" */\n");
}

void usage(char *s)
/*
 * Prints usage message. If s is not NULL we then print *s and exit
 * from the program.  Otherwise we just return.
 */
{
int i;
static char *Usage_text[] = {
"usage:    ray initfile.ray [cmnd1] [cmnd2] ... [cmndN] [flags]" ,
"where:",
"   initfile.ray       is a ray initialization file" ,
"   cmnd1 ... cmndN    are command line parameters used in the" ,
"                      initialization file via $1$ ... $N$" ,
"flags:" ,
"   -d    (debug)      send some debugging info to stdout" ,
"   -e    (e_only)     do NOT trace any rays" ,
"   -h    (help)       send  detailed help to stdout",
"   -l    (license)    display the license agreement",
"   -m    (makeinit)   generate easy initialization file",
"   -M    (Makeinit)   generate FULL initialization file",
"   -n    (whatsnew)   display a list of whats new to this version",
"   -o    (output)     forces output file to be next argument",
"   -q    (quiet)      print as little as possible",
"   -s    (screenmode) formats verbose output for screen"};

   for(i=0; i<Num_els(Usage_text); i++) mputs(Usage_text[i]);
   if (s != NULL) {
      mputs(s);
      exit(-1);
   }
}

/*--------------- routines to open files -----------------------------------*/

void require_file(char *name, char *file_type)
{
   if (name != NULL) return;
   printf("Need to specify %s file\n",file_type);
   fatal(NULL);
}

iofiles *safe_open_iofile(char **name,int len,char *format,char *s,char *ext)
{
iofiles *iof;

   if (name == NULL) return(NULL);
   if (*name == NULL) return(NULL);
   if (ext != NULL) default_extension(name,ext);
   iof = open_iofile(*name,len,format);
   if (iof == NULL) {
      printf("Could not open ==> %s <== for %s\n",*name,s);
		if (Log_file != NULL)
			fprintf(Log_file,"Could not open ==> %s <== for %s\n",*name,s);
      fatal(NULL);
   }
   return(iof);
}

FILE *safe_open_file(char **name,char *format,char *s,char *ext)
{
FILE *file;

   if (name == NULL) return(NULL);
   if (*name == NULL) return(NULL);
   default_extension(name,ext);
   file = fopen(*name,format);
   if (file == NULL) {
      printf("Could not open ==> %s <== for %s\n",*name,s);
      exit(-1);
   }
   return(file);
}

/*--------------- end of routines to open files ----------------------------*/

long Print_mat(char *s,FILE *f);

long Print_mat(char *s,FILE *file)
{
double d;
int    i,len;

   len = strlen(s);
   for(i = 0; i < len; i++) {
      d = (double) s[i];
      fwrite(&d,sizeof(double),1,file);
   }
   Num_prt += len;
   return(len);
}
      
void save_main_tab(FILE *file,char *name,intabs *tab)
{
long     where;
   
   if (file == NULL || tab == NULL || name == NULL) return;
   Num_prt = 0;
   where = ftell(file);
   start_mat(file,name,1,Num_prt,0,mat_type(0,0,1));
   print_main_tab(Print_mat,file,Main_tab);
   fseek(file,where,SEEK_SET);
   start_mat(file,name,1,Num_prt,0,mat_type(0,0,1));
   fseek(file,0,SEEK_END);
}

void write_asc_header(FILE *file,char *name,time_t *timer)
/*
 * writes a little ascii header to the ascii file
 */
{

   if (file == NULL) return;
   fprintf(file,"file <%s> created %s",name,ctime(timer));
   fprintf(file,"%s program version %5.2f %s\n\n",
      Prog_name,Version,Prog_date);
}

void do_saves1(params *param)
/*
 *  Saves some or all the information we have so far into the matlab file
 *  Mat_file depending on the flags in the Save structure.
 *  Also start the ascii file if there is one.
 */
{
time_t timer;

   time(&timer);
   write_asc_header(Log_file,Param.fname.log,&timer);
   save_string(NULL, Log_file,"init_file",Param.fname.init);
   save_string(NULL, Log_file,"mat_file",Param.fname.mat);
   save_string(NULL, Log_file,"ascii_file",Param.fname.ascii);
   save_string(NULL, Log_file,"prof_file",Param.fname.prof);
   save_string(NULL, Log_file,"bath_file",Param.fname.bath);
	save_string(NULL, Log_file,"loss_file",Param.fname.loss);
   if (Asc_file != NULL) write_asc_header(Asc_file,Param.fname.ascii,&timer);

   if (param->saveflg.initialization) {
      ldbug("Saving initialization.");
      if (Asc_file != NULL) print_main_tab(NULL,Asc_file,Main_tab);
      save_main_tab(Mat_file,Param.vname.init,Main_tab);
      save_string(Mat_file,NULL,"prog_date",Prog_date);
   }

   save_string(Mat_file,NULL,"run_date", ctime(&timer));
   save_doubles(Mat_file,&Version,1,"ray_version");

   if (param->saveflg.filenames) {
      ldbug("Saving filenames.");
      save_string(Mat_file, Asc_file,"init_file",Param.fname.init);
      save_string(Mat_file, Asc_file,"mat_file",Param.fname.mat);
      save_string(Mat_file, Asc_file,"ascii_file",Param.fname.ascii);
      save_string(Mat_file, Asc_file,"prof_file",Param.fname.prof);
      save_string(Mat_file, Asc_file,"bath_file",Param.fname.bath);
		save_string(Mat_file, Asc_file,"loss_file",Param.fname.loss);
   }
}

/*---------- routines for reading command line parameters ------------------*/

/*
 *   The three set_xxxx() routines below compare a command line argument *s
 *   with a fixed string *sflag.  If the two string are different we return(0)
 *   otherwise we return(1).  Before returning:
 *        set_flag   sets *where to one.
 *        set_double sets *where to the value after the "=" sign in *s
 *        set_long   sets *where to the value after the "=" sign in *s
 */

int set_flag(char *s, char *sflag,int *where)
{
   if ( strcmp(s,sflag) ) return(0);
   *where = 1;
   return(1);
}

int set_double(char *s, char *signal,double *where)
{
int len;

   len = strlen(signal);
   if (strncmp(s,signal,len) !=0) return(0);
   sscanf(s+len,"%lf",where);
   return(1);
}

int set_long(char *s, char *signal,long *where)
{
int len;

   len = strlen(signal);
   if (strncmp(s,signal,len) !=0) return(0);
   sscanf(s+len,"%li",where);
   return(1);
}

int set_flag_a(char s, char flag,int *where)
{
   if ( s != flag ) return(0);
   *where = 1;
   return(1);
}

int set_all_flags(char *s)
{
char c;
int i;
int len = strlen(s);

   if ( len < 2 ) return(0);
   for(i=1;i<len;i++) {
      c = s[i];
      if (set_flag_a(c,'d',   &Debug))        continue;
      if (set_flag_a(c,'e',   &Param.saveflg.e_only))  continue;
      if (set_flag_a(c,'h',    &Help))        continue;
      if (set_flag_a(c,'l', &License))        continue;
      if (set_flag_a(c,'m',&Makeinit))        continue;
      if (set_flag_a(c,'M',&Makefull))        continue;		
      if (set_flag_a(c,'n',&Whatsnew))        continue;
		if (set_flag_a(c,'q',   &Quiet))        continue;
      if (set_flag_a(c,'s',   &Screen))       continue;
      if (set_flag_a(c,'v', &Verbose))        continue;
      return(0);
   }
   return(1);
}

void read_commands(int argc, char *argv[])
/*
 *  Processes all of the command line arguments contained in *argv[],
 *  except the first which is just the name of this program.
 *  All arguments beginning with a "/" are treated as flags.
 *  Unknown flags cause the usage text to be displayed and then
 *  an exit(-1).  The first non-flag argument is sent to Param.fname.init
 *  since it should be the name of the intialization file. All other
 *  non-flag arguments are passed to the reada.c module via add_command()
 *  since they should be used for command line substitution when we use
 *  reada to read the initialization file.
 */
{
char s[100];
int  arg;
int  cmnd_cnt = 0;

   for(arg=1; arg < argc; arg++)    /* process all command line arguments */
   {
      strcpy(s,argv[arg]);
      if ((s[0] != '-') || (!isalpha(s[1]) ))  {
    if (cmnd_cnt++ == 0) {
       Param.fname.init = argv[arg];
       add_command(raw_nickname(Param.fname.init));
    }
    else add_command(argv[arg]);
    continue;
      }
      if  (strcmp(s,"-o") == 0) {
         if (++arg >= argc) continue;
    Param.fname.cmnd = strdup(argv[arg]);
    continue;
      }
      if (set_all_flags(s)) continue;
      printf("Unknown argument: ==> %s <==",s);
      exit(-1);
   }
}

/*------- end of routines for reading command line parameters --------------*/

void makeinit(FILE *log,int comment)
{
time_t timer;
intabs *tab;

   if ( Makeinit ) 
		tab = Small_tab;
	else if( Makefull )
		tab = Main_tab;
	else
		return;
	
   time(&timer);	
	Comment_out = comment;
	printf("\n// Ray initialization file created %s\n",ctime(&timer));
	print_main_tab(NULL,stdout,tab);
	if (log == NULL) return;
	fprintf(log,"// Ray initialization file created %s\n\n",ctime(&timer));
	print_main_tab(NULL,log,tab);
	exit(0);
}
	
void read_init_file(char **name,FILE *log)
/*
 *  Takes care of reading the intialization file and writing
 *  a sample initialization if requested to do so by Makeinit.
 */
{
   if (*name == NULL)
		makeinit(log,0);
	
   default_extension(name,".ray");
   require_file(*name,"initialization");
   if (Verbose)
      printf("===> reading <%s> as initialization file.\n",*name);
   Init_file =  safe_open_iofile(name,256,"rt","initialization",NULL);
   Init_file->debug         = Parse;
   Init_file->log_file      = log;
   Init_file->line_echo_off = 0;
   Init_file->prog_name     = Prog_name;

   read_main_tab(Init_file,Main_tab);
   makeinit(log,1);

}

/*----------- resolve ambiguites in initialization here --------------------*/

/*
 *  This section of code makes sure that the initialization file
 *  contained enough information to do what we think was requested.
 *  We also make sure that there are no conflicting requests.
 */

int resolve_filenames(params *param)
{
/*
 *  [Just read the printf's]
 */

int errors = 0;

   if (param->fname.mat == NULL && param->fname.ascii == NULL) {
      printf("Ambiguity: output file must be specified, as in\n");
      printf("      output mat_file = \"fname.mat\";\n");
      printf(" or   output ascii_file = \"fname.txt\";\n\n");
      errors++;
   }

   if ( param->fname.prof == NULL) {
      printf("Ambiguity: an input prof_file must be specified, as in\n");
      printf("      input prof_file = \"fname.ssp\";\n\n");
      errors++;
   }

   return(errors);
}

void resolve_saves(params *param,pinfos *fan,pinfos *eigenfan)
{
int count = 0;

   count += param->saveflg.initialization;
   count += param->saveflg.filenames;
   count += param->saveflg.sound_speeds;
   count += param->saveflg.bathymetry;
   count += param->saveflg.fan;
   count += param->saveflg.eigenrays;
   count += param->saveflg.wavefront;
   if (count == 0) param->saveflg.everything = 1;

   if (param->saveflg.everything) {
      param->saveflg.initialization = 1;
      param->saveflg.filenames      = 1;
      param->saveflg.sound_speeds   = 1;
      param->saveflg.bathymetry     = 1;
      param->saveflg.wavefront      = 1;
      param->saveflg.turning_points = 1;
      if (fan->selected) param->saveflg.fan = 1;
      if(eigenfan->selected ) param->saveflg.eigenrays = 1;
   }
   if(param->saveflg.fan == 0 &&
      param->saveflg.wavefront == 0 &&
      param->saveflg.eigenrays == 0) param->saveflg.e_only = 1;
}

int resolve_pinfo(params *param,pinfos *pinfo,int flag,char *name)
{
int errors = 0;

   pinfo->selected = (pinfo->dr > 0.0) + (pinfo->dt > 0.0) + (pinfo->dl > 0.0);
   if (pinfo->selected > 1) {
      printf("Ambiguity: only \"%s dr\" or \"dt\" or \"dl\" may be used.\n\n",name);
      errors++;
   }
   
   if ( pinfo->min_range == 0.0  &&  pinfo->max_range == 0.0 ) {
      pinfo->min_range = param->srange;
      pinfo->max_range = param->rrange;
   }
   if (flag && ! pinfo->selected) {
      printf("Ambiguity: \"output %s;\" selected but no \"%s dr = ;\"\n",
				 name,name);
      printf("or \"%s dt = ;\" or \"%s dt = ;\" specified.\n",name,name);
      errors++;
   }
   return(errors);
}

void resolve_pinfosaves(int flag,pinfos *pinfo)
{
int i,count = 0;

   for(i=0; i < Fan_saves; i++)
      count += pinfo->saveflg[i];

   if (count == 0)  pinfo->save_everything = 1;

   if (pinfo->save_everything) {
      /* don't set turning points automatically */
      for (count=0; count < Fan_saves - 4; count++)
         pinfo->saveflg[count] = 1;
		pinfo->saveflg[Fan_saves-1] = 1;
		pinfo->saveflg[Fan_saves-2] = 1;
   }
   count = 0;
   for(i=0; i < Fan_saves; i++)
      count += pinfo->saveflg[i];

   pinfo->num_cols = count;
   if (flag == 0) pinfo->inc_bonks = 0;
}

int resolve_angles(params *param)
{
int errors = 0;

   if (param->angles != NULL) {
      if (param->num_angles != 0) {
    printf("Ambiguity: both \"angles specific\" and \"angles number\" were initialized.\n\n");
    errors++;
      }
      if (param->angle1 != 0.0) {
    printf("Ambiguity: both \"angles specific\" and \"angles first\" were initialized.\n\n");
    errors++;
      }
      if (param->angle2 != 0.0) {
    printf("Ambiguity: both \"angles specific\" and \"angles last\" were initialized.\n\n");
    errors++;
      }
      return(errors);
   }
   if (param->num_angles == 0) {
      printf("Ambiguity: an angular range must be specified, as in\n");
      printf("      angles first =  15 (degrees);\n");
      printf("      angles last  = -15 (degrees);\n");
      printf("      angles number = 100;\n");
      printf("or\n");
      printf("      angles specific = { 15,15,5,0,-5,-10,-15} (degrees);\n\n");
      errors++;
      return(errors);
   }
   if (( param->angle1 == param->angle2 ) && ( param->num_angles != 1) ) {
      printf("Ambiguity: an angular range must be specified, as in\n");
      printf("      angles first =  15 (degrees);\n");
      printf("      angles last  = -15 (degrees);\n\n");
      errors++;
   }
   return(errors);
}

int resolve_eigenrays(params *param)
{
int errors = 0;

   if (param->eigen.diff_miss < 0.0 ) param->eigen.diff_miss = 0.0; 
   if (param->eigen.geo_miss < 0.0 ) param->eigen.geo_miss = 5; 
   if (param->eigen.search_max < 0 ) param->eigen.search_max = 5; 
   if (param->eigen.diff_miss != 0.0){
       param->eigen.diff_miss = 0.0;
       printf("Ambiguity: diffracted rays haven't been implemented yet!\n");
       errors++;
      }
   return(errors);
}

int resolve_bounds(params* param)
{
int errors = 0;

   if (param->max_angle >= Pi)
		param->min_cos = 0.0;
	else
		param->min_cos = cos(param->max_angle);
	
   if (param->srange == param->rrange) {
      printf("Ambiguity: receiver range cannot be the same as source range.\n");
      printf("Specify a new range in the form\n");
      printf("      receiver range = 1000 (km);\n\n");
      errors++;
   }
   if (param->sdepth <= 0.0) {
      printf("Ambiguity: the source cannot be on a boundary. Try moving it as in\n");
      printf("      source depth = 0.1 (m);\n\n");
      errors++;
   }
   return(errors);
}

int resolve_smoothing(params *param)
{
int num_set;

   num_set = param->smooth.levitus;
   num_set += param->smooth.auto_max;
   if (param->smooth.specific != NULL) num_set++;
   if (num_set == 1) return(0);
   if (num_set == 0) {
      param->smooth.auto_max = 1;
      ldbug("Selecting automatic width selection.");
      return(0);
   }
   printf("Ambiguity: only one kind of prof_smoothing can be used.\n");
   printf("Please select from:\n");
   printf("      \"model prof_smoothing levitus;            /* default levitus widths */\"\n");
   printf("      \"model prof_smoothing auto_max;           /* maximum widths         */\"\n");
   printf("      \"model prof_smoothing specific = { } (m); /* user specified list    */\"\n");
   return(1);
}

void resolve_ambiguities(params *param, pinfos *fan,pinfos *eigenfan)
{
int errors = 0;

   errors += resolve_pinfo(param,fan,Param.saveflg.fan,"fan");
   errors += resolve_pinfo(param,eigenfan,Param.saveflg.eigenrays,"eigenrays");
   resolve_saves(param,fan,eigenfan);
   resolve_pinfosaves(Param.saveflg.fan,fan);
   resolve_pinfosaves(Param.saveflg.eigenrays,eigenfan);
   errors += resolve_filenames(param);
   errors += resolve_smoothing(param);

   if (!param->saveflg.e_only) {
      errors += resolve_angles(param);
      errors += resolve_bounds(param);
      errors += resolve_eigenrays(param);
   }
   if (errors == 0) return;
   printf("Error: %d ambiguit%s in initialization file <%s>.\n",errors, (errors == 1) ? "y" : "ies", param->fname.init);
   fatal(NULL);
}

/*----------- end of resolve ambiguites in initialization ------------------*/

void create_log_name(params *param)
/*
 *  uses the name of the init file to create a name for the log file
 */
{
    if (param->fname.log != NULL) return;
    if (param->fname.init == NULL) param->fname.log = "ray.log";
    else param->fname.log = raw_nickname(param->fname.init);
 }

void set_default_params(params *param)
{
    param->add_extrema.min_cnt = 1;
    param->step.mult        = 1.0;
    param->step.max         = 200;
    param->step.min         = 5;
    param->step.cos_factor  = 10.0;
    param->debias.tolerance = 0.0001;
    param->debias.iteration = 1000;
    param->eigen.geo_miss   = 5;
    param->eigen.search_max = 10;
    param->eigen.diff_miss  = 0.0;
    param->eigen.geo_dangle = 0.001 * Pi / 180.0; 
    param->bath_smooth      = 1.0e4;
    param->earth_radius     = 6378137.0;
    param->max_angle        = 89.5 * Pi / 180.0;
    param->margin           = 100;
    param->z_tol            = 1.0e-10;
    param->sign_dr          = 1.0;
    param->sign_z           = 1.0;
    param->vname.ctab       = "ctab";
    param->vname.btab       = "btab";
    param->vname.rtab       = "rtab"; 
    param->vname.idex       = "idex"; 
    param->vname.wavefront  = "wavefront"; 
    param->vname.fan        = "fan"; 
    param->vname.eigenrays  = "eigenrays"; 
    param->vname.eigenfront = "eigenfront";
    param->vname.init       = "init"; 
    param->bottom_type      = 0;
    param->units.bth_range.offset  = 1;
    param->units.bth_depth.offset  = 0;
    param->units.ssp_range.offset  = 1;
    param->units.ssp_depth.offset  = 0;
    param->units.ssp_speed.offset  = 0;

    param->units.loss_range.offset  = 1;
    param->units.loss_angle.offset  = 1;
    param->units.loss_atten.offset  = 0;

    param->units.loss_range.table  = Meters;
    param->units.loss_angle.table  = Radians;
    param->units.loss_atten.table  = Attenuation;

    param->units.bth_range.table  = Meters;
    param->units.bth_depth.table  = Meters;
    param->units.ssp_range.table  = Meters;
    param->units.ssp_depth.table  = Meters;
    param->units.ssp_speed.table  = Velocity;
}

void initialize_ray(int argc, char *argv[])
/*
 *  Read the command line arguments,
 *  set appropriate flags,
 *  read initialization file,
 *  open output file and auxillary files,
 *  save initialization information if requested.
 */
{

   disp_banner();
   if (argc < 2 ) {
      usage(NULL);
      exit(0);
   }
   set_default_params(&Param);
   ldbug("Reading command line arguments.");
   read_commands(argc,argv);

   if (Help)    do_help();
   if (License) do_license();
   if(Whatsnew) do_whatsnew();

	if( !Verbose && Quiet) Verbose = 0;
	else Verbose = 1;
	
   create_log_name(&Param);
   Log_file  = safe_open_file(&Param.fname.log,"wt","log",".log");

   read_init_file(&Param.fname.init,Log_file);
   close_iofile(Init_file);
   if ( check_commands() != 0) exit(-1);
	
   if (Param.fname.cmnd != NULL) Param.fname.mat = Param.fname.cmnd;
   
   ldbug("Resolving ambiguities.");
   resolve_ambiguities(&Param,&Fan,&Eigenfan);

   ldbug("Opening files.");
   Mat_file  = safe_open_file(&Param.fname.mat,"wb","matlab",".mat");
   Asc_file  = safe_open_file(&Param.fname.ascii,"wt","ascii",".asc");
   Prof_file = safe_open_file(&Param.fname.prof,"rb","sound speed profiles",".ssp");
   Bath_file = safe_open_file(&Param.fname.bath,"rb","bathymetry",".bth");
   Loss_file = safe_open_file(&Param.fname.loss,"rb","bottom loss",".bls");
   do_saves1(&Param);
}

/****************************************************************************/
/*                      end of initray.c                                    */
/****************************************************************************/

