/****************************************************************************/
/*                 program preray.c                                            */
/****************************************************************************/
/*
 *      Copyright (C) 1992,  Woods Hole Oceanographic Institution
 *    See LICENSE file for use, copying and distribution conditions
 *
 *
 *  jim bowlin feb 1992.
 *
 *  used by ray.c to pre-process profiles and generate tables
 *  for integration.
 */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "utils.h"
#include "ray.h"

/****************************************************************************/
/*                 GLOBAL VARIABLES                                         */
/****************************************************************************/

int Def_widths[] =
  { 5, 5,   5,   5,   5,  20,  12,  13,  12,  25,  25,  25,  50,
   50,  50,  50,  50,  50,  50,  50,  50,  50,  50,  50,  50,
  125, 125, 250, 250, 250, 250, 250, 250, 250, 250 ,250, 250};

/*  These widths were tuned for the Levitus depth spacing of sound speeds  */
/*  Users can now input their own values via "prof_smoothing".             */


/****************************************************************************/
/*                      Utilites                                            */
/****************************************************************************/

double tab_speed(c_tabs *tab,double z)
/*
 *   Computes sound speed from table look-up.  Used in tabtab_speed() below
 *   and in computing parabolic sound speed parameters.
 */
{
double dz = z - tab->z0;

   return ( tab->c0 + dz * ( tab->c1 + dz * tab->c2) );
}

double tab_grad(c_tabs *tab,double z)
/*
 *  computes dc/dz from table look-up.
 */
{
double dz = z - tab->z0;

   return ( tab->c1 + 2.0 * dz * tab->c2 );
}

double tabtab_speed(tabtabs *tt,int *idex,double range,double depth)
/*
 *  Computes linearly interpolated
 *  (in range) sound speed at range range and depth depth.
 */
{
double c1,c2,lambda;
int iz,iiz;

   iz     = (int) depth + idex[0];
   iiz    = idex[iz];
   lambda = (tt->r1 - range) * tt->r12_inv;
   c1 = tab_speed(tt->tab1 + iiz, depth);
   c2 = tab_speed(tt->tab2 + iiz, depth);
   return( c1 + lambda * (c2 - c1) );
}

double get_speed(tabtabs *tabtab,int *idex,double range,double depth)
/*
 *  Used for getting speed of sound at source.  Finds proper tabtab
 *  in linked list starting at *tabtab then calls tabtab_speed() to
 *  get the linearly interpolated (in range) sound speed.
 */
{
tabtabs *tt;

   for(tt = tabtab; tt != NULL; tt = tt->next)
      if ( tt->r1 <= range && tt->r2 >= range) break;
   if (tt == NULL) fatal("range not bracketed in \"get_speed\".");
   return(tabtab_speed(tt,idex,range,depth));
}

void interpolate_speed(c_tabs *rtab,c_tabs *tab1,c_tabs *tab2,double lambda,int numsect)
/*
 *   Creates a new sound speed profile linearly interpolated between
 *   tab1 and tab2.
 */
{
int i;

   for(i=0;i < numsect; i++) {
      rtab[i].z0 = tab1[i].z0;
      rtab[i].c0 = tab1[i].c0 + lambda * (tab2[i].c0 - tab1[i].c0 );
      rtab[i].c1 = tab1[i].c1 + lambda * (tab2[i].c1 - tab1[i].c1 );
      rtab[i].c2 = tab1[i].c2 + lambda * (tab2[i].c2 - tab1[i].c2 );
   }
}

void receiver_speed(c_tabs **rtable,tabtabs *tt,params *param)
/*
 *   Calls interpolate_speed() above to create a sound speed profile
 *   at the receiver range valid in my linear interpolate in r models.
 *   This profile will later be used to record the upper and lower turning
 *   points of all rays in the wave front.
 */
{
double    r = param->rrange;
int       numsect = param->num.sections;
double    lambda;
c_tabs    *rtab;

   Safe_malloc(*rtable,c_tabs,numsect);
   rtab = *rtable;
   for(;tt->next != NULL; tt = tt->next) if (tt->r2 >= r ) break;
   lambda = (tt->r1 - r) * tt->r12_inv;
   interpolate_speed(rtab,tt->tab1,tt->tab2,lambda,numsect);
}

void get_max_profz(environs *environ,params *param)
/*
 *  Finds the profile contained in *environ which has the largest number
 *  of depth / sound speed pairs.  Sets  parm->num.prof_pts to this number
 *  and sets param->mlevz to the ARRAY of depths in this profile.
 */
{
profs   *prof,*mprof= NULL;
int     i,max_num = 0;
int     prof_cnt = 0;
int     *profz;
char    temp[80];

   for(prof=environ->prof;prof !=NULL;prof=prof->next) {
      prof_cnt++;
      if (prof->num_pnts > max_num) {
         max_num = prof->num_pnts;
         mprof = prof;
      }
   }

   param->num.profz = max_num;
   param->num.profs = prof_cnt;
   Safe_calloc(profz,int,max_num);
                     /* Round here   */
   for( i = 0; i < max_num; i++) profz[i] = (int) (mprof->depth[i] + 0.5);
   param->profz   = profz;      
   param->mprofz  = profz[max_num-1];

   sprintf(temp,"Deepest profile occured at range %g (m).",mprof->range);
   ldbug(temp);
   sprintf(temp,"It has %d points, the deepest is at %d (m).\n",\
           max_num,param->mprofz);
   ldbug(temp);

/*** check for integer values of depths ***/

   for(i=0; i< max_num; i++) {
      if (profz[i] != mprof->depth[i]) {
         
         sprintf(temp,"Warning: depth %d rounded from %g to %d\n",\
                 i+1,mprof->depth[i],profz[i]);
    ldbug(temp);
      }
   }
}

/****************************************************************************/
/*                   end of Utilites                                        */
/****************************************************************************/


/****************************************************************************/
/*                     de-bias routines                                     */
/****************************************************************************/
/*
 *  These routines get rid of the "bias" (speed offset)  in the
 *  sound speed tables, which is created by smoothing, be putting
 *  offsets in the original sound speed data contained in environ.
 */


void tridiag(double a[],double b[], double c[], double r[], double u[], double gam[],int n)
{
int j;
double bet;

   if ( b[0] == 0.0) fatal("Error 1 in tridiag");
   u[0] = r[0] /( bet = b[0]);
   for ( j= 1; j < n; j++) {
      gam[j] = c[j-1]/bet;
      bet = b[j] - a[j] * gam[j];
      if ( bet == 0.0 ) fatal("Error 2 in tridiag");
      u[j] = ( r[j] - a[j] * u[j-1]) /bet;
   }
   for ( j = (n-2); j >= 0; j--)
      u[j] -= gam[j+1] * u[j+1];
}

void debias1_ex(profs *prof,double *a,double *b, double *c,double *temp)
{
int num_pnts = prof->num_pnts;
double *cin  = prof->raw_speed;
double *cout = prof->speed;

   tridiag(a,b,c,cin,cout,temp,num_pnts);
}

void debias_ex(environs *environ,params *param)
/*
 *   Calculates new sound speeds for every *prof in *environ so that
 *   when these sound speeds are smoothed with the widths in
 *   param->smooth.width[].  The stuff that is the same for each
 *   profile is done just once here and then the debiasing happens
 *   a profile at a time in unbais1().
 */
{
profs *prof;
double *dz,*a,*b,*c,*temp;
int    i;
int    *profz    = param->profz;
int    num_profz = param->num.profz;
int    *w        = param->smooth.width;
int    max_i     = num_profz -1;



   ldbug("Exact debias selected.\n\n");

/*** get space for four arrays to be passed to debias1() ***/

   Safe_malloc(  dz,double,num_profz);
   Safe_malloc(   a,double,num_profz);
   Safe_malloc(   b,double,num_profz);
   Safe_malloc(   c,double,num_profz);
   Safe_malloc(temp,double,num_profz);


/*** fill in the arrays with depth spacing info  ***/

   for(i=0; i < max_i; i++)  dz[i] = (profz[i+1] - profz[i]) * 4.0;

   c[0] = 0.0; /* w[0] / dz[0]; */
   b[0] = 1.0 - c[0];
   a[0] = 0.0;
   for(i=1; i < max_i; i++) {
      a[i] = w[i] / dz[i-1];
      c[i] = w[i] / dz[i];
      b[i] = 1.0 - a[i] - c[i];
   }
   a[max_i] = 0.0; /* w[max_i] / dz[max_i - 1]; */
   b[max_i] = 1.0 - a[max_i];
   c[max_i] = 0.0;

/*** for each profile: check that its depth spacings agree with profz[] ***/
/***                 then debias that profile.   ROund here also...                       ***/

   for(prof = environ->prof;prof!=NULL; prof=prof->next) {
      for(i=0;i<prof->num_pnts;i++) {
         if ( param->profz[i] != (int) (prof->depth[i] + 0.5) ) fatal("Depths not aligned amongst profiles.");
      }
      debias1_ex(prof,a,b,c,temp);
   }

/*** free up space that was created above ***/

   free(temp);
   free(c);
   free(b);
   free(a);
   free(dz);
}

void debias1(profs *prof,double dz[],double mdz[],double temp[],params *param,int pcnt)
/*
 *  Uses the debiasing algorithm as given the the Ray report
 *  to debias a single sound speed profiles by adjusting the
 *  sound speeds in *prof before the are sent to the routines
 *  that create the smooth parameterizations.
 */
{
double   f1,f2,diff_c,new_c, max_diff_c = 0;
int      iter,i,max_i;
double   tol      = param->debias.tolerance;
int      *w       = param->smooth.width;
int      max_iter = param->debias.iteration;
int      num_pnts = prof->num_pnts;
double  *cin      = prof->raw_speed;
double  *cout     = prof->speed;


   if (max_iter == 0) max_iter = 10000;
   max_i = num_pnts - 1;

   Memcopy(cout,cin,num_pnts * sizeof(double) );
   for(iter = 1;iter <= max_iter ;iter++) {
      Memcopy(temp,cout,num_pnts * sizeof(double));
      for(i=1; i < max_i; i++) {
         f1 = w[i] * (temp[i-1] / dz[i] + temp[i+1] / dz[i+1] );
         f2 = 1.0 - w[i] / mdz[i];
         cout[i] = (cin[i] - f1) / f2;
      }
      /***  check for convergence here ***/

      max_diff_c = 0.0;
      for(i=1; i < max_i; i++) {
         new_c = cout[i] * ( 1.0 - w[i] / mdz[i] ) +
                 cout[i-1] * w[i] / dz[i] +
                 cout[i+1] * w[i] / dz[i+1];
         diff_c = fabs(cin[i] - new_c);
         max_diff_c = Max(max_diff_c,diff_c);
      }
      if (max_diff_c <= tol) break;
   }
   fprintf(Log_file,"%12d %12d %12d %12.6g\n",pcnt,num_pnts,iter,max_diff_c);
   if (Debug) printf("%12d %12d %12d %12.6g\n",pcnt,num_pnts,iter,max_diff_c);

}

void debias(environs *environ,params *param)
/*
 *   Calculates new sound speeds for every *prof in *environ so that
 *   when these sound speeds are smoothed with the widths in
 *   param->smooth_width[].  The stuff that is the same for each
 *   profile is done just once here and then the debiasing happens
 *   a profile at a time in unbais1().
 * NOTE dz[] and mdz[] are multiplied by 4.
 */
{
profs *prof;
double *dz,*mdz,*temp;
int    i;
int    *profz    = param->profz;
int    num_profz = param->num.profz;
int    pcnt      = 0;

   fprintf(Log_file,"%12s %12s %12s %12s\n","profile","num_pnts","iteration","max_error");
   if (Debug) printf("%12s %12s %12s %12s\n","profile","num_pnts","iteration","max_error");

/*** first get space for all of the data we want to keep ***/

   for(prof = environ->prof;prof!=NULL; prof=prof->next) {
      for(i=0;i<prof->num_pnts;i++)
         if ( param->profz[i] != prof->depth[i] ) fatal("Depths not aligned amongst profiles.");
   }

/*** get space for three arrays to be passed to unbias1() ***/

   Safe_malloc(dz,double,num_profz);
   Safe_malloc(mdz,double,num_profz);
   Safe_malloc(temp,double,num_profz);

/*** fill in the arrays with depth spacing info  ***/

   for(i=1; i < num_profz;  i++)   dz[i] = (profz[i] - profz[i-1]) * 4.0;
   for(i=1; i < num_profz-1; i++)  mdz[i] = (dz[i] * dz[i+1] ) / (dz[i] + dz[i+1] );

/*** for each profile: check that its depth spacings agree with mprofzz[] ***/
/***                 then debias that profile.                          ***/

   for(prof = environ->prof;prof!=NULL; prof=prof->next) {
      debias1(prof,dz,mdz,temp,param,++pcnt);
   }

/*** free up space that was created above ***/

   free(temp);
   free(dz);
   free(mdz);
}

/****************************************************************************/
/*                     end of de-bias routines                              */
/****************************************************************************/

/****************************************************************************/
/*                     make index routines                                  */
/****************************************************************************/
/*
 *  These routines are used to create the array  int Idex[]
 *  each element of Idex corresponds to a meter of depth in the
 *  ocean.   The element of Idex tells which parameterization of
 *  sound speed to use for that meter of ocean.
 */

void auto_widths_max(params *param)
{
int    *width;
int    i,max_i = param->num.profz-1;
int    *z = param->profz;

/*** get room for one width per profile point ***/

   Safe_malloc(param->smooth.width,int,max_i + 1);
   width = param->smooth.width;
   width[0] = z[1] - z[0];
   for(i = 1; i < max_i; i++) {
      width[i] = Min( z[i+1] - z[i], z[i] - z[i-1]) ;
   }
   width[max_i] = z[max_i] - z[max_i-1];
}

void fill_widths(params *param)
/*
 *   Takes care of converting widths input by user ("prof_smoothing")
 *   into an array I can use.  Also repeats last width if the number of
 *   widths is less than the maximum number of depth/sound speed pairs
 *   in any one profile ( param->num.profz_pts).  Uses Def_widths defined
 *   as a global if user does not specify "prof_smoothing".
 */
{
int   *new_widths;
int   i,last_width,intw;
dubs  *dub;
int   num_profz  = param->num.profz;
int   num_widths = 0;
int   count      = 0;


   if(param->smooth.auto_max) {
      auto_widths_max(param);
      return;
   }

/*** get room for one width per profile point ***/

   Safe_malloc(param->smooth.width,int,num_profz);
   new_widths = param->smooth.width;
/*** if user has given us some widths then use them ***/

   if (param->smooth.specific != NULL) {
      for(dub = param->smooth.specific; dub != NULL; dub = dub->next) {

/*** check for integer values ***/

         intw = (int) floor(dub->d + 0.5);
         if (intw != dub->d) {
            fprintf(Log_file,"Warning: width %d rounded from %g to %d\n",count,dub->d,intw);
            if (Debug) printf("Warning: width %d rounded from %g to %d\n",count,dub->d,intw);
         }

         new_widths[num_widths++] = intw;
         if (num_widths == num_profz) break;
      }
   }

/*** otherwise use my set of default widths for profzitus ***/

   else {
      num_widths = Min(Num_els(Def_widths),num_profz);
      Memcopy(new_widths,Def_widths,num_widths * sizeof(int));
   }

/***  in either case pad width array with repeat of last value ***/

   last_width = new_widths[num_widths-1];
   for (i = num_widths; i < num_profz; i++) new_widths[i] = last_width;
}

void fill_index(int index[],int *p,int d,int *nsect,int max_ii,params *param)
/*
 *    Used only in make_index below. Fills in a portion of index[]
 *    with the value *nsect.  Fills from index[*p] to index[d],
 *    incrementing *p along the way.  Then increments *nsect.
 */
{
int width = d - *p;
int margin = param->margin;

   d = Min(d,max_ii - margin);
   param->smooth.minz[*nsect] = *p;
   param->smooth.sect_width[*nsect] =  width;
   if (width > 0 )
      param->smooth.valid_sect[(param->smooth.num_valid)++] = *nsect;
   for( ;*p < d; (*p)++ ) index[*p+margin] = *nsect;
   (*nsect)++;
   index[0] = margin;
}


void log_widths(int *z,int *w,int max_i)
{
int i;

   fprintf(Log_file,"\n%8s %8s %8s %8s %8s\n\n",\
           "i","z","w","z-w[z]","z+w[z]");
   for (i = 0; i < max_i; i++)
     fprintf(Log_file,"%8d %8d %8d %8d %8d\n",\
            i,z[i],w[i],z[i]-w[i],z[i]+w[i]);
}
void do_width_error(int *z,int i)
{
char temp[80];

   sprintf(temp,"\nError: prof_smoothing widths are too wide at z[%d] = %d.",
          i,z[i]);
   ldbug(temp);
   ldbug("This is an easy error to fix.");
   ldbug("Set \"model prof_smoothing auto_max;\"");
   ldbug("to have Ray automatically calculate smoothing widths for you.");
   ldbug("You can also input your own widths with the line:");
   ldbug("\"model prof_smoothing specific = {w1, w2, ..., wn} (m);\"\n");
   fatal(NULL);
}

void make_index(int **index_tab,params *param)
/*
 * makes an array **index_tab of integers with nz + 2 * Margin points
 * which point to c_tabs of coefficients which we use to calculate
 * sound speed and its derivative.  nsect is set to the number
 * of sections.
 */
{
int i,w1,d1,d2,minz,maxz,*index;
int     *profz  = param->profz;
int     *width  = param->smooth.width;
int     max_i   = param->num.profz - 1;
int     w2      = width[1];
int     margin  = param->margin;
int     ii      = -margin;
int     nsect   = 0;
int     max_ii  = param->num.depths + 2 * margin -1;

/*** get room for the array ***/

   Safe_malloc(index,int,max_ii+1);
   *index_tab = index;

/*** pass one merely counts the number of sections     ***/
/*** and check to see if widths are too wide any where ***/

   log_widths(profz,width,max_i);
   nsect++;
   for (i= 1; i < max_i-1; i++) {
      w1 = w2;
      w2 = width[i+1];
      d1 = profz[i]   + w1;
      d2 = profz[i+1] - w2;
      if ( d2 < profz[i] || d1 > profz[i+1])
         do_width_error(profz,i);
      minz = Min(d1,d2);
      maxz = Max(d1,d2);
      nsect++;
      if ( minz == maxz ) continue;
      nsect++;
   }
   nsect++;
   nsect++;
   param->num.sections = nsect;

/*** now in pass two we actually do the filling in of idex ***/
/*** and we also save the minz and maxz of each section    ***/

   Safe_malloc(param->smooth.minz,int,nsect);
   Safe_malloc(param->smooth.sect_width,int,nsect);
   Safe_malloc(param->smooth.valid_sect,int,nsect);

/*** fill in the top margin area ***/

   nsect = 0;
   w2    = param->smooth.width[1];
   fill_index(index,&ii,profz[1]-w2,&nsect,max_ii,param);

   for (i= 1; i < max_i-1; i++) {
      w1 = w2;
      w2 = param->smooth.width[i+1];
      d1 = profz[i]   + w1;
      d2 = profz[i+1] - w2;
      minz = Min(d1,d2);
      maxz = Max(d1,d2);
      fill_index(index,&ii,minz,&nsect,max_ii,param);
      if ( minz == maxz ) continue;
      fill_index(index,&ii,maxz,&nsect,max_ii,param);
   }
   fill_index(index,&ii,profz[i] + w2,&nsect,max_ii,param);
   fill_index(index,&ii,param->num.depths + margin,&nsect,max_ii,param);
   if ( param->num.sections != nsect) fatal("internal error in fill_index");
}


/****************************************************************************/
/*                  end of make index routines                              */
/****************************************************************************/

/****************************************************************************/
/*                     auto range step routines                             */
/****************************************************************************/
/*
 *   These routines are used to constuct a parameterized table of
 *   range steps as a function of depth accessed via Idex as in  the
 *   ctab tables.  In each section the range step is controlled by
 *   the width of the section and is designed to be a continuous
 *   function of depth.
 *
 */

void fill_rstep(rstep_tabs *rstep,double z0,double dr0,double dr1,double dr2)
/*
 *  This should be a macro.  We just fill the three dr# components
 *  of rstep with the values passed in.
 */
{
   rstep->z0  =  z0;
   rstep->dr0 = dr0;
   rstep->dr1 = dr1;
   rstep->dr2 = dr2;
}

double scale_width(params * param,double w1,double w2)
{
double w;

   w = Min(w1,w2) * param->step.mult;
   w = Min(w,param->step.max);
   w = Max(w,param->step.min);
   return(w);
}

void compile_auto_rstep(rstep_tabs **rstep_tab,params *param)
/*
 *   Creates a table of rstep_tabs which can be accessed via Idex[z-Margin]
 *   to give as range steps as a function of depth.
 */
{
int         i,i2;
double      w,lastw,nextw;
rstep_tabs *rstep;
int         *width  = param->smooth.sect_width;
int         *minz   = param->smooth.minz;
int         *vs     = param->smooth.valid_sect;
int         max_i   = param->smooth.num_valid -1;
int         num_sect = param->num.sections;
double      min_width = param->step.max;
double      max_width = param->step.min;

/*** get room for the table of rstep parameters ***/

   Safe_calloc(rstep,rstep_tabs,num_sect);
   *rstep_tab = rstep;
   nextw = scale_width(param,width[0],width[1]);
   min_width = Min(min_width,nextw);
   max_width = Max(max_width,nextw);

   for(i=0;i <= max_i ;i++) {
      i2 = vs[i];
      lastw = nextw;
      w     = width[i2];
      if ( i+1 < max_i) {
         nextw = scale_width(param,w,width[vs[i+1]]);
         min_width = Min(min_width,nextw);
         max_width = Max(max_width,nextw);
      }
      fill_rstep(rstep+i2,minz[i2],lastw,(nextw - lastw)/ w,0.0);
   }
   fprintf(Log_file,"Minimum step_size = %g (m)\n",min_width);
   fprintf(Log_file,"Maximum step_size = %g (m)\n",max_width);
}

/****************************************************************************/
/*                 end of auto range step routines                          */
/****************************************************************************/

/****************************************************************************/
/*              compile sound speeds routines                               */
/****************************************************************************/

/*
 *   These routines create the smooth sound speed parameterizations.
 *   Each sound speed profile is compiled into an array of c_tabs.
 *   Each  c_tab gives the parameterization for the sound speed for
 *   a certain depth region of the ocean.  Which depths use which table
 *   is determined by Idex[] discussed above.  In addition, we create
 *   a linked list of tabtab structures which point to two "adjacent"
 *   ctabs and also contain the range information needed to linearly
 *   interpolate between the two ctabs.
 */

void fill_linear(c_tabs *tab,int sect,double z0,double c0, double beta)
/*
 *   Fills in a  tab entry where the sound-speed is linear.
 *   Used only by compile_speed() below.
 */
{
   tab += sect;
   tab->z0 = z0;
   tab->c0 = c0;
   tab->c1 = beta;
   tab->c2 = 0.0;
}

void fill_parab1(c_tabs *tab,int sect,double z0,double c0, double beta1,double beta2,double width)
/*
 *   Fills in a  tab entry where the sound-speed is a parbolic
 *   interpolation between two linear sections.
 *   Used only by compile_speed() below.
 */

{
   tab += sect;
   tab->z0 = z0;
   tab->c0 = c0 + width * (beta2 - beta1)/4.0;
   tab->c1 = (beta2 + beta1) / 2.0;
   tab->c2 = (beta2 - beta1) / ( 4.0 * width);
}

void fill_parab2(c_tabs *tab,int sect,c_tabs *ltab,c_tabs *ntab,double d1,double d2)
/*
 *   Fills in a  tab entry where the sound-speed is the overlap
 *   of two parabolic interpolations as in fill_parb1.
 *   Used only by compile_speed() below.
 *   Note:    d1 > d2
 */
{
double grad1,grad2;

   tab += sect;
   tab->z0 = d2;
   tab->c0 = tab_speed(ltab,d2);
   grad1   = tab_grad(ltab,d2);
   grad2   = tab_grad(ntab,d1);
   tab->c1 = grad1;
   tab->c2 = (grad2 - grad1) /(d1 - d2)/2.0 ;
}
void compile_one_speed(c_tabs **table,profs *prof,params *param)
/*
 * makes a table of coefficients to compute piece-wise parabolic
 * sound speed and its derivitive given a sound speed profile
 * and a table of widths.
 */
{
double  *profc;
int     *profz;
double  beta1,beta2,beta3;
int     i,w1,w2,max_i,section = 0;
c_tabs  *tab,*ltab;
c_tabs  ntab;
int     d1,d2;

   Safe_malloc(*table,c_tabs,param->num.sections);
   tab = *table;
   profz = param->profz;
   profc = prof->speed;
   max_i = prof->num_pnts - 1;

   w2 = param->smooth.width[1];
   beta2 = (profc[1] - profc[0]) / (profz[1] - profz[0]);
   if (max_i >= 2)
      beta3 = (profc[2] - profc[1]) / (profz[2] -profz[1]);
   else beta3 = beta2;
   fill_linear(tab,section,profz[0],profc[0],beta2);
   for (i= 1; i < max_i-1; i++) {
      beta1 = beta2;
      beta2 = beta3;
      beta3 = (profc[i+1] - profc[i+2]) / (profz[i+1] - profz[i+2]);
      w1 = w2;
      w2 = param->smooth.width[i+1];
      d1 = profz[i]   + w1;
      d2 = profz[i+1] - w2;
      fill_parab1(tab,++section,profz[i],profc[i],beta1,beta2,w1);
      if (d1 == d2) continue;
      if (d1  < d2) {
         fill_linear(tab,++section,profz[i],profc[i],beta2);
         continue;
      }
      fill_parab1(&ntab,0,profz[i+1],profc[i+1],beta2,beta3,w2);
      ltab = tab + section;
      fill_parab2(tab,++section,ltab,&ntab,d1,d2);
   }
   fill_parab1(tab,++section,profz[i],profc[i],beta2,beta3,w2);
   fill_linear(tab,++section,profz[max_i],profc[max_i],beta3);
   while( section < param->num.sections-1) {
      Memcopy(&tab[section+1],&tab[section],sizeof(c_tabs));
      section++;
   }
}

void compile_speeds(tabtabs **tabtab1,environs *environ,params *param)
/*
 *  For each profile (*prof) in *environ we create a ctab table which parameterizes
 *  the sound speeds in that profile.  For each adjacent pair of profiles we
 *  create a tabtab which points to two ctabs and contains their ranges.
 */
{
profs *prof1,*prof;
tabtabs *tabtab,*last;
double   srange = param->srange;

   /*** find last profile < Srange ***/

   for(prof1 = environ->prof; prof1->next !=NULL; prof1 = prof1->next)
      if (prof1->next->range > srange) break;

   if (prof1->range > srange) {
      ldbug("===> Warning: first profile is past the source.");
      prof1->range = srange;
   }
   last = NULL;
   param->num.tabs = 0;
   for( prof = prof1; prof != NULL; prof = prof->next) {
      (param->num.tabs)++;
      Safe_calloc(tabtab,tabtabs,1);
      tabtab->r1 = prof->range;
      compile_one_speed(&(tabtab->tab1),prof,param);
      if (last == NULL) *tabtab1 = tabtab;
      else {
         last->next    = tabtab;
         last->r2      = tabtab->r1;
         last->tab2    = tabtab->tab1;
         last->r12_inv = 1.0 /(last->r1 - last->r2);
      }
      if (tabtab->r1 >= param->rrange) return;
      last = tabtab;
   }
   last->r2      = param->rrange;
   last->tab2    = last->tab1;
   last->r12_inv = 0.0;
}

/****************************************************************************/
/*              end of compile sound speeds routines                        */
/****************************************************************************/

/****************************************************************************/
/*               compile bathymetry routines                                */
/****************************************************************************/

/*
 *   These routines are used to "compile" the bathymetry.
 *   They take as input the input bathymetry located in environ->bathy
 *   which is a  linked list of range depth pairs and produce as
 *   output a linked list of btabs contain the parabolic parameterization
 *   of the bathymetry.
 *   NOTE: The other two inputs needed are  param->bath_smooth,
 *   which is the maximum width we use for smoothing, and
 *   we also rely on Idex and Rstep to figure out the range steps
 *   in order to calculate safe minimum-z values (see fill_btab_minz() )
 *   NOTE: if param->bath_smooth is set to 0 then we don't do
 *   any smoothing and we just fill the btabs with linear segments.
 */

btabs *linlin_btab(bathys *bin1,bathys *bin2)
/*
 *  Used for filling in a linear btab segment connecting bin1 to bin2
 *  Only used by compile_lin_bathy() below.
 */
{
double dr = bin2->r - bin1->r;
double dz = bin2->z - bin2->z;
btabs *btab;

   Safe_calloc(btab,btabs,1);
   btab->r  = bin2->r;
   btab->z0 = bin2->z;
   btab->z1 = dz/dr;
   btab->z2 = 0.0;
   return(btab);
}

btabs *compile_lin_bathy(bathys *first_bathy)
/*
 *  Creates a linked list of btabs which linearly interpolate
 *  between all the points in the bathy linked list.
 *  Only called if param->bath_smooth was zero.
 */
{
bathys *bin1,*bin2;
btabs *btab,*first_btab;

   bin1 = first_bathy;
   bin2 = bin1->next;
   first_btab = btab = linlin_btab(bin1,bin2);
   while (bin2->next != NULL) {
      bin1 = bin2;
      bin2 = bin2->next;
      btab = btab->next = linlin_btab(bin1,bin2);
   }
   return(first_btab);
}

btabs *linear_btab(bathys *bin1,bathys *bin2,bathys *bin3,double max_width)
/*
 *  Fills in a btab with a linear section that would connect bin1 with bin2.
 *  We need bin3 in order to figure out the width of the parabolic section
 *  around the point at bin2 so we can know where this section will end.
 */
{
double dr1 = bin2->r - bin1->r;
double dr2 = bin3->r - bin2->r;
double width,beta;
btabs *btab;

   Safe_calloc(btab,btabs,1);
   width = Min(dr1,dr2) / 4.0;
   width = Min(width,max_width);
   beta     = (bin2->z - bin1->z) / (bin2->r - bin1->r);
   btab->r  = bin2->r - width;
   btab->z0 = bin2->z - width * beta;
   btab->z1 = beta;
   btab->z2 = 0.0;
   return(btab);
}

btabs *parabolic_btab(bathys *bin1,bathys *bin2,bathys *bin3,double max_width)
/*
 *  Fills in a btab with a parabolic section around bin2.  It will
 *  connect smoothly (matching value and 1st derivative) with a line
 *  from bin1 to bin2 at one end and a line from bin2 to bin3 at the
 *  other end.
 */
{
double dr1 = bin2->r - bin1->r;
double dr2 = bin3->r - bin2->r;
double width,beta1,beta2;
btabs *btab;

   Safe_calloc(btab,btabs,1);
   width = Min(dr1,dr2) / 4.0;
   width = Min(width,max_width);
   beta1 = (bin2->z - bin1->z) / dr1;
   beta2 = (bin3->z - bin2->z) / dr2;

   btab->r  = bin2->r + width;
   btab->z0 = bin2->z + beta2 * width;
   btab->z1 = beta2;
   btab->z2 = (beta2 - beta1)/width / 4.0;
   return(btab);
}

btabs *compile_parab_bathy(bathys *first_bathy,double max_width)
/*
 *  This routines makes the linked list of btabs that parameterize
 *  the bathymetry as a alternating series of linear and parbolic
 *  sections.
 */
{
bathys *bin1,*bin2,*bin3;
btabs *btab,*first_btab;

   bin1 = first_bathy;
   bin2 = bin1->next;
   bin3 = bin2->next;
   first_btab = btab = linear_btab(bin1,bin2,bin3,max_width);
   btab = btab->next = parabolic_btab(bin1,bin2,bin3,max_width);
   while (bin3->next != NULL) {
      bin1 = bin2;
      bin2 = bin3;
      bin3 = bin3->next;
      btab = btab->next = linear_btab(bin1,bin2,bin3,max_width);
      btab = btab->next = parabolic_btab(bin1,bin2,bin3,max_width);
   }
   btab->next= linear_btab(bin2,bin3,bin3,max_width);
   return(first_btab);
}

void compile_bathy(btabs **out,bathys *first_bathy,params *param)
{
/*
 *   Makes a lined list of btabs that parameterize the bathymetry points
 *   which exit in the linked list first_bathy.   Adds an extra bathy at
 *   at the beginning and end of the bathy list which will tend to make
 *   the first and last btab section flat.  To avoid this flattening effect
 *   just make sure that the source and receiver ranges
 *   lie within the range of the first and last bathymetry points which
 *   you specify.
 */

bathys *bathy,*last_bathy;
double max_width = param->bath_smooth;

/*** if no bathymetry points set **out to NULL and return ***/

   if (first_bathy == NULL) {
      *out = NULL;
      return;
   }

/*** add an extra bathymetry point 1 km before source range ***/

   Safe_calloc(bathy,bathys,1);
   bathy->r   = Min(param->srange,first_bathy->r) - 1000.0;
   bathy->z    = first_bathy->z;
   bathy->next = first_bathy;
   first_bathy = bathy;

/*** add an extra bathymetry point 1 km past receiver range ***/

   for(bathy= first_bathy;bathy->next != NULL;bathy=bathy->next);
   last_bathy  = bathy;
   Safe_calloc(bathy,bathys,1);
   bathy->r  = Max(param->rrange,last_bathy->r) + 1000.0;
   bathy->z  = last_bathy->z;
   last_bathy->next = bathy;

/*** choose whether to smooth or not ***/

   if (max_width > 0.0) *out = compile_parab_bathy(first_bathy,max_width);
   else                 *out = compile_lin_bathy(first_bathy);
}

double get_btab_minz(btabs *btab,double ra,double rb)
/*
 *    Finds minimum z value for all ranges between ra and rb.
 */
{
double minz,za,zb;
double z0 = btab->z0;
double z1 = btab->z1;
double z2 = btab->z2;
double dra = ra - btab->r;
double drb = rb - btab->r;

   za = z0 + dra * ( z1 + dra * z2);
   zb = z0 + drb * ( z1 + drb * z2);
   minz = Min(za,zb);
   if (z2 <= 0.0) return(minz);
   if ((z1 + 2.0 * dra * z2) * (z1 + 2.0 * drb * z2) >= 0.0) return(minz);
   return(z0 - z1 * z1 / (z2 * 4.0) );
}

double get_next_minz(btabs *btab,double r1,double r2)
/*
 *  Goes through the btab linked list getting minz from range r1 to range r2.
 */
{
double minz,next_minz;

   minz = get_btab_minz(btab,r1,Max(r2,btab->r) );
   while ( r2 > btab->r) {
      r1 = btab->r;
      btab = btab->next;
      if (btab == NULL) return(minz);
      next_minz = get_btab_minz(btab,r1,Max(r2,btab->r));
      minz = Min(minz,next_minz);
   }
   return(minz);
}

void fill_btab_minz(params *param,btabs *btab,int *idex,rstep_tabs *rstep)
/*
 * fills in the minz component of btab and its decendents with
 * the minimum z value over each section
 */
{
double dr;
double last_r = param->srange;

   while(btab->next != NULL) {
      dr         = 2.0 * tab_dr(rstep,idex,btab->z0);
      btab->minz = get_next_minz(btab,last_r,btab->r + dr);
      last_r     = btab->r;
      btab       = btab->next;
   }
   btab->minz = get_btab_minz(btab,last_r,Max(btab->r,param->rrange));
}

/****************************************************************************/
/*               end of compile bathymetry routines                         */
/****************************************************************************/


/****************************************************************************/
/*                reverse the ray trace routines                            */
/****************************************************************************/

void reverse_profiles(environs *environ)
/*
 *  Reverses the order of the profs linked list in environ and
 *  also multiplies each range in the profs by -1.0.
 */
{
profs *temp,*last_prof = NULL, *prof = environ->prof;

   while(prof != NULL) {
      prof->range *= -1.0;
      temp = prof->next;
      prof->next = prof->prev;
      prof->prev = temp;
      last_prof = prof;
      prof = prof->prev;
   }
   environ->prof = last_prof;
}

void reverse_bathymetry(environs *environ)
/*
 *  Reverses the order of the bathys linked list in    environ
 *  and also multiplies each range in the bathys by -1.0.
 */
{
bathys *last,*next,*bathy = environ->bathy;

   last = NULL;
   while( bathy != NULL) {
      bathy->r *= -1.0;
      next = bathy->next;
      bathy->next = last;
      last = bathy;
      bathy = next;
   }
   environ->bathy = last;
}

void reverse_ranges(params *param,environs *environ, \
          pinfos *pinfo1,pinfos *pinfo2)
/*
 * 1) Multiply all ranges in an environ structure by minus 1 (easy).
 * 2) reverse the linkage so that they all get processed from the
 *    bottom up. (harder).
 */
{
   reverse_profiles(environ);
   reverse_bathymetry(environ);
   param->srange     *= -1.0;
   param->rrange     *= -1.0;
   pinfo1->min_range *= -1.0;
   pinfo1->max_range *= -1.0;
   pinfo2->min_range *= -1.0;
   pinfo2->max_range *= -1.0;
}

/****************************************************************************/
/*            end of reverse the ray trace routines                         */
/****************************************************************************/

void get_max_bathy(params *param,environs *environ)
/*
 *  Makes sure that param->bottom is greater than the maximum
 *  bathymetry depth in environ->bathy.  If it is not we
 *  increase param->bottom to the new maximum value.
 *  Also sets param->num.depths to the appropriate value.
 */
{
bathys *bathy;
double max_profz = param->mprofz;
double max_depth = 0.0;

   if ( environ->bathy != NULL)  {
      for(bathy = environ->bathy; bathy != NULL; bathy = bathy->next)
         max_depth = Max(max_depth,bathy->z);
   }
	if (max_depth == 0.0 ) max_depth = param->bottom;
	if (max_depth == 0.0 ) max_depth = max_profz;	
   param->bottom = max_depth;
   param->num.depths = (int) param->bottom + 1;

   fprintf(Log_file,"bottom set to %g (m).\n",max_depth);
   if (Debug) printf("bottom set to %g (m).\n",max_depth);

   if ( environ->bathy != NULL)  return;

   /*** add fake bathymetry here for flat bottom case ***/

   fprintf(Log_file,"Creating flat bottom at %g (m).\n",max_depth);
   if (Debug) printf("Creating flat bottom at  %g (m).\n",max_depth);

   Safe_calloc(bathy,bathys,1);
   environ->bathy = bathy;
   bathy->z = max_depth;
   bathy->r = Min(param->srange,param->rrange);
   Safe_calloc(bathy,bathys,1);
   environ->bathy->next = bathy;
   bathy->z = max_depth;
   bathy->r = Max(param->srange,param->rrange);
}


/****************************************************************************/
/*               main pre-processing routine                                */
/****************************************************************************/

void pre_process(tabtabs **tt,c_tabs **rtab,int **idex,rstep_tabs **rstep,\
       btabs **btab,environs *environ,params *param,            \
       pinfos *pinfo1,pinfos *pinfo2)
/*
 *  This routine is the sole interface for preray.c.  It is almost the only
 *  fuction declared in preray.h so if you want to know what happens
 *  in preray.c and in what order LOOK HERE.
 */
{
   if (param->rrange - param->srange < 0.0) {
      param->sign_dr = -1.0;
      ldbug("\n===> Reverse ray trace <===\n\n");
      reverse_ranges(param,environ,pinfo1,pinfo2);
                          /* Reverse order of profs & bathys,  */
                          /* multiply all ranges by -1.0       */
   }
   else param->sign_dr = 1.0;

   ldbug("Getting deepest sound speed profile.");
   get_max_profz(environ,param);              /* Find prof with maximum number    */
                                              /* of levels.                       */
   ldbug("Getting maximum bathymetry depth.");
   get_max_bathy(param,environ);      /* Set param->bottom to maximum     */
                                            /* depth from bathymetry            */

   ldbug("Resolving profile smoothing.");
   fill_widths(param);                  /* Load in smooth.specific widths       */
                                        /* or my defaults.  Pad extra       */
                                        /* levels with last value.          */
   ldbug("Making Idex.");
   make_index(idex,param);              /* Create Idex[] which points to    */
                                        /* ctabs and rstep_tabs as a        */
                                        /* function of depth.               */
   ldbug("Compiling range step table.");
   compile_auto_rstep(rstep,param);     /* Create table of rstep_tabs which */
                                        /* give range step as a fuction of  */
                                        /* depth.                           */

   ldbug("Debiasing sound speed profiles.");
   if ( param->debias.exact) debias_ex(environ,param);
   else debias(environ,param);
                                        /* Change values of sound speeds    */
                                        /* in *environ so that when they are    */
                                        /* smoothed the smoothed curves     */
                                        /* will hit orignal points.         */

   ldbug("Compiling sound speed tables.");
   compile_speeds(tt,environ,param);  /* Create an array of ctabs for     */
                                        /* prof and create a linked list of */
                                        /* tabtabs point to adjacent ctabs  */
                                        /* arrays.                          */
   ldbug("Compiling profile at receiver.");
   receiver_speed(rtab,*tt,param);/* Make a ctab for the sound speed  */
                                        /* at the receiver range.           */

   ldbug("Getting sound speed at source.");      /* find sound speed at source       */
   param->source_c = get_speed(*tt,*idex,param->srange,param->sdepth);

   if (environ->bathy != NULL)
   {
      ldbug("Compiling bathymetry.");
      compile_bathy(btab,environ->bathy,param); /* Make linked list of btabs       */
                                                  /* which contain parameterization  */
                                                  /* of bathymetry.                  */

      fill_btab_minz(param,*btab,*idex,*rstep);   /* Fill in the minz component with */
                                                  /* the highest we can hit while    */
                                                  /* stepping in each btab.          */
   }
   ldbug("End of preprocessing.");
}

/****************************************************************************/
/*                  end of PRE--PROCESSING                                  */
/****************************************************************************/

