/****************************************************************************/
/*                        program ray.c                                     */
/****************************************************************************/
/*
 *      Copyright (C) 1992-1994,  Woods Hole Oceanographic Institution
 *    See LICENSE file for use, copying and distribution conditions
 *
 *
 *  jim bowlin Mar. 1994.
 *
 *  designed to do fast ray tracing.
 *
 * Changes: Rich Pawlowicz Aug 93
 *            - small changes to make the output for eigenray searches
 *              look nicer; added extra parameter to shoot_all_angles
 */

#define  Main

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <setjmp.h>

#include "utils.h"
#include "ray.h"
#include "preray.h"
#include "outray.h"
#include "eigenray.h"
#include "readray.h"

#undef Main

#if !defined(__MSDOS__)
   #include <unistd.h>
   #include <memory.h>
#endif

/******************** Pointers to functions *************************/

/*
 *   NOTE:  The alignment in the arrays below MUST match the alignment
 *          in the associated intab structures in initray.c :
 *               ran_dep_lst, integ_lst, bathy_lst
 *
 *   Derivative: points to the routine we use for calculating the
 *               derivatives of a ray element.  This choice controls
 *               how we handle the range dependence.
 *
 */

derivatives Derivative;

void  dds_frd(rels *s,double *c_inv,double *dtheta);
void  dds_nrd(rels *s,double *c_inv,double *dtheta);
void  dds_zrd(rels *s,double *c_inv,double *dtheta);

derivatives Derivative_tab[] = {
   dds_frd,
   dds_nrd,
   dds_zrd,
};

void    integrate_dl(double dl);
void    integrate_dr(double dr);
double  bottom(double r,btabs *btab);

/******************* GLOBALS USED IN CRITICAL LOOP **************************/

/****************************************************************************/
/*                 HERE ARE THE LOOK-UP TABLES                              */
/****************************************************************************/
/*
 *   Idex
 *      is a table of integers.  For a given depth z, Idex[z-Margin]
 *      will give the index of the tables Tab1,Tab2,Rtab,and Rstep_tab.
 *
 *   Tab1,Tab2
 *      give the sound speed as a fuction of z at ranges Range1 and Range2.
 *      See e.g. preray.tab_speed() for their use.
 *
 *   TabTab
 *      is a linked list of pointers to tab1, tab2, range2, range2 and
 *      r12_inv.  As we shoot a ray we walk through this table to get
 *      a range dependent c(z).
 *
 *   Rstep_tab
 *      is similar to Tabs above.  There is just one though and it
 *      contains to coefficients to calculate a depth dependent
 *      range step size.  See tab_dr() below.
 *
 *   B_tab
 *     is a linked list of tables used to calculate where the bottom
 *     is see e.g. bathymetry() for its use.
 */

int        *Idex;          /* [nz+2*Saftey]  contains index into Tab */
c_tabs     *Tab1,
           *Tab2;   /* [nsections] {double z0, c0, dc/dz, d2c/dz2 /2 }*/
tabtabs    *Tabtab;
c_tabs     *Rtab;
rstep_tabs *Rstep_tab;
btabs      *First_btab,   /* beginning of btab linked list */
           *B_tab;        /* current  pointer in btab list */

/****************************************************************************/
/*                     end of LOOK-UP TABLES                                */
/****************************************************************************/

int Num_dots;
int Saving_pinfo;
pinfos *Pinfo;

jmp_buf Abort_buf;

rinfos Rinfo1;

rels S1,S2,S3,S4,S5,S6;
rels Ds1,Ds2,Ds3,Ds4;
static rels *Find_speed_rel;

double One_half = 1.0 /2.0;
double One_sixth = 1.0 /6.0;

double S_earth;

double Range1;         /*  range of last profile                          */
double Range2;         /*  range of next profile                          */
double R12_inv;        /*  1.0 / (Range1 - Range2)                        */
double R_earth_inv;    /*  1/R_earth  used for spherical earth correction */

double Cos_b;
double Cos_a;
environs        *Environ;
losses          *Loss;


/************** End of GLOBALS USED IN CRITICAL LOOP  ***********************/

/****************************************************************************/
/*                 end of global variables                                  */
/****************************************************************************/

int maybe_reflect(void);
void write_pinfo(rels *rel);
void advance_pinfo(void);
time_t      start_timing(long n);                               /* ray.c     */
void        end_timing(long n,time_t start_time);               /* ray.c     */

/****************************************************************************/
/*                     debugging stuff                                      */
/****************************************************************************/

#define Abort_normal 1
#define Abort_error  2

void process_exception(int abort_flag,long ray_number)
/*
 *  print out everything we can about S1,S2,Rinfo1 into logfile.
 */

{
char s[100];
   
   sprintf(s,"Ray %ld aborted %s. Launch angle was %12g (degrees)",
         ray_number, (abort_flag == Abort_normal) ? "normally" : "ABNORMALLY",
         Rinfo1.sangle *180.0/Pi);
   if ( abort_flag == Abort_error) {
      fputs("\n",stdout);
      vlputs(s);
   }
   else {
      ldbug(s);
}
   sprintf(s,"%12s %12s %12s %12s %12s",
           "step","range","depth","angle","bottom");
   ldbug(s);
   sprintf(s,"%12d %12g %12g %12g %12g",
           1,S1.r,S1.z,atan2(S1.s,S1.c) * 180 /Pi,bottom(S1.r,B_tab));
   ldbug(s);
   sprintf(s,"%12d %12g %12g %12g %12g",
           2,S2.r,S2.z,atan2(S2.s,S2.c) * 180 /Pi,bottom(S2.r,B_tab));
   ldbug(s);
   Num_dots = 0;
}

/****************************************************************************/
/*                       turning point stuff                                */
/****************************************************************************/
double rspeed(double z);

double rspeed(double z)
{
c_tabs *tab;
double dz;

   tab = Rtab + Idex[(int) z + Idex[0]];
   dz = z - tab->z0;
   return(tab->c0 + dz * ( tab->c1 + dz * tab->c2));
}

double find_tp(double z1, double angle, double z2, double (*speed)(double))
{
double zm;
double c1,c2,cm,c0;
double ztol = 0.5;

   c1 = (*speed)(z1);
   c2 = (*speed)(z2);
   c0 = c1 / cos(angle);
   if (c2 <= c0) return(z2);
   do {
      zm = (z1 + z2)/2.0;
      cm = (*speed)(zm);
      if ( (cm - c0) * (c2 - c0) >= 0.0) {
         c2 = cm;
         z2 = zm;
      }
      else {
         c1 = cm;
         z1 = zm;
      }
   } while ( fabs( z1 -z2 ) > ztol );
   return( (z1 + z2) / 2.0);
}

double find_utp(double z,double angle,double (*speed)(double) )
{
   return( find_tp(z,angle,Param.top,speed) );
}

double find_ltp(double z,double angle,double (*speed)(double) )
{
   return( find_tp(z,angle,Param.bottom,speed));
}

void find_tps(rinfos *rinfo)
{
double angle = rinfo->rangle;
double z     = rinfo->rdepth;

   rinfo->utp = find_utp(z,angle,rspeed);
   rinfo->ltp = find_ltp(z,angle,rspeed);
}

/****************************************************************************/
/*                  end of turning point stuff                              */
/****************************************************************************/

/****************************************************************************/
/*                     bottom interaction routines                          */
/****************************************************************************/


double bottom_grad(double r,btabs *btab)
{
double dr;

   dr = r - btab->r;
   return( btab->z1 + 2.0 * dr * btab->z2 );
}

int get_roots(double roots[],double z[])
/*
 * finds real roots of equation:
 *                0 = z[0] + x * z[1] + x^2 * z[2]
 * returns number of real roots found.
 * Sets *root1 to smallest root
 * Sets *root2 to largest  root
 * if no real roots are found sets *root1 and *root2 at
 * extremum.
 */
{
double temp,q;
int    cnt  = 0;
double c    = z[0];
double b    = z[1];
double a    = z[2];
double disc = b * b - 4.0 * a * c;

   if ( disc <= 0.0) {
      if ( a != 0.0 )   *roots = - b / ( 2.0 * a);
      else              return(-1);
      if ( disc == 0.0) return(1);
      return(0);
   }
   q = (b > 0.0) ? -0.5 *(b + sqrt(disc)) : -0.5 * (b - sqrt(disc));
   if (a != 0.0 ) roots[cnt++] = q/a;
   if (q != 0.0 ) roots[cnt++] = c/q;
   if (cnt < 2 )  return(cnt);
   if (roots[0] > roots[1]) {
      temp = roots[0];
      roots[0] = roots[1];
      roots[1] = temp;
   }
   return(cnt);
}

int get_smp_root(double *root,double z[])
/*
 * returns smallest positive root.
 *
 */
{
double roots[2];
int    num = get_roots(roots,z);

   *root = roots[0];
   switch (num) {
   case 2 :
      if (roots[0] >= 0.0) return(num);
      if (roots[1] >=  0.0) {
         *root = roots[1];
         return(1);
      }
      break;
   case 1 :
      if (roots[0] >= 0.0) return(1);
      break;
   case 0 :
      return(0);
   }
   return(-1);
}

void param_approach(double z[],rels *si,rels *so,btabs *btab)
/*
 * approach to bottom parameterized as
 *
 *      z[0] + dr * z[1] + dr^2 * z[2]
 *
 * Use parabolic approximation to the ray path and use
 * bathymetry parameterization at si->r
 * Fills z[] with appropriate values
 */
{
double dr,rz0,rz1,rz2,bot;

/*** parameterize trajectory from *si to *so as parabolic ***/

   dr  = so->r - si->r;
   rz0 = si->z;
   rz1 = si->s / si->c * (1.0 -  si->z * R_earth_inv);
   rz2 = (so->z - rz0 - dr * rz1 )/ (dr * dr);

/*** parameterize bathymetry from *si to *so as parabolic ***/
/*** and subtract the ray parameterization                ***/
  /* ( could make this more compact and more obscure ) */

   bot  = bottom(si->r,btab);
   z[0] = bot                     - rz0;
   z[1] = bottom_grad(si->r,btab) - rz1;
   z[2] = btab->z2                - rz2;
}

int hit_root(btabs *btab,double dr,double max_dr)
/*
 *   Tries to step the ray to just touch the bottom.
 *   Since the bottom is parabolic the only reason we won't
 *   hit exactly on first go round is because we approximated
 *   the ray path as parabolic.  Returns 1 if we find a positive
 *   root within max_dr, otherwise returns 0.
 */
{
int    num_roots, count = 0;
double z[3],roots[2],bottom_miss;

   integrate_dr(dr);
   bottom_miss = S2.z - bottom(S2.r,btab);
   while (  fabs(bottom_miss) > Param.z_tol ) {
      if (count++ > 10) return(0);
      param_approach(z,&S2,&S1,btab); /* note that we expand about S2 */
      num_roots = get_roots(roots,z);
      if (num_roots <= 0) return(0);
      dr += (dr + roots[0] < 0.0) ? roots[1] : roots[0];
      if (dr < 0.0) {
         ldbug("dr negative in \"hit_root()\"");
         longjmp(Abort_buf,Abort_error);
      }
      if (dr > 2.0 * max_dr) return(0);
      integrate_dr(dr);
      bottom_miss = S2.z - bottom(S2.r,btab);
   }
   /* S2.z = bottom(S2.r,btab); */
   return(1);
}

void compute_loss(double s1,double c1,double s2, double c2,double range)
{
double atten,atten1,atten2,angle,angle1,angle2;
double lambda,s,c;
int    i;
   
   if (Loss == NULL) return;

   /*** advance to proper loss table ***/
   
   while ( Loss->next != NULL) {
      if (Loss->next->range >= range) break;
      Loss = Loss->next;
   }

   /*** calculate angle with respect to normal ***/

   if ( Loss->num_pnts < 1) return;   
   if ( Loss->num_pnts == 1) {
      Rinfo1.bottom_loss += Loss->atten[0];
      return;
   }
   
   s = s1 * c2 - c1 * s2;
   c = c1 * c2 + s1 * s2;
   angle = atan2( fabs(s),c);       /* grazing angle */

   /*** find angle in lookup table ***/

   for ( i = 1; i < Loss->num_pnts; i++) 
      if (Loss->angle[i] > angle) break;

   angle1 = Loss->angle[i-1];
   angle2 = Loss->angle[i];
   atten1 = Loss->atten[i-1];
   atten2 = Loss->atten[i];

   /*** interpolate attenuation ***/

   if ( angle1 == angle2) atten = atten1;
   else {
      lambda = (angle - angle1) / (angle2 - angle1);
      atten  = ( 1.0 - lambda) * atten1 + lambda * atten2;
   }
   Rinfo1.bottom_loss += atten;
}

void reflect_ray(void)
{
double sinb;
double cosb;
double new_angle;
double dot_prod,snorm;

   if ( Param.bottom_type == 1) {
      ldbug("Ray absorbed.");
      longjmp(Abort_buf,Abort_normal);
   }
   sinb = bottom_grad(S2.r,B_tab);
   sinb = sinb / sqrt( 1 + sinb * sinb);     /* sin(angle) of bottom */
   cosb = sqrt( 1 - sinb * sinb);            /* cos(angle) of bottom */
   snorm = 1.0 / sqrt( S2.s * S2.s + S2.c * S2.c);
   S2.s  *= snorm;
   S2.c  *= snorm;
   compute_loss(S2.s,S2.c,sinb,cosb,S2.r);
   dot_prod =  S2.s * sinb + S2.c * cosb;
   S2.s = 2.0 * (dot_prod * sinb) - S2.s;
   S2.c = 2.0 * (dot_prod * cosb) - S2.c;
   new_angle = atan2(S2.s,S2.c);
   if (fabs(new_angle) > Param.max_angle) {
      ldbug("Maximum angle exceeded in \"reflect_ray()\"");
      longjmp(Abort_buf,Abort_normal);
   }
}

void save_bounce_point(double sign_dz)
{
                                              /*    sign_dz:   1      -1   */
double tb_adjust =  Max(0.0,-sign_dz/2.0);    /*               0       1/2 */
double bb_adjust =  Max(0.0, sign_dz/2.0);    /*              1/2       0  */

   Rinfo1.top_bonks += tb_adjust;
   Rinfo1.bot_bonks += bb_adjust;
   Rinfo1.tot_bonks += bb_adjust + tb_adjust;
   if( Saving_pinfo) if (Pinfo->inc_bonks != 0) write_pinfo(&S2);
   Rinfo1.top_bonks += tb_adjust;
   Rinfo1.bot_bonks += bb_adjust;
   Rinfo1.tot_bonks += bb_adjust + tb_adjust;
   if (Rinfo1.tot_bonks > Param.max_bonks) {
      if (Param.max_bonks > 0) {
         ldbug("maximum bounces exceeded.");
         longjmp(Abort_buf,Abort_normal);
      }
   }
}

int check_seam(void)
/*
 * makes sure that the end of this step does not lie past the end of the
 * current btab.  If it does then we
 *           step up to the seam
 *           advance the bathymetry
 *           check for refections from the new bathymetry section
 *
 *  Note that check_seam and maybe_reflect are a recursive pair.
 */
{
   if ( S2.r <= B_tab->r) return(0);
   integrate_dr(B_tab->r - S2.r);
   B_tab = B_tab->next;
   if (B_tab == NULL) fatal("End of bathymetry inside of check_seam().");
   return(maybe_reflect());
}

int maybe_reflect()
/*
 *   ASSUMES that S2 was the result of a step from S1
 *   Does more carefull checking of crossing the bathymetry,
 *   IF the bathymetry was crossed we bounce off of it correctly,
 *   otherwise we just take a normal step.
 *   Returns 1 if there was a bounce, otherwise returns 0.
 *   S1 unchanged,
 *   S2 contains result of bounce (or unchanged if none)
 */
{
double z[3],root;
double max_dr;
rels temp_s2,temp_s1;
double dr;
int ret_val;

   dr = S2.r - S1.r;
   Memcopy(&temp_s2,&S2,sizeof(rels));
   Memcopy(&temp_s1,&S1,sizeof(rels));

   max_dr = 2.0 * dr;
   param_approach(z,&S1,&S2,B_tab);

   if (z[0]  < 0.0) {
      ldbug("started past boundary in \"maybe_reflect().\"");
      longjmp(Abort_buf,Abort_error);
   }

   get_smp_root(&root,z);

   if ((root <= 0.0)||(root > max_dr)) {
      ret_val = check_seam();
      Memcopy(&S2,&temp_s2,sizeof(rels));
      return(ret_val);
   }

   if (hit_root(B_tab,root,max_dr) == 0 ) {
         ret_val = check_seam();
      Memcopy(&S2,&temp_s2,sizeof(rels));
      return(ret_val);
   }
   if ((S2.r > B_tab->r) ||( S2.r > temp_s2.r)) {
      ret_val = check_seam();
      Memcopy(&S2,&temp_s2,sizeof(rels));
      return(ret_val);
   }

/*** we only get here if we are going to reflect a ray ***/

   Memcopy(&S1,&temp_s1,sizeof(rels));
   if ( Saving_pinfo )     advance_pinfo();
   save_bounce_point(1.0);
   reflect_ray();
   Memcopy(&S1,&S2,sizeof(rels));
   integrate_dr(temp_s2.r - S2.r);
   if ( Saving_pinfo )     advance_pinfo();
   Memcopy(&S1,&temp_s1,sizeof(rels)); 
   return(1);
}

/****************************************************************************/
/*                 end of bottom interaction routines                       */
/****************************************************************************/

/****************************************************************************/
/*                         bounce stuff                                     */
/****************************************************************************/

int maybe_bounce(double bound,double sign_dz)
/*
 *   ASSUMES that S2 was the result of a step from S1
 *   Does more carefull checking of crossing the boundary bound.
 *   IF the boundary was crossed we bounce off of it correctly,
 *   otherwise we just take a normal step.
 *   Returns 1 if there was a bounce, otherwise returns 0.
 *   S1 unchanged,
 *   S2 contains result of bounce (or unchanged if none)
 */
{

double z[3],root;
rels   temp_s1,temp_s2;
btabs  btab;
double dr = S2.r - S1.r;
double max_dr = 2.0 * dr;

   btab.r = S1.r;
   btab.z0 = bound;
   btab.z1 = 0.0;
   btab.z2 = 0.0;
   param_approach(z,&S1,&S2,&btab);

   if (z[0] * sign_dz < 0.0) {
      ldbug("started past boundary \"maybe_bounce().\"");
      longjmp(Abort_buf,Abort_error);
   }
   get_smp_root(&root,z);

   if ((root <= 0.0)||(root > max_dr )) return(0);

   Memcopy(&temp_s2,&S2,sizeof(rels));
   Memcopy(&temp_s1,&S1,sizeof(rels));
   if (hit_root(&btab,root,max_dr) == 0 ) {
      Memcopy(&S2,&temp_s2,sizeof(rels));
      return(0);
   }
   if (S2.r > temp_s2.r){
      Memcopy(&S2,&temp_s2,sizeof(rels));
      return(0);
   }

/*** we only get here if we are going to bounce a ray ***/

   Memcopy(&S1,&temp_s1,sizeof(rels));
   if ( Saving_pinfo )     advance_pinfo();  
   save_bounce_point(sign_dz);
   S2.s = - S2.s;                           /* this reflects the ray */
   Memcopy(&S1,&S2,sizeof(rels));
   integrate_dr(temp_s2.r - S1.r);
   if ( Saving_pinfo )     advance_pinfo();
   Memcopy(&S1,&temp_s1,sizeof(rels));
   return(1);
}

void count_turns(void)
{
  if (S2.s * S1.s > 0.0) return;
  if (S2.s == 0.0)       return;
  if (S2.s < 0.0)  (Rinfo1.bot_turns)++;
  else             (Rinfo1.top_turns)++;
}

/****************************************************************************/
/*                     end of bounce stuff                                  */
/****************************************************************************/


/****************************************************************************/
/*                 tracing routines                                         */
/****************************************************************************/

double tab_dr(rstep_tabs *rstep,int *idex,double z)
{
double dz;

   rstep += idex[(int) z + idex[0] ];
   dz = z - rstep->z0;
   return( rstep->dr0 + dz * rstep->dr1);
}


void vprint_dot(long bounces)
/*
 * if Verbose is on we print either a "." if no new bounces have
 * occured or a "*" if we have some new bounces.
 */
{
   if(!Verbose) return;
   bounces -= (int) Rinfo1.tot_bonks;
   Num_dots++;
   putchar((bounces) ? '*' : '-');
   fflush(stdout);
}


/****************************************************************************/
/*                 tracing routines while saving path                       */
/****************************************************************************/

/*
 *  These routines piggy back ontop of the tracing routines that don't
 *  save path info as much as possible.  Put the pot on the floor
 *  and procede as in problem A.
 */

double find_speed(double z)
{
int     iiz;
c_tabs  *tab1,*tab2;
double  dz1,dz2,speed1,speed2;

   iiz    = Idex[(int) z + Idex[0]];
   tab1   = Tab1 + iiz;
   tab2   = Tab2 + iiz;
   dz1     = z - tab1->z0;
   dz2     = z - tab2->z0;   
   speed1 = tab1->c0 + dz1 * ( tab1->c1 + dz1 * tab1->c2 );
   speed2 = tab2->c0 + dz2 * ( tab2->c1 + dz2 * tab2->c2 );
   return(speed1 + (Range1 - Find_speed_rel->r) * R12_inv * (speed2 - speed1));
}

double find_grad(rels *s)
{
int     iiz;
c_tabs  *tab1,*tab2;
double  z,dz1,dz2,grad1,grad2;

   z      = s->z;
   iiz    = Idex[(int) z + Idex[0]];
   tab1   = Tab1 + iiz;
   tab2   = Tab2 + iiz;

   dz1     = z - tab1->z0;
   dz2     = z - tab2->z0;   
   grad1  = tab1->c1 + dz1 * (tab1->c2 + tab1->c2);
   grad2  = tab2->c1 + dz2 * (tab2->c2 + tab2->c2);
   return(grad1 + (Range1 - s->r) * R12_inv * (grad2 - grad1));
}

void begin_pinfo_mat(FILE *file,pinfos *pinfo,char *name)
/*
 *  Writes a matlab header to *file, remembers the file for future
 *  writes. Also rememebers the location in the file to rewrite the
 *  header with the correct column and row information.
 *  Also sets default values of min_range and max_range.
 *  This must be called AFTER reverse_ranges() if we are
 *  doing a reverse ray trace in order for the order
 *  of min_range and max_range to be correct.
 */
{
double temp;

   save_pinfo_help(file,name,pinfo);
   pinfo->header = ftell(file);
   pinfo->file   = file;
   pinfo->write_count  = 0;
   pinfo->ray_count = 0.0;
   start_swapmat(file,name,pinfo->write_count,pinfo->num_cols);
   if (pinfo->min_range > pinfo->max_range) {
      temp = pinfo->min_range;
      pinfo->min_range = pinfo->max_range;
      pinfo->max_range = temp;
   }
}

void write_asc_head(FILE *file,int flag,char *label)
{
   if (flag != 1) return;
   fprintf(file,"%15s",label);
}

void begin_pinfo_asc(FILE *file,pinfos *pinfo,char *name)
{
   if (file == NULL) return;
   fprintf(file,"\nbegin_%s\n",name);
   write_asc_head(file,pinfo->saveflg[0],"range");
   write_asc_head(file,pinfo->saveflg[1], "depth");
   write_asc_head(file,pinfo->saveflg[2], "time");
   write_asc_head(file,pinfo->saveflg[3],  "angle");
   write_asc_head(file,pinfo->saveflg[4],"speed");
   write_asc_head(file,pinfo->saveflg[5],"grad_z");
   write_asc_head(file,pinfo->saveflg[6],"length");
   write_asc_head(file,pinfo->saveflg[7],"length_c2");
   write_asc_head(file,pinfo->saveflg[8],"p_z");
   write_asc_head(file,pinfo->saveflg[9],"top_bounces");
   write_asc_head(file,pinfo->saveflg[10],"bot_bounces");
	   write_asc_head(file,pinfo->saveflg[11],"tot_bounces");
   write_asc_head(file,pinfo->saveflg[12],"upper_tp");
   write_asc_head(file,pinfo->saveflg[13],"lower_tp");
      write_asc_head(file,pinfo->saveflg[14],"bottom_loss");
   write_asc_head(file,pinfo->saveflg[15],"ray_number");
   fprintf(file,"\n");
}


void write_pinfo_1asc(FILE *file,double value)
{
   fprintf(file,"%15g",value);
}

void write_pinfo_asc(rels *rel)
{

FILE *file;
double speed;
double z = rel->z;
double a = -atan2(rel->s,rel->c);

   if (Asc_file == NULL) return;
   file = Asc_file;
   Find_speed_rel = rel;
   speed = find_speed(z);
   if (Pinfo->saveflg[0]) write_pinfo_1asc(file, Param.sign_dr * rel->r);
   if (Pinfo->saveflg[1]) write_pinfo_1asc(file, Param.sign_z  * z);
   if (Pinfo->saveflg[2]) write_pinfo_1asc(file,rel->t);
   if (Pinfo->saveflg[3]) write_pinfo_1asc(file,a);
   if (Pinfo->saveflg[4]) write_pinfo_1asc(file,speed);
   if (Pinfo->saveflg[5]) write_pinfo_1asc(file,find_grad(rel));
   if (Pinfo->saveflg[6]) write_pinfo_1asc(file,rel->l);
   if (Pinfo->saveflg[7]) write_pinfo_1asc(file,rel->l_c2);
   if (Pinfo->saveflg[8]) write_pinfo_1asc(file,rel->s / speed);
   if (Pinfo->saveflg[9]) write_pinfo_1asc(file,Rinfo1.top_bonks);
   if (Pinfo->saveflg[10]) write_pinfo_1asc(file,Rinfo1.bot_bonks);
	if (Pinfo->saveflg[11]) write_pinfo_1asc(file,Rinfo1.bot_bonks +
														       Rinfo1.top_bonks);
   if (Pinfo->saveflg[12]) write_pinfo_1asc(file,find_utp(z,a,find_speed) );
   if (Pinfo->saveflg[13]) write_pinfo_1asc(file,find_ltp(z,a,find_speed) );
   if (Pinfo->saveflg[14]) write_pinfo_1asc(file,Rinfo1.bottom_loss );
   if (Pinfo->saveflg[15]) write_pinfo_1asc(file,Pinfo->ray_count);
   fprintf(file,"\n");
}

void save_pindex(pinfos *pinfo,char *name)
{
FILE    *file = pinfo->file;
char    idex_name[80];
pindexs *pindex;
	
	strcpy(idex_name,name);
	strcat(idex_name,"_idex");
	start_swapmat(file,idex_name,pinfo->ray_count,2);
	for( pindex = pinfo->first_index; pindex != NULL; pindex = pindex->next) {
		fwrite(&pindex->starti,sizeof(double),1,file);
		fwrite(&pindex->endi,  sizeof(double),1,file);
	}
}

void flush_pinfo_mat(pinfos *pinfo,char *name)
/*
 * rewinds file to pinfo header.  rewrites the header with the
 * now known number of rows.  then rewinds to end of file.
 * Then write out the index file.
 */
{
FILE *file;

   if (pinfo == NULL) return;
   if ((file = pinfo->file) == NULL) return;
   fseek(file,pinfo->header,SEEK_SET);
   start_swapmat(file,name,pinfo->write_count,pinfo->num_cols);
   fseek(file,0,SEEK_END);

	save_pindex(pinfo,name);
}

void flush_pinfo_asc(FILE *file,char *name)
{
   fprintf(file,"end_%s\n",name);
}

void begin_pinfo(FILE *mat_file,FILE *asc_file,pinfos *pinfo,char *name,int flag)
{
   if (flag == 0) return;
	Pinfo->index_count = 0.0;
   if (mat_file != NULL) begin_pinfo_mat(mat_file,pinfo,name);
   if (asc_file != NULL) begin_pinfo_asc(asc_file,pinfo,name);
}


void flush_pinfo(FILE *mat_file,FILE *asc_file,pinfos *pinfo,char *name,int flag)
{
   if (flag == 0) return;
   if (mat_file != NULL) flush_pinfo_mat(pinfo,name);
   if (asc_file != NULL) flush_pinfo_asc(asc_file,name);
}

void write_1col(FILE *file,double value)
{
   fwrite(&value,sizeof(double),1,file);
}

void write_pinfo_mat(rels *rel)
/*
 *   Write out pinfo information and update counts.
 *   Note: we use Sign_dr here for "reverse" ray trace.
 *   Note: [t|b]b_adjusts are used to increment top and bot bonks by 1/2
 *   at bounce points.
 */
{
FILE *file = Pinfo->file;
double z = rel->z;
double a = -atan2( rel->s,rel->c);

   Find_speed_rel = rel;
   if (Pinfo->saveflg[0]) write_1col(file, Param.sign_dr * rel->r);
   if (Pinfo->saveflg[1]) write_1col(file, Param.sign_z  * z);
   if (Pinfo->saveflg[2]) write_1col(file, rel->t);
   if (Pinfo->saveflg[3]) write_1col(file, a);
   if (Pinfo->saveflg[4]) write_1col(file,find_speed(z));
   if (Pinfo->saveflg[5]) write_1col(file,find_grad(rel));
   if (Pinfo->saveflg[6]) write_1col(file,rel->l);
   if (Pinfo->saveflg[7]) write_1col(file,rel->l_c2);
   if (Pinfo->saveflg[8]) write_1col(file,rel->s / find_speed(z));
   if (Pinfo->saveflg[9]) write_1col(file,Rinfo1.top_bonks);
   if (Pinfo->saveflg[10]) write_1col(file,Rinfo1.bot_bonks);
   if (Pinfo->saveflg[11]) write_1col(file,Rinfo1.bot_bonks +
												       Rinfo1.top_bonks);	
   if (Pinfo->saveflg[12]) write_1col(file,find_utp(z,a,find_speed) );
   if (Pinfo->saveflg[13]) write_1col(file,find_ltp(z,a,find_speed) );
	if (Pinfo->saveflg[14]) write_1col(file,Rinfo1.bottom_loss);
   if (Pinfo->saveflg[15]) write_1col(file,Pinfo->ray_count);
   Pinfo->steps_taken = 0;
   (Pinfo->write_count)++;
}

void write_pinfo(rels *rel)
{
	Pinfo->index_count++;
   if ( Mat_file != NULL) write_pinfo_mat(rel);
   if ( Asc_file != NULL) write_pinfo_asc(rel);
}

void start_pinfo(rels *rel)
{
pindexs *pindex;
	
   Pinfo->ray_count++;
	Safe_calloc(pindex,pindexs,1);
	if (Pinfo->index == NULL)
		Pinfo->first_index = pindex;
	else
		Pinfo->index->next = pindex;
	Pinfo->index = pindex;
	Pinfo->index->starti = Pinfo->index_count+1;
   Pinfo->next_range = Pinfo->min_range;
   if ( Pinfo->next_range <= rel->r) {
      write_pinfo(rel);
      Pinfo->next_range += Pinfo->dr;
      if ( Pinfo->next_range > Pinfo->max_range)
         Pinfo->next_range = Pinfo->max_range;
   }
}

void end_pinfo()
{
   Pinfo->index->endi = Pinfo->index_count;
}

void rinfo_2_rel(rels *si,params *param,double angle)
/*
 *   Load *si with the starting information contained in *rinfo.
 *   should be used only once per ray.
 */
{
   si->z = Rinfo1.sdepth = param->sdepth;
   si->r = Rinfo1.srange = param->srange;
   Rinfo1.rrange         = param->rrange;
   Rinfo1.sangle         = angle;
   Rinfo1.c_at_s         = Param.source_c;
   si->s                 = -sin(angle);
   si->c                 =  cos(angle);
   si->t                 = 0.0;
   si->l                 = 0.0;
   si->l_c2              = 0.0;
   Rinfo1.top_bonks      = 0.0;
   Rinfo1.bot_bonks      = 0.0;
   Rinfo1.tot_bonks      = 0.0;
   Rinfo1.top_turns      = 0.0;
   Rinfo1.bot_turns      = 0.0;
   Rinfo1.tot_turns      = 0.0;
   Rinfo1.bottom_loss    = 0.0;
}

void reverse_z(void)
{
   Rinfo1.sdepth *= Param.sign_z;
   Rinfo1.rdepth *= Param.sign_z;
   Rinfo1.utp    *= Param.sign_z;
   Rinfo1.ltp    *= Param.sign_z;
}

void rel_2_rinfo(rels *so,params *param)
/*
 *   Takes the ray trace results in *so and puts them into *rinfo.
 *   Note we use Sign_dr here for "reverse" ray trace.
 *
 */
{
   Rinfo1.rangle = atan2(-so->s,so->c);
   Rinfo1.rdepth = so->z;
   Rinfo1.time   = so->t;
   Rinfo1.rrange = param->sign_dr * so->r;
   Rinfo1.srange   *= param->sign_dr;
   Rinfo1.tot_turns = Rinfo1.top_turns + Rinfo1.bot_turns;
   if (param->saveflg.turning_points) find_tps(&Rinfo1);
   Rinfo1.length = so->l;
   Rinfo1.length_c2 = so->l_c2;
   Rinfo1.c_at_r = rspeed(so->z);
   Rinfo1.p_z = so->s / Rinfo1.c_at_r;
   reverse_z();
}
/*************************************************************************/
/*************************************************************************/
/***************** BEGIN CRITICAL LOOP ***********************************/
/*************************************************************************/
/*************************************************************************/

void advance_btab(double r)
{
   if (B_tab == NULL) {
      printf("at range %g\n",r);
      fatal("End of bathymetry inside of bottom().");
   }
   while(B_tab->r < r) {
      B_tab = B_tab->next;
      if (B_tab == NULL) {
         printf("at range %g\n",r);
         fatal("End of bathymetry inside of bottom().");
      }
   }
}

double  bottom(double r,btabs *btab)
{
double dr;

   dr = r - btab->r;
   return( btab->z0 + dr * ( btab->z1 + dr * btab->z2) );
}

/****************************************************************************/
/*                     Integration routines                                 */
/****************************************************************************/

void  integrate_dr(double dr)
/*
 *  4th order Runge-Kutta integration over range.
 */
{
double dr2,dr6,dr1;
double cos_inv,se_inv;
double c_inv,dtheta;

   dr1 = dr;
   dr2 = dr * One_half;
   dr6 = dr * One_sixth;

   S_earth = 1.0 / ( se_inv = 1.0 - S1.z * R_earth_inv);
   cos_inv = 1.0 / S1.c;

   Derivative(&S1,&c_inv,&dtheta);

   S2.r = S1.r + dr2;
   S2.z = S1.z + dr2 * (Ds1.z =   S1.s * cos_inv          * se_inv);
   S2.s = S1.s + dr2 * (Ds1.s =   dtheta                  * se_inv);
   S2.c = S1.c + dr2 * (Ds1.c = - S1.s * dtheta * cos_inv * se_inv);
   Ds1.l_c2 = c_inv * (Ds1.t = c_inv * (Ds1.l = cos_inv * se_inv));

   Derivative(&S2,&c_inv,&dtheta);

   S_earth = 1.0 / (se_inv =  1.0 - S2.z * R_earth_inv);
   cos_inv = 1.0 / S2.c;

   S3.r = S1.r + dr2;
   S3.z = S1.z + dr2 * (Ds2.z =   S2.s * cos_inv         * se_inv);
   S3.s = S1.s + dr2 * (Ds2.s =   dtheta                 * se_inv);
   S3.c = S1.c + dr2 * (Ds2.c = - S2.s * dtheta *cos_inv * se_inv);
   Ds2.l_c2 = c_inv * (Ds2.t = c_inv * (Ds2.l = cos_inv * se_inv));

   Derivative(&S3,&c_inv,&dtheta);
   cos_inv = 1.0 / S3.c;

   S4.r = S1.r + dr1;
   S4.z = S1.z + dr1 * (Ds3.z =   S3.s * cos_inv          * se_inv);
   S4.s = S1.s + dr1 * (Ds3.s =   dtheta                  * se_inv);
   S4.c = S1.c + dr1 * (Ds3.c = - S3.s * dtheta * cos_inv * se_inv);
   Ds3.l_c2 = c_inv * (Ds3.t = c_inv * (Ds3.l = cos_inv * se_inv));

   Derivative(&S4,&c_inv,&dtheta);
   cos_inv = 1.0 / S4.c;

   Ds4.z =   S4.s * cos_inv          * se_inv;
   Ds4.s =   dtheta                  * se_inv;
   Ds4.c = - S4.s * dtheta * cos_inv * se_inv;
   Ds4.l_c2 = c_inv * (Ds4.t = c_inv * ( Ds4.l = cos_inv * se_inv));

   S2.r = S1.r + dr1;
   S2.z = S1.z + dr6 * ( Ds1.z + 2.0 * (Ds2.z + Ds3.z) + Ds4.z);
   S2.s = S1.s + dr6 * ( Ds1.s + 2.0 * (Ds2.s + Ds3.s) + Ds4.s);
   S2.c = S1.c + dr6 * ( Ds1.c + 2.0 * (Ds2.c + Ds3.c) + Ds4.c);
   S2.t = S1.t + dr6 * ( Ds1.t + 2.0 * (Ds2.t + Ds3.t) + Ds4.t);
   S2.l = S1.l + dr6 * ( Ds1.l + 2.0 * (Ds2.l + Ds3.l) + Ds4.l);
   S2.l_c2 = S1.l_c2 + dr6 * ( Ds1.l_c2 + 2.0 * (Ds2.l_c2 + Ds3.l_c2) +
                               Ds4.l_c2);

}

void integrate_dl(double dl)
/*
 *  4th order Runge-Kutta integration over path length.
 *
 */
{
double dl2,dl6,dl1;
double c_inv,dtheta;

   dl1 = dl;
   dl2 = dl * One_half;
   dl6 = dl * One_sixth;

   S_earth = 1.0 / ( 1.0 - S1.z * R_earth_inv);

   Derivative(&S1,&c_inv,&dtheta);

   S2.r = S1.r + dl2 * (S1.c) * S_earth;
   S2.z = S1.z + dl2 * (S1.s);
   S2.s = S1.s + dl2 * (Ds1.s =   S1.c * dtheta);
   S2.c = S1.c + dl2 * (Ds1.c = - S1.s * dtheta);
   Ds1.l_c2 = c_inv * ( Ds1.t = c_inv);

   Derivative(&S2,&c_inv,&dtheta);

   S_earth = 1.0 / ( 1.0 - S2.z * R_earth_inv);

   S3.r = S1.r + dl2 * (S2.c) * S_earth;
   S3.z = S1.z + dl2 * (S2.s);
   S3.s = S1.s + dl2 * (Ds2.s =   S2.c * dtheta);
   S3.c = S1.c + dl2 * (Ds2.c = - S2.s * dtheta);
   Ds2.l_c2 = c_inv * (Ds2.t = c_inv);

   Derivative(&S3,&c_inv,&dtheta);

   S4.r = S1.r + dl1 * (S3.c) * S_earth;
   S4.z = S1.z + dl1 * (S3.s);
   S4.s = S1.s + dl1 * (Ds3.s =   S3.c * dtheta);
   S4.c = S1.c + dl1 * (Ds3.c = - S3.s * dtheta);
   Ds3.l_c2 = c_inv * (Ds3.t = c_inv);

   Derivative(&S4,&c_inv,&dtheta);

   S2.l = S1.l + dl1;
   S2.r = S1.r + dl6 * (  S1.c + 2.0 * ( S2.c +  S3.c) + S4.c) * S_earth;
   S2.z = S1.z + dl6 * (  S1.s + 2.0 * ( S2.s +  S3.s) + S4.s);
   S2.s = S1.s + dl6 * ( Ds1.s + 2.0 * (Ds2.s + Ds3.s) + S4.c * dtheta);
   S2.c = S1.c + dl6 * ( Ds1.c + 2.0 * (Ds2.c + Ds3.c) - S4.s * dtheta);
   S2.t = S1.t + dl6 * ( Ds1.t + 2.0 * (Ds2.t + Ds3.t) + c_inv);
   S2.l_c2 = S1.l_c2 + dl6 * ( Ds1.l_c2 + 2.0 * (Ds2.l_c2 + Ds3.l_c2) +
                               c_inv * c_inv);
}

/****************************************************************************/
/*              end of Integration routines                                 */
/****************************************************************************/

/****************************************************************************/
/*                    Derivative routines                                   */
/****************************************************************************/

void  dds_nrd(rels *s,double *c_inv, double *dtheta)
/*
 *  This is the range independent integration subroutine.
 */
{
double dz,z,c1,c2dz;
c_tabs *tab;

   z      = s->z;
   tab    = Tab1 + Idex[(int) z + Idex[0]];
   dz     = z - tab->z0;
   c1     = tab->c1;
   c2dz   = tab->c2 * dz;

   *c_inv  = 1.0 / (tab->c0 + dz * ( c1 + c2dz ));
   *dtheta = - s->c * ( R_earth_inv * S_earth + (c1 + c2dz + c2dz) * *c_inv);
}

void  dds_zrd(rels *s,double *c_inv, double *dtheta)
/*
 *   This integration subroutine gives linear range-dependent c and dc/dz
 *   but ignores dc/dr.
 */
{
int     iiz;
double  z,lambda;
c_tabs  *tab1,*tab2;
double  speed1,speed2;
double  dz1,dz2,t1c2dz,t2c2dz;
double  grad1,grad2;

   z      = s->z;
   iiz    = Idex[(int) z + Idex[0]];
   tab1   = Tab1 + iiz;
   tab2   = Tab2 + iiz;

   dz1     = z - tab1->z0;
   speed1 = tab1->c0 + dz1 * ( tab1->c1 + (t1c2dz = tab1->c2 * dz1 ));
   grad1  = tab1->c1 + t1c2dz + t1c2dz;
   
   dz2     = z - tab2->z0;   
   speed2 = tab2->c0 + dz2 * ( tab2->c1 + (t2c2dz = tab2->c2 * dz2 ));
   grad2  = tab2->c1 + t2c2dz + t2c2dz;

   lambda = (Range1 - s->r) * R12_inv;   
   *c_inv  = 1.0 / (speed1 + lambda * (speed2 - speed1));   
   *dtheta = -s->c * (R_earth_inv * S_earth + *c_inv * \
          (grad1 + lambda * (grad2 - grad1)));
}

void  dds_frd(rels *s,double *c_inv, double *dtheta)
/*
 *   This integration subroutine gives linear dependent c and dc/dz
 *   and  includes dc/dr.
 */
{
int     iiz;
double  z,lambda;
c_tabs  *tab1,*tab2;
double  speed1,speed2,delta_c;
double  dz1,dz2,t1c2dz,t2c2dz;
double  grad1,grad2;

   z      = s->z;
   iiz    = Idex[(int) z + Idex[0]];
   tab1   = Tab1 + iiz;
   tab2   = Tab2 + iiz;

   dz1     = z - tab1->z0;
   dz2     = z - tab2->z0;   
   lambda = (Range1 - s->r) * R12_inv;
   speed1 = tab1->c0 + dz1 * ( tab1->c1 + (t1c2dz = tab1->c2 * dz1 ));
   grad1  = tab1->c1 + t1c2dz + t1c2dz;
   speed2 = tab2->c0 + dz2 * ( tab2->c1 + (t2c2dz = tab2->c2 * dz2 ));
   grad2  = tab2->c1 + t2c2dz + t2c2dz;

   *c_inv  = 1.0 / (speed1 + lambda * (delta_c = speed2 - speed1));
   *dtheta = - s->c * (R_earth_inv * S_earth  \
             + *c_inv * (grad1 + lambda * (grad2 - grad1))) \
             + *c_inv * s->s * delta_c *R12_inv;
}

/****************************************************************************/
/*               end of Derivative routines                                 */
/****************************************************************************/

void take_step(double newz)
/*
 *   This routine will be turned into a macro someday.
 *   This contains all of the boundary checking for each step.
 */
{
int bonks = 0;

   if (S1.c < Param.min_cos || S2.c < Param.min_cos) {
      ldbug("max angle exceeded");
      longjmp(Abort_buf,Abort_normal);
   };
   advance_btab(S1.r);
   if (newz < Param.top)
      bonks += maybe_bounce(Param.top,  -1.0);
   else if (newz > B_tab->minz)
      bonks += maybe_reflect();
   if ( bonks == 0)
      if (Saving_pinfo) advance_pinfo();
   count_turns();
}

double fast_dl(void)
{
double dz;
rstep_tabs *rstep;
double z = S1.z;

   rstep = Rstep_tab + Idex[(int) z + Idex[0]];
   dz = z - rstep->z0;
   return( (S1.c * Cos_a + Cos_b) * ( rstep->dr0 + dz * rstep->dr1) );
}


void  advance_pinfo(void)
/*
 *   Assumes a valid step was taken from S1 to S2.
 *   If at least one pinfo range occurs between S1 and S2
 *   we save S1 into S3 and S2 into S4
 *   and march S1 -> S2 to all intervening
 *   pinfo ranges saving pinfo as we go.  S2 is restored to
 *   its original value.
 *   We ASSUME no bouncing occurs between S1 and S2 so
 *   we don't worry about btabs or maybe_ ... here.
 */
{
double final_range;
rels temp_s1,temp_s2;

   if ( S2.r < Pinfo->next_range ) return;
   final_range = S2.r;
   Memcopy(&temp_s1,&S1,sizeof(rels));           /* save S1 */
   Memcopy(&temp_s2,&S2,sizeof(rels));           /* save S2 */

   while( Pinfo->next_range <= final_range) {
      if (S1.r > Pinfo->next_range) {
         ldbug("S1.r > Pinfo->next_range in \"advance_pinfo()\"");
         fprintf(Log_file,"Pinfo->next_range = %g\n",Pinfo->next_range);
         longjmp(Abort_buf,Abort_error);
      }
      integrate_dr(Pinfo->next_range - S1.r);
      write_pinfo(&S2);
      if ( Pinfo->next_range == Pinfo->max_range) {
         Saving_pinfo = 0;
    break;
      }
      Pinfo->next_range += Pinfo->dr;
      if ( Pinfo->next_range >= Pinfo->max_range) {
         Pinfo->next_range = Pinfo->max_range;
      }
   }
   Memcopy(&S1,&temp_s1,sizeof(rels));           /* restore S1 */
   Memcopy(&S2,&temp_s2,sizeof(rels));           /* restore S2 */
}

void  shoot_region(double range)
/*
 *       shoots one ray from S1 to range.
 *       This is the start of the CRITICAL LOOP.
 */
{
double dr,dl;
double targ_r;

   targ_r = range;
   for (;;) {
      dl = fast_dl();
      integrate_dl(dl);
      if (S2.r  > targ_r ) break;
      take_step(S1.z + 2.0 * dl * S1.s);
      Memcopy(&S1,&S2,sizeof(rels));
   }
   integrate_dr(dr = range - S1.r);
   take_step(S1.z + 2.0 * dr * S1.s / S1.c);
   Memcopy(&S1,&S2,sizeof(rels));
}

tabtabs *shoot_one_lrd(void)
/*
 * This now only shoots a ray from the source to the receiver.
 * We use repeated calls to shoot_region to shoot a ray through
 * a linearly interpolated pair of profiles.
 * fiddle with table pointers along the way to make everything
 * work out.
 */
{
double  range;
long    bounces;
tabtabs *tabtab;

   range = Rinfo1.rrange;
   B_tab = First_btab;
   tabtab = Tabtab;
   Loss   = Environ->loss;
   Tab1   = tabtab->tab1;
   Tab2   = tabtab->tab2;

   Saving_pinfo = Param.saving_pinfo;
   if (Saving_pinfo) start_pinfo(&S1);

   for( ;tabtab != NULL; tabtab = tabtab->next) {
      Tab1   = tabtab->tab1;
      Tab2   = tabtab->tab2;
      Range1 = tabtab->r1;
      Range2 = tabtab->r2;
      R12_inv = tabtab->r12_inv;
      if ( range <= Range2 ) break;
      bounces = (int) Rinfo1.tot_bonks;
      shoot_region(Range2);
      vprint_dot(bounces);
   }
   bounces = (int) Rinfo1.tot_bonks;
   shoot_region(range);
   vprint_dot(bounces);
	if (Param.saving_pinfo) end_pinfo();
   return(tabtab);
}

/*************************************************************************/
/*************************************************************************/
/***************** END OF CRITICAL LOOP **********************************/
/*************************************************************************/
/*************************************************************************/

void shoot_one(params *param,double angle,long ray_number)
{
int abort_flag;

   rinfo_2_rel(&S1,param,angle);
   if ( (abort_flag = setjmp(Abort_buf)) == 0) {
      shoot_one_lrd();
   }
   else {
		process_exception(abort_flag,ray_number);
		if( param->saving_pinfo) end_pinfo();
	}
   rel_2_rinfo(&S1,param);
}

time_t start_timing(long n)
{
time_t start_time;
char s[100];

   sprintf(s,"Shooting %ld ray%s.\n",n, (n==1) ? "" : "s");
   if ( n > 0) vlputs(s);
   time(&start_time);
   return(start_time);
}

void end_timing(long n,time_t start_time)
{
time_t    end_time;
double    elapsed;
char     *time_str;
char      temp[120];
double    megameters = fabs(Param.srange - Param.rrange) * 1.0e-6;

   time(&end_time);
   elapsed = (end_time - start_time);
   if ( elapsed < 1) elapsed = 1;
   time_str = s2dhms(elapsed);
   sprintf(temp,"%ld rays took %s. %.2g rays/sec. %.2g ray-Mm/sec\n",
         n,time_str,n/elapsed,n/elapsed * megameters);
   vlputs(temp);
}

rinfos *shoot_one_ray(params *param,double angle,long cnt,long num_a)
/*
 * shoots a single ray and fills an rinfo structure with the
 * results.  
 */
{
rinfos *rinfo;
int dots;

   if (Verbose) {
      if (cnt != 0) {
         if ( num_a != 0 )
            printf("%6ld %8.4f %6.1f ",cnt,angle*180.0/Pi,cnt*100.00/num_a);
         else
            printf("%6ld %8.4f        ", cnt, angle*180.0/Pi);          
         if (Screen) {
            for(dots = 0; dots <= Num_dots; dots++) putchar(' ');
            for(dots = 0; dots <= Num_dots; dots++) putchar('\b');
            fflush(stdout);
         }
         else putchar('[');
      }
      Num_dots = 0;
   }
   shoot_one(param,angle,cnt);
   if (Verbose) {
      if (!Screen) putchar(']');
      putchar( (Screen) ? '\r' : '\n');
      fflush(stdout);
   }
   Safe_calloc(rinfo,rinfos,1);
   Memcopy(rinfo,&Rinfo1,sizeof(rinfos));
   return(rinfo);
}

rinfos *shoot_all_angles(params *param)
/*
 *   This routine shoots rays on angles contained in Params.angles.
 *   It use the routine shoot_one() to shoot one ray.  We prepare to run
 *   shoot_one() by loading *rinfo and *s with the starting info for the
 *   ray.  Shoot() will fill *s with the final values for this ray which
 *   we then put into *rinfo.  Along the way shoot also recorded bounces and
 *   turns in *rinfo.
 *
 */
{
double angle;
rinfos *rinfo = NULL,*first=NULL,*last = NULL;
dubs   *dub;
int cnt = 0,num_a = param->num_angles;

   if (Verbose) printf("%6s %6s %6s\n","ray","angle","   %");
   for(dub = param->angles;dub != NULL;dub = dub->next) {
      angle = dub->d;
      rinfo = shoot_one_ray(param,angle,++cnt,num_a);
      merge_1rinfo(&first,last,rinfo);
      last  = rinfo;
   }
   vprint("\n");
   return(first);
}

rinfos *shoot(pinfos *pinfo,char *p_name,int path_flag)
{
rinfos *rinfo;
time_t start_time;

   Pinfo = pinfo;
   start_time = start_timing(Param.num_angles);
   begin_pinfo(Mat_file,Asc_file,Pinfo,p_name,path_flag);
   Param.saving_pinfo = path_flag;
   rinfo = shoot_all_angles(&Param);
   flush_pinfo(Mat_file,Asc_file,Pinfo,p_name,path_flag);
   end_timing(Param.num_angles,start_time);
   return(rinfo);
}

void do_add_extrema(params *param,rinfos **rinfo)
{
long count;

   if(Verbose) printf("Adding extrema to wavefront.\n");
   fprintf(Log_file,"Adding extrema to wavefront.\n");
   count = add_extrema(param,rinfo);
   if (Verbose) printf("%ld rays were added to the wavefront.\n",count);
   fprintf(Log_file,"%ld rays were added to the wavefront.\n",count);
}

/****************************************************************************/
/*                      Main routine and misc                               */
/****************************************************************************/

void select_angles(params *param)
/*
 * Creates a linked list of angles if param->angles == NULL,
 * otherwise counts the number of angles that exist in param->angles.
 */
{
double  da;
dubs    *dub, *last=NULL;
int     i, num_a=0;

   if (param->angles != NULL)
   {                            /* count the input angles */
      for(dub = param->angles;dub != NULL;dub=dub->next) num_a++;
      param->num_angles = num_a;
      return;
   }
   else
   {
      num_a = param->num_angles;
      da =  (num_a > 1) ? (param->angle2 - param->angle1)/(num_a-1) : 0.0;
      for(i=1;i <= num_a;i++) {
         Safe_calloc(dub,dubs,1);
         dub->d = param->angle1 + (i-1) * da;
         if (last == NULL) param->angles = dub;
         else last->next = dub;
         last = dub;
      }
   }
}

derivatives select_derivative(params *param)
{
int which = param->range_depend;

   if (which < 0 || which >= Num_els(Derivative_tab)) fatal("range_dependence out of range");
   return(Derivative_tab[which]);
}

void select_routines(params *param)
/*
 *  Figures out what combination of routines should be used to
 *  do the ray integration.
 *  also set the global variable R_earth_inv
 */
{
   Derivative = select_derivative(param);
   R_earth_inv = (param->earth_radius > 0.0) \
                       ? 1.0 / param->earth_radius : 0.0;
   Cos_a = (param->step.cos_factor  == 0.0) \
                            ? 0.0 : 1.0 / param->step.cos_factor;
   Cos_b = 1.0  - Cos_a;
}

/****************************************************************************/
/*                       Main routine                                       */
/****************************************************************************/

int main(int argc, char *argv[])
{
rinfos          *rinfo;

   Prog_name = "Ray";
   Prog_date = "Tuesday April 25, 1994";
   Version   = 1.47;

   initialize_ray(argc, argv);
   Environ = read_inputs(Prof_file,Bath_file,Loss_file);
   pre_process(&Tabtab,&Rtab,&Idex,&Rstep_tab,&First_btab,\
          Environ,&Param,&Fan,&Eigenfan);

   if (Param.saveflg.sound_speeds)
      save_speeds(Mat_file,Tabtab,Idex,Rstep_tab,Rtab,&Param,Environ);
   if (Param.saveflg.bathymetry)
      save_bathymetry(Mat_file,First_btab,Param.vname.btab);

   if (Param.saveflg.e_only) {
      lputs("No rays were computed.");
   }
   else {
      select_angles(&Param);
      select_routines(&Param);
      rinfo = shoot(&Fan,Param.vname.fan,Param.saveflg.fan);
      if ( Param.add_extrema.max_iter > 0) do_add_extrema(&Param,&rinfo);
      add_all_geo_amps(&Param,rinfo);
      if (Param.saveflg.wavefront)
           save_rinfo(Mat_file, Asc_file,rinfo,Param.vname.wavefront);
      Param.saving_pinfo = 0;
      if (Param.saveflg.eigenrays)  find_eigenrays(&Param,&rinfo);
   }
   fprintf(Log_file,"Done.\n");
   if (Asc_file != NULL) fclose(Asc_file);
   if (Mat_file != NULL) fclose(Mat_file);
   if (Log_file != NULL) fclose(Log_file);
   printf("Done.\n");
   return(0);
}

/****************************************************************************/
/*                      End of ray.c                                        */
/****************************************************************************/


