/****************************************************************************/
/*                 program reada.c                                          */
/****************************************************************************/
/*
 *     Copyright (C) 1993,  Woods Hole Oceanographic Institution
 *     See LICENSE file for use, copying and distribution conditions
 *
 *     Jim Bowlin jbowlin@whoi.edu                May 10, 1993
 *
 *     Implements several function based around the iofile structure.
 *     This structure is a superset of a FILE. See reada.h for a full
 *     description of all of the structure elements.
 *
 *  iofiles *iof_open(char *name,int line_len,char *format)
 *
 *     creates an iofile just like fopen() would create a FILE.
 *     you need to specify a line length, (256 is a good value to use).
 *
 *  void iof_close(iofiles *iof)
 *
 *     Reports the number of errors recorded in an iofile and then
 *     closes it.
 *
 *     Other fuctions are available for error reporting and reading
 *     strings, numbers, etc. See reada.h for a listing of available
 *     functions.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include "utils.h"
#include "reada.h"


static    commands  *First_command,*Last_command;
static    int       Ncommands;

void raw_iof_err(iofiles *iof,long linecnt,int charcnt,char *s);

void iof_puts(iofiles *iof,char *s)
{
   fputs(s,stdout);
   if (iof->log_file == NULL) return;
   fputs(s,iof->log_file);
}

void tok_puts(iofiles *iof,char *s)
{
   iof_puts(iof, "<");
   iof_puts(iof,s);
   iof_puts(iof,">");
}

/****************************************************************************/
/*                Open and Close iofiles                                    */
/****************************************************************************/

char *raw_nickname(char *name)
{
char *nick;
int  i,first,last,end,size;

   end = strlen(name);
   for(i=0,first=0;i<end;i++)
      if (name[i] == '\\' || name[i] == '/' ) first = i+1;
   for (last = first;last<end;last++)
      if (name[last] == '.') break;
   size = last - first;
   Safe_malloc(nick,char,size+1);
   Memcopy(nick,name+first,size);
   nick[size] = 0;
   return(nick);
}

char *raw_extension(char *name)
{
char *ext;
int  i,first,last,end,size;

   end = strlen(name);
   for(i=0,first=0; i < end;i++)
      if (name[i] == '\\' || name[i] == '/' ) first = i+1;

   last = end;
   for (i = first;i < end;i++)
      if (name[i] == '.') last = i;
   size = end - last;
   if (size == 0) return(NULL);
   Safe_malloc(ext,char,size+1);
   Memcopy(ext,name+last,size);
   ext[size] = 0;
   return(ext);
}

void default_extension(char **name,char *ext)
/*
 * if **name has no "." in   it we change *name to point to a
 * copy of itself with ext appended on the end.
 */
{
   if ( name == NULL) return;
   if (*name == NULL) return;
   if (  ext == NULL) return;
   if (strchr(*name,'.') != NULL) return;
   *name = Strcpycat(*name,ext);
}


void make_nickname(iofiles *iof)
/*
 * strips off extension and path from name and copies whats left into
 * nickname.
 */
{
   iof->nickname  = raw_nickname(iof->name);
   iof->extension = raw_extension(iof->name);
}

void raw_open_iofile(iofiles *iof,int linelen,char *format)
{
   if (linelen > 0) {
      Safe_malloc(iof->tok,char,linelen);
      Safe_malloc(iof->string,char,linelen);
      Safe_malloc(iof->line,char,linelen);
      Safe_malloc(iof->err_buf,char,256);
      iof->line[0] = 0;
   }
   iof->prog_name = (Prog_name != NULL) ? Prog_name: "Prog_name";
   iof->log_file  = Log_file;
   iof->max_len   = linelen;
   iof->file      = fopen(iof->name,format);
   make_nickname(iof);
   return;
}

iofiles *open_iofile(char *name,int linelen,char *format)
{
iofiles *iof;

   Safe_calloc(iof,iofiles,1);
   iof->dynamic = 1;
   iof->name = Strdup(name);
   raw_open_iofile(iof,linelen,format);
   if(iof->file == NULL) return(NULL);
   return(iof);
}

int close_iofile(iofiles *iof)
{
   if (iof == NULL) return(-1);
   fclose(iof->file);
   if (iof->max_len > 0) {
      free(iof->line);
      free(iof->string);
      free(iof->err_buf);
      free(iof->tok);
   }
   if (iof->err_cnt != 0) {
      sprintf(iof->err_buf,"\n%s: %d errors reading %s\n",iof->prog_name,iof->err_cnt,iof->name);
      iof_puts(iof,iof->err_buf);
      exit(-1);
   }
   if (iof->dynamic == 1) free(iof);
   return(0);
}

/****************************************************************************/
/*                end of Open and Close iofiles                             */
/****************************************************************************/

/***************************************************************************/
/*                    Reading Lines                                        */
/***************************************************************************/

char *get_line(iofiles *iof)
/*
 * reads in just one line from file and counts lines
 */
{

   if (iof->end_of_file ==0) (iof->linecnt)++;
   if ( fgets(iof->line,iof->max_len,iof->file) == NULL){
      if ((iof->end_of_file)++ >10 ) {
         iof_puts(iof,"Fatal error. 10 reads past end of file\n");
         exit(2);
      }
      return(NULL);
   }
   iof->start = iof->end = 0;
   if (iof->debug ) iof_puts(iof,"\n");
   return(iof->line);
}

/*
char *get_a_line(iofiles *iof)
{
char *s;
   s = get_line(iof);
   iof->end = strlen(s);
   return(s);
}
*/
/***************************************************************************/
/*                    end of Reading Lines                                 */
/***************************************************************************/


/***************************************************************************/
/*                        Token Parsing                                    */
/***************************************************************************/


void debug_iof(iofiles *iof)
{
   if (!iof->debug) return;
   tok_puts(iof,iof->tok);
   if (iof->tok[0] != '\"') return;
   tok_puts(iof,iof->string);
}

int find_end_comment(iofiles *iof,int start)
/*
 *  reads lines from iof->file until end-of-comment () or end-of-file
 *  is found. End-of-file results in a fatal error.  End-of-comment
 *  results in our returning the offset just past the comment.
 */
{
char c;
char *line;
long  linecnt;
int   comment_nest;

   line = iof->line;
   linecnt = iof->linecnt;
   comment_nest = 0;
   for(;;){
      c = iof->line[start];
      switch (c) {
        case '*':
           if (line[++start] == '/') {
              if (--comment_nest < 0) {
                 iof->end = start+1;
                 return(start+1);
              }
              start++;
           }
           break;
       case '/':
           if (line[++start] == '*') {
              comment_nest++;
              start ++;
           }
           break;
        case '\0':
        case '\n':
           if ( (line = get_line(iof)) == NULL) {
              sprintf(iof->err_buf,
							 "end-of-file in comment started on line %ld.",linecnt);
				  	raw_iof_err(iof,iof->linecnt,iof->start,NULL);
              exit(-2);
           }
           start = 0;
           break;
        default:
           start++;
      }
   }
}

void find_matching(iofiles *iof,char opening,char closing)
/*
 * Searches through iof for a matching closing.
 */
{
int   count = 1;
char *tok;

   do {
		tok = get_token(iof);
		if (*tok == opening)
			count++;
		else if ( *tok == closing)
			count--;
	} while ( count > 0);
}

void add_command(char *s)
/*
 * adds *s to our linked list of commands.
 * Increments the count of the number of commands.
 */
{
commands *command;

   Safe_calloc(command,commands,1);
   command->command = Strdup(s);
   command->number  = Ncommands;
   if (Ncommands++ == 0) First_command = command;
   else Last_command->next = command;
   Last_command = command;
}

int get_command(iofiles *iof,char *dest, int start)
{
int   end,n,len;
char *line,c,*sub;
char  cmnd[80];
commands *command;
char *end_num;

   line = iof->line;
   iof->start = start;
   start++;                   /* now points to char past '$' */
   for(end = start; ;end++) {
      c = line[end];
      iof->end = end;
      if (c == '$') break;
      if (c == '\0') {
         iof_err(iof,"Missing trailing '$' in command substitution");
         return(0);
      }
   }
   iof->end = end+1;
   len = end - start;
   if (len ==0) return(len);
   Memcopy(cmnd,line + start,len);
   cmnd[len] = 0;
   n = strtol(cmnd,&end_num,10);
   if (end_num[0] != 0) {
      iof_err(iof,"Expected a number between \"$\" and \"$\".");
      return(len);
   }

   if ( n < 0 || n >= Ncommands ) {
      sprintf(iof->err_buf,"Command line argument number %d not found",n);
      iof_err(iof,NULL);
      return(len);
   }
   for(command = First_command; command != NULL; command = command->next)
      if (command->number == n) break;

   sub = command->command;
	command->invoked++;
   strcpy(dest,sub);
   return(strlen(sub));
}

int  check_commands()
/*   Goes through the linked list of commands and makes sure
 *   that each one has been envoked at least once.  If any
 *   have not been used we give an error msg and return non-zero.
 */
{
commands *cmnd;
int      errors = 0;	
	for(cmnd = First_command->next; cmnd != NULL; cmnd = cmnd->next) {
      if (cmnd->invoked != 0) continue;
		printf("Warning: command line parameter %d <%s> not used.\n",
				 cmnd->number,cmnd->command);
		fprintf(Log_file,"Warning: command line parameter %d <%s> not used.\n",
				  cmnd->number,cmnd->command);
		errors++;
	}
	return(errors);
}
					
char get_c(iofiles *iof)
{
  return( (iof->line)[(iof->end++)] );
}

void unget_c(iofiles *iof)
{
   iof->end--;
}

char get_escape(iofiles *iof,char termchar)
/*
 *  reads char from iof to make conversion following the
 *  backslash character.
 */
{
char c;
char temp[4];
int i;

   c = get_c(iof);
   if (c == 0) {
      iof_err(iof,"string and char constants cannot end in \"\\\"");
      return(c);
   }

   /*** process octal numbers ***/

   if ( isdigit(c) ) {
      temp[0] = c;
      for (i = 1; i< 3; i++) {
         if ((temp[i] = get_c(iof) ) == termchar) {
            unget_c(iof);
            break;
         }
      }
      temp[i] = 0;
      c = strtol(temp,NULL,8);
      return(c);
   }

   /*** process hex numbers ***/

   if( c == 'x' )
   {
      for (i = 0; i< 2; i++) {
         if ((temp[i] = get_c(iof) ) == termchar) {
            unget_c(iof);
            break;
         }
      }
      temp[i] = 0;
      c = strtol(temp,NULL,16);
      return(c);
   }
   switch (c)
   {
      case 'a' : c = '\a'; break;
      case 'b' : c = '\b'; break;
      case 'f' : c = '\f'; break;
      case 'n' : c = '\n'; break;
      case 'r' : c = '\r'; break;
      case 't' : c = '\t'; break;
      case 'v' : c = '\v'; break;
      case '\\': c = '\\'; break;
      case '\?': c = '\?'; break;
      case '\'': c = '\''; break;
      case '\"': c = '\"'; break;
      case '$' : c = '$' ; break;
      default  : iof_err(iof,"invalid escape sequence");
   }
   return(c);
}

int read_iof_str(iofiles *iof,char termchar,int start)
/*
 *  read chars from token stream translating /chars until a naked
 *  termchar is found.
 */
{
int  len;
char c;
char *dest = iof->string;

   strcpy(iof->tok,"\"");
   iof->end = iof->start = start+1;
   for(len = 0;len < iof->max_len;) {
      c = get_c(iof);
      if (c == termchar) {
         dest[len] = 0;
         return(0);
      }
      switch (c) {
         case '\n' :
         case '\0' :
            iof_err(iof,"end of line encountered inside a string");
            return(-1);
         case '\\' :
            dest[len++] = get_escape(iof,termchar);
            break;
         case '$' :
            len += get_command(iof,dest+len,iof->end-1);
            break;
         default  :
            dest[len++] = c;
      }
   }
   iof_err(iof,"string constant too long");
   return(-1);
}

char *raw_get_token(iofiles *iof) {

char c1,c2,endc;
char *line;
int   start,end;

   if (iof->unget) {
      iof->unget = 0;                /* turn off unget flag          */
      return(iof->tok);              /* and just return the last one */
   }
   for(;;) {
      start = iof->end;
      line = iof->line;
      while(isspace(line[start])) start++;   /* line[start] is now non-white */

      /*** check for end of line ***/

      c1 = line[start];
      if (c1 == '\0') {
         if (get_line(iof) != NULL) return("\n");
         strcpy(iof->tok,End_of_file);
         debug_iof(iof);
         return(iof->tok);
      }
      if (strchr(DELIMITS,c1) == NULL) {
         /*** read in a normal token ***/
         for(end = start;line[end] != '\0';end++) {
            endc = line[end];
            if (isspace(endc) || (strchr(DELIMITS,endc)!=NULL)) break;
         }
         break;
      }

      /*** do command line substitution ***/

      if ( c1 == '$') {
         get_command(iof,iof->tok,start);
         debug_iof(iof);
         return(iof->tok);
      }
      /***  process a string            ***/

      if ( c1 == '\"' ) {
         read_iof_str(iof,'\"',start);
         debug_iof(iof);
         return(iof->tok);
      }

      end = start + 1;
      c2 = line[start+1];
      if ( c1 == '/' && c2 == '*') {
         find_end_comment(iof,start+2);
         continue;
      }
      if ( c1 == '/' && c2 =='/') {
         start = iof->end = strlen(line);
			continue;
      }	      		
      break;
   }
   iof->tok[0] =0;

	/* copy the token into iof->tok */
	
   strncat(iof->tok,iof->line+start,end-start); 
   iof->start = start;
   iof->end  = end;
   debug_iof(iof);
   return(iof->tok);
}

int unget_token(iofiles *iof) {

   if ((iof->unget)++ >= 1) {
      iof_err(iof,"Internal double unget token error");
      return(1);
   }
   return(0);
}

char *get_token(iofiles *iof)
{
char *tok;

   do {
      tok = raw_get_token(iof);
   }  while( tok[0] == '\n');
   return(tok);
}


/***************************************************************************/
/*                   end of Token Parsing                                  */
/***************************************************************************/

/**************************************************************************/
/*                 basic conversion utilities                             */
/**************************************************************************/




long get_a_long(iofiles *iof)
{
char *tok,*end;
long i;

   tok = get_token(iof);
   i = strtol(tok,&end,10);
   if (end[0] == 0) return(i);
   iof_err(iof,"expected an integer value here.");
   return(0);
}


double get_a_dub(iofiles *iof)
{
char *tok;
double  d;

   tok = get_token(iof);
   d = atof(tok);
   return(d);
   /******                      does not work on SUN
   d = Strtod(tok,&end);
   if (end[0] == 0) return(d);
   iof_err(iof);
   printf("expected a numerical value here.");
   return(0.0);
   *******/
}

int get_an_int(iofiles *iof)
{
char *tok,*end;
int  i;

   tok = get_token(iof);
   i = strtol(tok,&end,10);
   if (end[0] == 0) return(i);
   iof_err(iof,"expected an integer value here.");
   return(0);
}

char *get_a_string(iofiles *iof)
{
char *tok = get_token(iof);

   if (tok[0] != '\"') {
      iof_err(iof,"Expected a ==> \" <== here.");
      return(NULL);
   }
   return(iof->string);
}

void eof_err(iofiles *iof)
{
   (iof->err_cnt)++;
   printf("\nError %s  (%d)\nUnexpected End-of-File ",iof->name,iof->err_cnt);
}

void raw_iof_err(iofiles *iof,long linecnt,int charcnt,char *s)
{
char temp[80];
int i;
char *line = iof->line;

   (iof->err_cnt)++;
   if (s ==NULL) s = iof->err_buf;
   sprintf(temp,"Error %s %ld:%d (%d) ",iof->name,linecnt,charcnt+1,iof->err_cnt);
   iof_puts(iof,temp);
   iof_puts(iof,s);
   iof_puts(iof,"\n");
   if (iof->line_echo_off != 0) return;
   iof_puts(iof,line);
   if(line[strlen(line)-1] != '\n') iof_puts(iof,"\n");
   for(i=0;i<charcnt;i++) iof_puts(iof,"-");
   iof_puts(iof,"^\n");

}

void skip_past_eos(char *tok,iofiles *iof)
{
long linecnt = iof->linecnt;

   while(tok[0] != ';' && linecnt == iof->linecnt) tok = get_token(iof);
	if (linecnt != iof->linecnt) unget_token(iof);
}

int want_tok(iofiles *iof,char *s)
{
char *tok;

   if (s[0] == ';') {
      tok = raw_get_token(iof);
      if (tok[0] == '\n' || tok[0] == ';') return(1);
   }
   else {
      tok = get_token(iof);
      if (strcmp(s,tok) == 0) return(1);
   }
   sprintf(iof->err_buf,"Expected \"%s\" here.\n",s);
   iof_err(iof,NULL);
	return(0);
}

char *get_token_eos(iofiles *iof)
{
char *tok;

    do tok = get_token(iof);
    while ( tok[0] == ';'  && iof->end_of_file == 0);
    return(tok);
}
/**************************************************************************/
/*                end of basic conversion utilities                       */
/**************************************************************************/

/**************************************************************************/
/*                         error handling                                 */
/**************************************************************************/

void iof_err(iofiles *iof,char *s)
{
	raw_iof_err(iof,iof->linecnt,iof->start,s);
	if(iof->end_of_file == 0) return;
   raw_iof_err(iof,iof->linecnt,0,"Unexpected end of file.");
}

/**************************************************************************/
/*                      end of error handling                             */
/**************************************************************************/


void test_reada(void ){
iofiles *iof;
char *token;

   iof = open_iofile("dum.mpp",256,"rt");
   while (iof->end_of_file == 0) {
      token = get_token(iof);
      tok_puts(iof,token);
   }
}

/****************************************************************************/
/*                    end of reada.c                                        */
/****************************************************************************/

