/****************************************************************************/
/*                 program readray.c                                        */
/****************************************************************************/
/*
 *    Copyright (C) 1992,  Woods Hole Oceanographic Institution
 *    See LICENSE file for use, copying and distribution conditions
 *
 *    Jim Bowlin March 4 1994.
 *
 *    Contains the code to read in bathymetry
 *                                 soundspeed profiles
 *                                 bottom loss
 *    from ascii files.
 *    Entry is via
 *    environs *read_inputs(FILE *prof_file,FILE *bath_file)
 *    which reads in the inputs and puts all data in an environs
 *    structure.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include "utils.h"
#include "reada.h"
#include "init.h"
#include "ray.h"
#include "readray.h"

#if !defined(__MSDOS__)
#include <unistd.h>
#include <memory.h>
#endif

#define    Line_buf_len  1000
char       Line_buf[Line_buf_len];



int read_1line(FILE *file,double value[],int n)
/*
 *  reads a line from a file and then reads the first n numbers
 *  from the line into the array value[].  value[0-(n-1)] are
 *  initialized to 0.0.  We return as soon non-numeric text is
 *  encountered.
 */
{
char *start,*end;
int  i;

   for(i=0; i < n; i++) value[i] = 0.0;
   if ((start = fgets(Line_buf,Line_buf_len,file) ) == NULL) return(0);
   end = start;
   for (i=0;i<n;i++)
   {
      while( isspace(start[0])) start++;
      if ( start[0]==0 ) return(i);
      value[i] = Strtod(start,&end);
      if (end == NULL || isalpha(end[0]) ) return(i);
      start = end;
   }
   return(n);
}

int read_doubles(FILE *file,double value[],int n,long *line)
{
int cols =0;

   while( cols == 0  && !feof(file) ) {
      cols = read_1line(file,value,n);
      (*line)++;
   }
   return(cols);
}

void format_msg(void)
{
   mputs("It appears that you input a prof_file that is not in the new format.\n");
   mputs("The new prof_file format is\n\n");
   mputs("\tneg       r_1 \t r_2 \t r_3 ...  r_n \n");
   mputs("\tz_1       c_11 \tc_12 \tc_13 ... c_1n \n");
   mputs("\tz_2       c_21 \tc_22 \tc_23 ... c_2n \n\n");
   mputs("\t          .  .  .                     \n\n");
   mputs("\tz_m       c_m1 \tc_m2 \tc_m3 ... c_mn \n\n");
   mputs("Where:\n");
   mputs("\t neg\tis any negative number.\n");
   mputs("\t r_j\tis the range of the j'th profile (km).\n");
   mputs("\t z_i\tis the depth of the i'th point (m).\n");
   mputs("\tc_ij\tis the sound speed at range r_j and depth z_i (m/s).\n\n");
   mputs("This block structure can be repeated indefinitely.\n\n");
}


char *unit_name(units *unit)
/*
 *   returns a string containing the name of the units
 *   specified in unit.
 */
{
intabs *intab = (intabs*) unit->table;

   return( intab[unit->offset].name);
}

double unit_value(units *unit)
/*
 *   returns conversion value for specified unit.
 */
{
intabs *intab = (intabs*) unit->table;

   return(*(double*) (intab[unit->offset].where) );
}

void read_profiles(environs *environ,FILE *file,char *name)
/*
 */
{
double  value[25];
profs   *prof,*prev,*first,*last = NULL;
int     profi,pnt,num_pnts,num_profs;
int     cols,num_cols;
long    old_line,line = 0;
long    where;
double  range_units,speed_units,depth_units;

   if (Verbose) printf("===> reading <%s> as profile file.\n",name);
   if (Verbose) printf("     range (%s), depth (%s), speed (%s).\n", 
                         unit_name(&Param.units.ssp_range),     
                         unit_name(&Param.units.ssp_depth),     
                         unit_name(&Param.units.ssp_speed) );

   fprintf(Log_file,"===> reading <%s> as profile file.\n",name);
   fprintf(Log_file,"     range (%s), depth (%s), speed (%s).\n", 
                         unit_name(&Param.units.ssp_range),     
                         unit_name(&Param.units.ssp_depth),     
                         unit_name(&Param.units.ssp_speed) );

   range_units = unit_value(&Param.units.ssp_range);     
   depth_units = unit_value(&Param.units.ssp_depth);     
   speed_units = unit_value(&Param.units.ssp_speed);

   do {
      where = ftell(file);                           /* save starting point */
      old_line = line;
      num_cols = read_doubles(file,value,25,&line);  /* read in first line  */
   } while( value[0] >= 0.0 && !feof(file));
   
   while (!feof(file)) {

      num_profs = num_cols -1;

/*** figure out how many points are in this profile ***/

      for( num_pnts = 0; num_pnts < 30000; num_pnts++) {
                                           /*** read in all of the points ***/
         cols = read_doubles(file,value,num_cols,&line);
         if (value[0] < 0.0 ) break;
         if (cols != num_cols ) {
            if (feof(file) ) break;
            printf("Error reading <%s> as prof_file on line %ld.\n",name,line);
            fatal(NULL);
         }
      }
      fseek(file,where,SEEK_SET);
      line = old_line;
      num_cols = read_doubles(file,value,25,&line);  /* read in first line  */

/*** generate linked prof list ***/

      first = prev = NULL;
      for( profi = 1; profi <= num_profs; profi++)  {
         Safe_calloc(prof,profs,1);
         if (prev == NULL) {
            first = prof;
            prof->prev = NULL;
         }
         else {
            prev->next = prof;
            prof->prev = prev;
         }
         prev = prof;
         last = prof;
         prof->range  = value[profi] * range_units;   /* km -> m */
         prof->num_pnts = num_pnts;
         Safe_calloc(prof->depth,double,num_pnts);
         Safe_calloc(prof->raw_speed,double,num_pnts);
         Safe_calloc(prof->speed,double,num_pnts);
      }

      /*** read in all of the points ***/
      
      for( pnt = 0; pnt < num_pnts; pnt++) {  
         cols = read_doubles(file,value,num_cols,&line);
         if (cols != num_cols ) {
            if (feof(file) ) {
               printf("Unexpected end-of-file reading <%s> as prof_file on line %ld.\n",name,line);
               fatal(NULL);
            }
            printf("Error reading <%s> as prof_file on line %ld.\n",name,line);
            fatal(NULL);
         }
         prof = first;
         for(profi = 1; profi <= num_profs; profi++) {
            prof->depth[pnt] = value[0] * depth_units;
            prof->raw_speed[pnt] = value[profi] * speed_units;
            prof = prof->next;
         }
      }

       /*** now add this linked list to environ linked list ***/

      if (environ->prof == NULL) {
         environ->prof       = first;
         environ->last_prof  = last;
      }
      else {
         first->prev = environ->last_prof;
         environ->last_prof->next = first;
         environ->last_prof = last;
      }
      do {
         /* save starting point */
         where = ftell(file);                      
         old_line = line;
         /* read in first line  */
         num_cols = read_doubles(file,value,25,&line);  
      } while( value[0] >= 0.0 && !feof(file));
   }
   if (environ->prof != NULL) return;
   format_msg();
   fatal(NULL);
}
void read_loss(environs *environ,FILE *file,char *name)
/*
 */
{
double  value[25];
losses   *loss,*prev,*first,*last = NULL;
int     lossi,pnt,num_pnts,num_losses;
int     cols,num_cols;
long    old_line,line = 0;
long    where;
double  range_units,angle_units,atten_units;

   if (Verbose) printf("===> reading <%s> as bottom loss file.\n",name);
   if (Verbose) printf("     range (%s), angle (%s), attenuation (%s).\n", 
                         unit_name(&Param.units.loss_range),     
                         unit_name(&Param.units.loss_angle),     
                         unit_name(&Param.units.loss_atten) );

   fprintf(Log_file,"===> reading <%s> as bottom loss file.\n",name);
   fprintf(Log_file,"     range (%s), angle (%s), attenuation (%s).\n", 
                         unit_name(&Param.units.loss_range),     
                         unit_name(&Param.units.loss_angle),     
                         unit_name(&Param.units.loss_atten) );

   range_units = unit_value(&Param.units.loss_range);     
   angle_units = unit_value(&Param.units.loss_angle);     
   atten_units = unit_value(&Param.units.loss_atten);

   do {
      where = ftell(file);                           /* save starting point */
      old_line = line;
      num_cols = read_doubles(file,value,25,&line);  /* read in first line  */
   } while( value[0] >= 0.0 && !feof(file));
   
   while (!feof(file)) {

      num_losses = num_cols -1;

/*** figure out how many points are in this region ***/

      for( num_pnts = 0; num_pnts < 30000; num_pnts++) {
                                           /*** read in all of the points ***/
         cols = read_doubles(file,value,num_cols,&line);
         if (value[0] < 0.0 ) break;
         if (cols != num_cols ) {
            if (feof(file) ) break;
            printf("Error reading <%s> as loss_file on line %ld.\n",name,line);
            fatal(NULL);
         }
      }
      fseek(file,where,SEEK_SET);
      line = old_line;
      num_cols = read_doubles(file,value,25,&line);  /* read in first line  */

/*** generate linked loss list ***/

      first = prev = NULL;
      for( lossi = 1; lossi <= num_losses; lossi++)  {
         Safe_calloc(loss,losses,1);
         if (prev == NULL) {
            first = loss;
            loss->prev = NULL;
         }
         else {
            prev->next = loss;
            loss->prev = prev;
         }
         prev = loss;
         last = loss;
         loss->range  = value[lossi] * range_units;   /* km -> m */
         loss->num_pnts = num_pnts;
         Safe_calloc(loss->angle,double,num_pnts);
         Safe_calloc(loss->atten,double,num_pnts);
      }

      /*** read in all of the points ***/
      
      for( pnt = 0; pnt < num_pnts; pnt++) {  
         cols = read_doubles(file,value,num_cols,&line);
         if (cols != num_cols ) {
            if (feof(file) ) {
               printf("Unexpected end-of-file reading <%s> as loss_file on line %ld.\n",name,line);
               fatal(NULL);
            }
            printf("Error reading <%s> as loss_file on line %ld.\n",name,line);
            fatal(NULL);
         }
         loss = first;
         for(lossi = 1; lossi <= num_losses; lossi++) {
            loss->angle[pnt] = value[0] * angle_units;
            loss->atten[pnt] = value[lossi] * atten_units;
            loss = loss->next;
         }
      }

       /*** now add this linked list to environ linked list ***/

      if (environ->loss == NULL) {
         environ->loss       = first;
         environ->last_loss  = last;
      }
      else {
         first->prev = environ->last_loss;
         environ->last_loss->next = first;
         environ->last_loss = last;
      }
      do {
         /* save starting point */
         where = ftell(file);                      
         old_line = line;
         /* read in first line  */
         num_cols = read_doubles(file,value,25,&line);  
      } while( value[0] >= 0.0 && !feof(file));
   }
   if (environ->loss != NULL) return;
   format_msg();
   fatal(NULL);
}

void read_bathy(environs *environ,FILE *file,char *name)
/*
 *  reads in range (km)  depth (m) pairs of bathymetry information,
 *  one pair per line.  Stops when end-of-file is reached.
 */
{
long i;
double  value[2];
bathys *bathy;
bathys *last = NULL;
int cols;
long line = 0;
double  range_units,depth_units;

   if (Verbose) 
      printf("===> reading <%s> as bathymetry file.\n",name);
   if (Verbose) 
      printf("     range (%s), depth (%s).\n", 
             unit_name(&Param.units.bth_range),     
             unit_name(&Param.units.bth_depth));

   fprintf(Log_file,"===> reading <%s> as bathymetry file.\n",name);
   fprintf(Log_file,"     range (%s), depth (%s).\n", 
             unit_name(&Param.units.bth_range),     
             unit_name(&Param.units.bth_depth));

   range_units = unit_value(&Param.units.bth_range);     
   depth_units = unit_value(&Param.units.bth_depth);     

   for(i=1; !feof(file); i++) {
      Safe_calloc(bathy,bathys,1);
      cols = read_doubles(file,value,2,&line);
      if (cols < 2) {
           break;
      }
      bathy->r  = value[0] * range_units;   /* km -> m */
      bathy->z   = value[1] * depth_units;
      if (last == NULL) environ->bathy = bathy;
      else last->next = bathy;
      last = bathy;
   }
   environ->num_bathy = i;
}

environs *read_inputs(FILE *prof_file,FILE *bath_file,FILE *loss_file)
/*
 *  Creates an environ structure and fills it with information from the
 *  appropriate prof_file and bath_file and loss_file.
 */
{
environs * environ;

   Safe_calloc(environ,environs,1);
   if (prof_file != NULL) {
   
      read_profiles(environ,prof_file,Param.fname.prof);
      fclose(prof_file);
   }
   if (bath_file != NULL) {
      read_bathy(environ,bath_file,Param.fname.bath);
      fclose(bath_file);
   }
   if (loss_file != NULL) {
      read_loss(environ,loss_file,Param.fname.loss);
      fclose(loss_file);
   }
   return(environ);
}

/****************************************************************************/
/*                      end of readray.c                                    */
/****************************************************************************/

